#include "cache.h"
#include "comp.h"
#include "cstack.h"
#include "file.h"
#include "form.h"
#include "graphic.h"
#include "owbasic.h"
#include "shared.h"


#ifndef SMALL_NO_CACHE /* caching functions (MB,4.50) */

#define  CACHE_VERSION  0x000B
#define  CACHE_SIG      stringstack2
#define  CACHE_SIGSIZE  (NAMELEN + 18)
#ifdef __HITACHI__
 #define CACHE_BUFSIZE  (1024 * 32 - 24)
 #define CACHE_OFFSET   (int*) (((int) (FBB + CACHE_SIGSIZE) + 3) & ~3)
#endif
#ifdef LSI_C
 #define CACHE_BUFSIZE  (1024 * 3 - 24)
 #define CACHE_OFFSET   (FBB + CACHE_SIGSIZE)
#endif
#define  CACHE_LASTELEM ((int*) (FBB + CACHE_BUFSIZE) - 1)


static void NEAR cache_getname (void);
static void NEAR cache_init (void);

int              cache_find (void);

int              cache_load (void);
int              cache_save (void);

static int NEAR  cache_geti (void);
static void NEAR cache_puti (int i);

int*     cache_bptr;
int      cache_fnr;

char FAR cache_filename[] = PROGNAME ",cache%02.2i,%s";


static void NEAR cache_getname (void)
{
  sprintf ((char*) CACHE_SIG, Near (cache_filename), cache_fnr++, (char*) (progname));
}

int cache_find (void)
{
  int found = FALSE;
  int fwd = 2;

  while (fwd > 0)
  {
    --fwd;
    while ((!found) && (LibFileFindNext (&fb, &finf, 0x00)))
    {
      if (!LibFileReadEx (&fb, &finf, FILE_BLOCK_SIZE /* load first block only */))
        return (FALSE); /* Read error */
      found = (strcmp ((char*) CACHE_SIG, (char*) (FBB)) == 0);
    }

    if (!found)
    {
      finf.fp = 0xFFFF; /* prepare FILE_INFO for "write new" or for search from beginning */
      finf.kind = FILE_KIND_BIN;
    }
  }

  return (found);
}

static void NEAR cache_init (void)
{
  cache_getname ();
  sffindfile (&fb);
  cache_find ();
  cache_bptr = (int*) CACHE_OFFSET;
  strcpy ((char*) FBB, (char*) CACHE_SIG);
}

int cache_save (void)
{
  int res;

  FBBNUM = ((uchar*) (cache_bptr)) - FBB;
  res = LibFileWrite (&fb, &finf);
  cache_init ();

  return (res);
}

int cache_load (void)
{
  cache_getname ();
  sffindfile (&fb);
  cache_find ();
  LibFileReadEx (&fb, &finf, (CACHE_BUFSIZE + 24) / FILE_BLOCK_SIZE); /* load all blocks */
  cache_bptr = (int*) CACHE_OFFSET;
}

static void NEAR cache_puti (int i)
{
  if (cache_bptr > CACHE_LASTELEM)
    cache_save ();

  *(cache_bptr++) = i;
}

static int NEAR cache_geti (void)
{
  if (cache_bptr > CACHE_LASTELEM)
    cache_load ();

  return (*(cache_bptr++));
}

int cache_read (void)
{
  int* c_pbin = (int*) bin;
  int* c_pbmax;
  GlobalVar FAR* gv;
  type_t type;

  int FAR* c_pxdata = xdata;
  int FAR* c_pxmax;
  int FAR* c_px2;

  int c_tmp;

  cache_fnr = 0;
  finf.fp = 0xFFFF; /* search from beginning */
  finf.kind = FILE_KIND_BIN; /* binary files */

  if (!cache_load ())
    return (FALSE);

  /*
   * cache file:
   *   uint file_version;
   *   uint owb_version;
   *   uint desired_version;
   *
   *   uint binstart;
   *   uint binsize, datanext, firstconst;
   *
   *   int  global_variable_addresses[...];
   *
   *   uint bin[binsize];
   *   int  data[datanext];
   *   int  const[DATASIZE - firstconst];
   */

  if (( cache_geti ()) != CACHE_VERSION)
    return (FALSE);
  if (cache_geti () != OWBVersion)
    return (FALSE);
#ifndef SMALL_NO_OLDCRCMODE
  if ((desired_version = cache_geti ()) < 325)
    default_hash_mode = 2;
#endif

  bin_start.c = bin + cache_geti ();
  c_tmp = cache_geti (); datanext = cache_geti (); firstconst = cache_geti ();
  pc.c = bin + c_tmp;

  for (gv = vartable; type = gv->type; ++gv)
    if (!(type & KCONST))
    {
      if (type & KPOINTER)
        *(gv->u.ptr) = xdata + cache_geti ();
      else
        *(gv->u.index) = cache_geti ();
    }

  /* read bin data */
  for (c_pbmax = (int*) &bin[c_tmp]; c_pbin < c_pbmax; ++c_pbin)
    *c_pbin = cache_geti ();

  /* read compressed xdata_var */
  for (c_pxmax = &xdata[datanext]; c_pxdata < c_pxmax;)
  {
    c_tmp = cache_geti ();
    if (c_tmp == 0) /*next number contains amount of following 0s */
    {
      for (c_px2 = c_pxdata + cache_geti (); c_pxdata < c_px2; ++c_pxdata)
        *c_pxdata = 0;
    }
    else
      *(c_pxdata++) = c_tmp;
  }

  /* read xdata_const */
  for (c_pxmax = &xdata[DATASIZE], c_pxdata = &xdata[firstconst]; c_pxdata < c_pxmax; ++c_pxdata)
    *c_pxdata = cache_geti ();

  return (TRUE);
}

int cache_write (void)
{
  int* c_pbin = (int*) bin;
  int* c_pbmax;

  int FAR* c_pxdata = xdata;
  int FAR* c_pxmax;
  GlobalVar FAR* gv;
  type_t type;

  int c_tmp;

  cache_fnr = 0;

  /*
   * cache file:
   *   uint file_version;
   *   uint owb_version;
   *   uint desired_version;
   *
   *   uint binstart;
   *   uint binsize, datanext, firstconst;
   *
   *   int  global_variable_addresses[...];
   *
   *   uint bin[binsize];
   *   int  data[datanext];
   *   int  const[DATASIZE - firstconst];
   */

  cache_init ();

  cache_puti (CACHE_VERSION);
  cache_puti (OWBVersion);

#ifndef SMALL_NO_OLDCRCMODE
  cache_puti (desired_version);
#endif

  cache_puti (bin_start.c - bin);
  cache_puti (pc.c - bin); cache_puti (datanext); cache_puti (firstconst);

  for (gv = vartable; type = gv->type; ++gv)
    if (!(type & KCONST))
      cache_puti ((type & KPOINTER) ? (*(gv->u.ptr) - xdata) : *(gv->u.index));


  /* write binary data (program codes) */
  for (c_pbmax = (int*) (pc.c); c_pbin < c_pbmax; ++c_pbin)
    cache_puti (*c_pbin);

  /* write compressed xdata */
  for (c_pxmax = &xdata[datanext]; c_pxdata < c_pxmax; ++c_pxdata)
  {
    cache_puti (*c_pxdata);
    if (*c_pxdata == 0) /* next number contains amount of following 0s */
    {
      for (c_tmp = 0; (c_pxdata < c_pxmax) && (*c_pxdata == 0); ++c_tmp, ++c_pxdata)
        ;
      cache_puti (c_tmp);
      --c_pxdata; /* incremented in outer loop again */
    }
  }

  /* write constants in xdata */
  for (c_pxmax = &xdata[DATASIZE], c_pxdata = &xdata[firstconst]; c_pxdata < c_pxmax; ++c_pxdata)
    cache_puti (*c_pxdata);

  /* save last file */
   cache_save ();
}

#endif
