#include "modeldef.h"
#include <string.h>
#include "owbasic.h"
#include "cache.h"
#include "comp.h"
#include "cstack.h"
#include "file.h"
#include "shared.h"
#include "form.h"
#include "ostring.h"
#include "graphic.h"
#include "rc.h"


uchar* prog;          /* the program pointer for compile */
uchar* aprog;         /* pointer for debug output lex () */
uchar* prog_start;    /* start of the program */

uchar*     token_start;
uchar*     token_end;
int        token_type;
int        tok;
datatype_t datatype;

/* variables for overlay managment */
bool a_overlay;
PC   a_overlaystart;
PC   a_overlayend;

jmp_buf errorret;

int last_proc;

/* --- include stack --- */
SourcePtr  IncStack[INCSTACKSIZE];
SourcePtr* IncStackPtr;

PC nullpc = {NULL};

int ctos;               /* index to top of control stack */
type_t default_datatype;
int compile_procedure (void);

int desired_version = OWBVersion;

int FAR* cpp; /* pointer to current procedure definition */

int varlistaddr; /* symbol table at start of procedure */


/*#ifndef LSI_C
 uchar _macro_expansion_space[MACRO_EXPANSION_SIZE + 2];
#endif*/


type_t DT (type_t dtyp)
{
  /*
     apply default type of identifiers
     if type is unknown (not given)
  */
  if ((dtyp & TYPEMASK) != DUNKNOWN)
    return (dtyp);
  return ((dtyp & ~TYPEMASK) | default_datatype);
}

static uchar FAR cstack_fl[] = "bc";
static csstype FAR cstack_ff[] = {C_BRK, C_CNT};
void compile_break (int jmp, int n) /* break and continue in control structure (MB,4.40) */
{
  char nls[IDENTLEN];
  unsigned int loc;
  int ref;
  datatype_t type = {DREFERENCE, 0};

  nls[0] = '~';
  nls[1] = cstack_fl[n];
  ref = cstack[ctos - 1].brkuse & C_NMASK;
  cstack[ref].brkuse |= cstack_ff[n];
  sprintf (nls + 2, "%d", cstack[ref].brkcnt);

  loc = find_label ((uchar*) nls);
  ref = newentry (type, (uchar*) "");
  xdata[ref] = loc;
  xdata[ref + 1] = (int) pc.p;
  ADDJMP (jmp, 0);
}

void cstack_nextlabel (cstack_entry* cs, int n)
{
  char nls[IDENTLEN];

  if (cs->brkuse & cstack_ff[n])
  {
    sprintf (nls + 2, "%d", cs->brkcnt);
    nls[0] = '~';
    nls[1] = cstack_fl[n];
    xdata[find_label ((uchar*) nls)] = (int) pc.c; /* position of label */
  }
}

static signed char d_cnt = 1;
static signed char d_pos = 1;

int compile_command (void)
{
  bool is_fptrcall = FALSE;
  int i;
  datatype_t dtype = {DUNKNOWN, 0};
  int rc = RC_CONTINUE;
  uchar name[IDENTLEN];
  CompileFunctionTable FAR* ctbl = comptable;

  if (LEX_CHAR == 0)
    return (RC_EXIT);

  if (LEX_CHAR == '#') /* compile directive (MB,4.40) */
  {
    LEX_INCCHAR ();
    lex ();
    tok = look_up (commandtable);
    if (tok == ENDIF)
    {
      if (d_cnt < 0)
      {
        if ((d_cnt < 0) && (d_cnt == d_pos))
        {
          ++d_cnt;
          d_cnt = -d_cnt;
        }
        else
          ++d_cnt;
      }
      else
        --d_cnt;
      return (RC_CONTINUE);
    }
    if (tok != IF)
    { /* constant assignment */
      dtype.type = DCONSTANT;
      i = findentry_token (&dtype);
      xdata[i + 1] = DINTEGER;
      xdata[i + 2] = 0;
      expect ('=');
      xdata[i] = expect_simple_constexp (DINTEGER);
      return (RC_CONTINUE);
    }
    if (d_cnt < 0)
    {
      --d_cnt;
      prognl ();
      return (RC_CONTINUE);
    }
    ++d_cnt;

    if (expect_simple_constexp (DBOOLEAN) == 0)
      d_pos = d_cnt = -d_cnt;

    return (RC_CONTINUE);
  }

  if (d_cnt < 0)
  {
    prognl ();
    ++ITOS.line; /* TODO: is this necessary? */
    return (RC_CONTINUE);
  }
  lex ();
  if ((tok == '(') || (tok == '*'))
  {
    compile_assignment ();
    return (RC_CONTINUE);
  }
  if (token_type == 0)
    serror (0);
  switch (token_type)
  {
  case TNUMBER:
    /* numbers as labels */
    if (datatype.type != DINTEGER)
      serror (0);
  case TLABEL:
    /* labels starting with '$' */
    i = find_token_label ();
    if (xdata[i] != 0)
      serror (6); /* label exists */
    xdata[i] = (int) pc.c; /* position of label */
    break;
  case TKEY:
    switch (tok)
    {
    case PROP_FBBSIZE:
    case CLIPBOARD:
    case FILEBUFFER: compile_assignment (); break;
    case WHILE:      compile_while ();      break;
    case ELSE:       compile_else ();       break;
    case REM:        /* nothing to do */;   break;
    case PRINT:      compile_print ();      break;
    default:         serror (0);
    }
    break;
  case TIDENT:
    if ((LEX_CHAR == '=') || (LEX_CHAR == '[') || (LEX_CHAR == '{'))
      compile_assignment ();
    else
    { /* commands and procedures */
      tok = look_up (commandtable);

      for (; ctbl->tok; ++ctbl) /* search comptable for command */
        if (ctbl->tok == tok)
        {
          (ctbl->func) ();
          tok = -1;
          break;
        }

      switch (tok)
      {
      case -1: /* found in comptable */
        break;

      case BREAK:
      case CONTINUE:
        compile_break (JMP, tok - BREAK);
        break;

      case GOTO:
        compile_goto (JMP);
        break;

      case PROC:
        compile_procfunc (FALSE);
        break;
      case FUNC:
        compile_procfunc (TRUE);
        break;
      case ENDP:
        compile_endp (FALSE);
        break;
      case FRETURN:
        if (cpp)
          comp_simple_exp (*(cpp - 2));
        ADDCODE (B_RETURN);
        break;
      case RETURN:
        if (cpp)
          compile_endp (TRUE);
        else
          ADDCODE (B_RETURN);
        break;

      case INCLUDE:
        if (compile_include (FALSE))
          rc = RC_INCLUDE;
        break;
      case APPEND:
        compile_include (TRUE);
        rc = RC_APPEND;
        break;

      case CALL:
        is_fptrcall = TRUE;
      case UNKNOWN:
        if (!compile_userprocedure (NULL, is_fptrcall))
          if (!compile_procedure ())
            serrortoken (2040); /* Unknown Procedure */
        break;
      case DEFAULT:
        lex ();
        if (datatype.type == DUNKNOWN)
        {
          tokcpy_ident (name);
          if (strcmp_nf (name, FN_INT) == 0)
            default_datatype = DINTEGER;
          if (strcmp_nf (name, FN_FLOAT) == 0)
            default_datatype = DFLOAT;
          if (strcmp_nf (name, FN_STRING) == 0)
            default_datatype = DSTRING;
        }
        else
          default_datatype = datatype.type;
        break;
      default:
        serrortoken (2020);
      }
    }
    break;
  case TDELIM: /* OK */
    break;
  default:
    serror (0);
  }
  return (rc);
}

/* ---------------------- compiler ----------------------*/

void compile_enum (void)
{
  int addr;
  register int caddr;
  uchar name[IDENTLEN];
  datatype_t type = {DTYPEDEF, DENUM};
  datatype_t etype = {DINTEGER, 0};
  int constval = 0;

  expect (IDENT);
  tokcpy_ident (name);
  if (searchentry_name (&type, name) >= 0)
    serror (12);
  addr = newentry (type, name);
  expect ('=');
  expect ('(');
  type.type = DCONSTANT;
  type.ext = 0;
  while (lex () == IDENT)
  {
    caddr = findentry_token (&type);
    xdata[caddr] = constval++;
    xdata[caddr + 1] = DENUM;
    xdata[caddr + 2] = addr;
    if (LEX_CHAR == '=')
    {
      LEX_INCCHAR ();
      constval = (xdata[caddr] = expect_constexp (&etype)) + 1;
    }
    if (LEX_CHAR == ',')
      lex ();
  }
  if (tok != ')')
    serror (1000 + IDENT);
}

#if FALSE
void compile_macro (void)
{
  char name[IDENTLEN];
  char buf[IDENTLEN];
  char paranames[IDENTLEN * MAXPARAM];
  int paracount = 0;
  char* sptr = MACRO_EXPANSION_SPACE;
  char* tptr = MACRO_EXPANSION_SPACE;
  char FAR* dptr;
  int i;
  bool par;
  type_t type = DMACRO;

  expect (IDENT);
  tokcpy_ident (name);

  if (searchentry_name (&type, name) >= 0)
    serror (12);

  if (LEX_CHAR == '=')
  {
    lex ();
    paracount = -1;
  }
  else
  {
    expect ('(');
    while (lex () != ')')
    {
      if (tok != IDENT)
        serror (1000 + IDENT);
      tokcpy_ident (paranames + (IDENTLEN * paracount));
      ++paracount;
      if (LEX_CHAR == ',')
        lex ();
    }
  }

  while (!isdelim (LEX_CHAR))
  {
    if (ismalpha (LEX_CHAR))
    {
      par = FALSE;
      lex ();
      tokcpy_ident (buf);
      for (i = 0; i < paracount; ++i)
        if (strcmp (buf, paranames + (IDENTLEN * i)) == 0)
        { /* macro parameter */
          par = TRUE;
          *(tptr++) = 128 + i; /* char >128: parameter */
          break;
        }
      else
      {
        tokcpy_ident (tptr);
        tptr += strlen (tptr) + 1;
      }
    }
    else
      *(tptr++) = LEX_CHARPP ();
  }
  *tptr = 0;
  dptr = (char FAR*) custom_data (newentry_custom (DMACRO, name, (tptr + 1 - MACRO_EXPANSION_SPACE + 1) / sizeof (int)));
  while (sptr <= tptr)
    *(dptr++) = *(sptr++);
}
#endif


void compile_assignment (void)
{
  datatype_t ltype;
  type_t rtype;
  bool is_fb;
  int var = -1;
  bool schar = FALSE;
  PC cpc = pc;


  if ((is_fb = (tok == FILEBUFFER)) || (tok == CLIPBOARD))
  { /* special case filebuffer */
    expect ('[');
    comp_simple_exp (DINTEGER);
    expect (']');
    expect ('=');
    comp_simple_exp (DINTEGER);
    ADDCODE (is_fb ? B_FBPUT : B_CBPUT);
    return;
  }
  if (tok == PROP_FBBSIZE)
  {
    expect ('=');
    comp_simple_exp (DINTEGER);
    ADDCODE (B_SETFBBSIZE);
    return;
  }

  if (comp_lvalue (&ltype))
    var = varaddr;

  rtype = typesize (ltype.type);
  if ((var >= 0) && (rtype == DINTEGER))
    pc = cpc;

  if (ltype.type == DSTRING)
  {
    if (LEX_CHAR == '{')
    {
      /* destination is char of string */
      lex ();
      comp_simple_exp (DINTEGER);
      expect ('}');
      ltype.type = DCHAR;
      schar = TRUE;
    }
  }

  /* check the = */
  expect ('=');

  /* get the value */
  comp_exp (&ltype);

  /* assign the value to the variable */
  if (schar)
  {
    ADDCODE (B_ASSIGN_SC);
  }
  else
    switch (rtype)
    {
    case DINTEGER: /* including integer sized types */
      if (var >= 0)
      {
        ADDCODE (B_ASSIGN);
        ADDVAL (var);
      }
      else
      {
        ADDCODE (B_ASSIGN_I);
      }
      break;
      case DFLOAT:
        ADDCODE (B_ASSIGN_F);
        break;
      case DSTRING:
        ADDCODE (B_ASSIGN_S);
        break;
    }
}

void compile_wait (void)
{
  comp_simple_exp (DINTEGER);
  ADDCODE (B_WAIT);
  ADDCODE (B_WAIT2);
}

void compile_print (void)
{
  byte newline = TRUE;
  datatype_t atyp;

  lex ();
  while (token_type != TDELIM) 	/* get next list item */
  {
    if (token_type == TEDELIM)  /* ';' or ',' */
    {
      if (tok == ',')
        ADDCODE (B_PRINT_TAB);
      newline = FALSE;
    }
    else                      /* expression */
    {
      putback ();
      comp_exput (&atyp);
      switch (atyp.type & TYPEMASK)
      {
      case DINTEGER:
        ADDCODE (B_PRINT_INT);
        break;
      case DFLOAT:
        ADDCODE (B_CONV_FS);
        ADDCODE (B_PRINT_STRING);
        break;
      case DSTRING:
        ADDCODE (B_PRINT_STRING);
        break;
      case DCHAR:
        ADDCODE (B_PRINT_CHAR);
        break;
      case DBOOLEAN:
      ADDCODE (B_PRINT_BOOLEAN);
        break;
      default:
        serror (22);
      }
      newline = TRUE;
    }
    lex ();
  }
  if (newline)
    ADDCODE (B_PRINT_NL);
  ADDCODE (B_PRINT_END);
}

bool compile_include (bool append)
{
  int i;
  uchar h[NAMELEN + 2];

  lex ();
  if ((token_type != TIDENT) && (token_type != TSTRING))
    serror (1000 + IDENT);
  if (append)
    return (TRUE);
  /* check if already included */
  h[0] = '#';
  if (token_type == TIDENT)
    tokncpy (&h[1], NAMELEN);
  else
    sconstncpy (&h[1], NAMELEN);
  i = find_label (h);
  if (xdata[i] != 0)
    return (FALSE);
  xdata[i] = 1;
  return (TRUE);
}

void compile_overlay (void)
{
  if (cpp)
    serror (0); /* no OVERLAY in proc declaration */

#ifdef LSI_C
  if (expect_simple_constexp (DINTEGER) == 1)  /* 0 - OFF, 1 - ON */
  {
    overlays_used = TRUE;
    /* start (new) overlaysegment */
    a_overlay = TRUE;
    a_overlayend = a_overlaystart = pc; /* segment length is zero */
  }
  else /* the following procedures are not overlayed */
    a_overlay = FALSE;
#endif
}

/*-------------------------------------------------------*/

/*static bool NEAR command_contains (uchar token)
{
  char* cprog = prog;

  for (; !isdelim (*cprog); ++cprog)
    if (*cprog == token)
      return TRUE;
  return FALSE;
}*/


void compile_procfunc (bool func)
{
  uchar pname[IDENTLEN];
  int htok;
  datatype_t dtyp, etype;
  int var;
  int         paracount;
  int         paravar[MAXPARAM]; /* variables for parameters */
  type_t FAR* paratype;
  int FAR*    paraval = NULL;
  int FAR*    paraext = NULL;
  int procnr; /* reference to procedure data */
  byte old_dr = declaration_required;
  bool default_args = FALSE;
  uchar* cprog;


  /*dtyp.ext = 0;*/
  declaration_required = 0;

  if (cpp) /* does already a procedure definition run? */
    serror (48);

  if (pc.c != bin_start.c) /* is there code outside proc. before this definition ?*/
    serror (50);

  expect (IDENT);   /* get name of procedure */

  if (func)
  {
    dtyp.type = DT (datatype.type);
    if (lex_ref == 1)
      dtyp.type |= KPOINTER; /* return value is pointer */
  }
  else
    dtyp.type = DPROCEDURE;

  tokcpy_ident (pname);

  if (search_procfunc (pname, dtyp.type) >= 0)
    serror (52); /* proc/func already exists */

  procnr = new_procfunc (pname, pc, dtyp.type); /* register new proc */
  cpp = &xdata[procnr]; /* procedure which is currently defined */

  etype.type = DDATA;
  etype.ext = 0;

  for (cprog = prog; !isdelim (*cprog); ++cprog)
  {
    if ((*cprog == '=') && !paraval)
      paraval = &xdata[cpp[2] = (newentry_custom (etype, (uchar*) "", MAXPARAM) + 5)];
    if ((*cprog == '\\') && !paraext)
      paraext = &xdata[cpp[3] = (newentry_custom (etype, (uchar*) "", MAXPARAM) + 5)];
  }

  varlistaddr = listnext;     /* begin of local variables in symbol table */
  paratype = (type_t FAR*) (custom_data (procnr));
  ++last_proc;

  if (a_overlay) /* are we in overlay mode ? */
  {
    pc = a_overlaystart; /* start compiling again at overlay startposition */
    *cpp = (int) pc.c;
    ADDVAL (last_proc);     /* store procedure number as first value to
                            identify the loaded segment */
    ADDVAL (0);          /* dummy; the procedure length is stored here later */
  }

  htok = lex ();
  dtyp = datatype; /* get first parameter */
  paracount = 0;

  while ((htok != DELIMITER) && (htok != END))
  {
    if (htok == VAR)  /* VAR found */
    {
      htok = lex (); /* next word is parameter name */
      dtyp.type = datatype.type | KREFERENCE;
      dtyp.ext = datatype.ext;
    }

    if (token_type != TIDENT)
      serror (1000 + IDENT);  /* name of parameter */
    if (paracount == MAXPARAM)
      serror (46); /* too much parameters */

    if (LEX_CHAR == '[') /* parameter is array */
    {
      dtyp.type |= KARRAY;
      if (!(dtyp.type & KREFERENCE))
        serror (64);
    }

    if (searchentry_token (&dtyp) >= 0) /* exist global variable ? */
      serrortoken (2028);

    var = findentry_token (&dtyp);

    paratype[paracount] = dtyp.type; /* register type of par. */
    paravar[paracount] = var; /* pointer to parameter */
    if (paraext)
      paraext[paracount] = dtyp.ext;

    if (dtyp.type & KARRAY)
    {
      lex (); /* get '[' */
      expect (']');
    }


    if (LEX_CHAR == '=') /* default parameter (MB,4.50) */
    {
      lex (); /* get '=' */
      default_args = TRUE;
      paratype[paracount] |= KLAST;
      etype.type = dtyp.type & TYPEMASK;
      etype.ext = dtyp.ext;
      if (dtyp.type & KREFERENCE)
      {
        etype.type |= KPOINTER;
        paraval[paracount] = expect_constexp (&etype);
      }
      else if (etype.type == DSTRING)
      {
        if (lex () != STRING)
          serror (22);
        comp_str ();
        paraval[paracount] = firstconst;
      }
      else
      {
        expect_constexp (&etype);
        if (dtyp.type & DFLOAT)
        {
          fconst (lexfloat);
          paraval[paracount] = firstconst;
        }
        else
          paraval[paracount] = lexint;
      }
    }
    else
      if (default_args)
        serror (19);


    htok = lex ();
    ++paracount;
    if (htok == ',')
      htok = lex ();
    dtyp = datatype;
  }

  paratype[paracount] = DUNKNOWN; /* mark end of para. list */

  /* add code to get parameters from stack */
  while (paracount > 0)
  {
    switch (typesize (paratype[--paracount]))
    {
    case DINTEGER: /* including integer sized types */
      ADDCODE (B_POPI);
      break;
    case DFLOAT:
      ADDCODE (B_POPIF);
      break;
    case DSTRING:
      ADDCODE (B_POPIS);
      break;
    default:
      serror (22);
    }

    ADDVAL (xdata[paravar[paracount]]);
  }

  declaration_required = old_dr;
}

void compile_endp (bool func)
{
  /* endp for procedures OR return for functions */
  if (cpp == NULL)
    serror (49); /* only valid in procedure definitions */
  if (func)
  {
    if (isdelim (LEX_CHAR))
    {
      ADDCODE (B_RETURN);
      return;
    }
    comp_simple_exp (*(cpp - 2)); /* compile expression (value on stack) */
  }
  if (ctos != 0)
    notclosederror (cstack[ctos - 1].type); /* unclosed control in procedure */
  ADDCODE (B_RETURN);       /* add code for return */
  listhide (varlistaddr);   /* hide local variables in symbol list */
  if (a_overlay) /* is procedure compiled in overlaymode ?*/
  {
    /**(cpp + 1) = pc.b - a_overlaystart.b;*/ /* length of procedure */
    *(cpp + 1) = last_proc;
    if (pc.b > a_overlayend.b)
      a_overlayend = pc; /* correct overlay segment end if neccessary */
    pc = a_overlayend; /* set pc to overlay segment end, in case normal code
                          follows */
    /* save overlay code to flash ram */
    if (!Overlay (OV_SAVE, last_proc, a_overlaystart.c, pc.b - a_overlaystart.b))
    /*if (!SaveOverlay (a_overlaystart.c, pc.b - a_overlaystart.b, last_proc))*/
      serror (68);
    *(a_overlaystart.i + 1) = a_overlayend.b - a_overlaystart.b; /* set procedure length */
  }
  bin_start = pc;            /* main starts after procedures */
  cpp = NULL;                /* not in procedure definition */
}


void comp_drop (type_t ltype)
{
  switch (typesize (ltype))
  {
  case DINTEGER:
    ADDCODE (B_DROP);
    break;
  case DFLOAT:
    ADDCODE (B_DROPF);
    break;
  case DSTRING:
    ADDCODE (B_DROPS);
    break;
  }
}

bool compile_userprocedure (type_t* ltype, bool is_fptrcall)
{
  /* call for user procedure (ltype==NULL) OR function (ltype!=NULL) */
  int procnr;
  type_t hdtyp, rtype;
  uchar pname[IDENTLEN + 1];
  int FAR *pp;
  uchar* rpos;
  uchar* vpos;
  PC pcp;
  datatype_t etype;
  type_t FAR* paratype;
  type_t pltype = DPROCEDURE;


  if ((datatype.type != DUNKNOWN) && (ltype == NULL)) /* function called as procedure (MB,4.25) */
  {
    pltype = datatype.type;
    ltype = &pltype;
  }
  if (is_fptrcall)
  {
    rtype = datatype.type;
    if (pltype != DPROCEDURE)
      expect ('('); /* function pointer call */
    vpos = prog;
    pcp = pc;
    etype.type = DFPTR;
    etype.ext = 0; /* will be assigned by comp_exp() */
    comp_exp (&etype);
    procnr = xdata[etype.ext];
    pc = pcp;
    datatype.type = rtype;
  }
  else
    tokcpy_ident (pname);

  if (ltype == NULL)
    hdtyp = DPROCEDURE;
  else
    *ltype = hdtyp = (DT (datatype.type) | ((lex_ref == 1) ? KPOINTER : 0));

  if (!is_fptrcall)
    procnr = search_procfunc (pname, hdtyp);  /* find user procedure */

  if (procnr < 0)
  {
    if (pltype != DPROCEDURE)
    {
      etype.type = *ltype;
      compile_function (&etype); /* call of built-in function as procedure */
      comp_drop (pltype);
      return (TRUE);
    }
    else
      return (FALSE);
  }

  pp = &xdata[procnr];
  paratype = (type_t FAR*) (custom_data (procnr));

  /* compile parameters */
  if (ltype == NULL)
    comp_para (paratype, pp, NULL, DELIMITER, 1);
  else
  {
    if (!is_fptrcall)
      expect ('(');
    comp_para (paratype, pp, NULL, ')', 1);
  }

  if (*(pp + 1) > 0) /* is procedure overlayed ?*/
  {
    if (cpp) /* if within procedure definition
                we have to check, that no procedure of the same
                overlay segment is called */
    {
      if (*pp == *cpp) /* procedure in same overlay can't be called */
        serror (70);
    }
    ADDCODE (B_CALL_OVERLAY); /* call of an overlayed procedure */
    ADDVAL (*(pp + 1)); /* the number of the procedure */
    pcp.c = (codet*) (*pp);
    ADDCODEP (pcp); /* the address where to put it */
  }
  else
  {
    if (is_fptrcall) /* is fptr */
    {
      rpos = prog;
      prog = vpos;
      comp_simple_exp (DFPTR);
      prog = rpos;
      ADDCODE (B_CALL_FPTR);
    }
    else /* normal procedure */
    {
      ADDCODE (B_CALL);
      pcp.c = (codet*) (*pp);
      ADDCODEP (pcp); /* code for call */
    }
  }

  if (pltype != DPROCEDURE)
    comp_drop (pltype);
  return (TRUE);
}

/*----------------------------------------------------------------*/
int find_token_label (void)
{
  uchar htoken[IDENTLEN];
  if (*token_start == '$')
    ++token_start;
  tokcpy_ident (htoken);
  return (find_label (htoken));
}

int find_label (uchar* s)
{
  datatype_t ht = {DLABEL, 0};
  return (findentry_name (&ht, s));
}


void compile_goto (int jmp)
{
  unsigned int loc;
  int ref;
  datatype_t type = {DREFERENCE, 0};

  expect_label ();
  loc = find_token_label ();
  if (xdata[loc] == 0) /* label not (yet) defined */
  {
    ref = newentry (type, (uchar*) ""); /* has to be linked later */
    xdata[ref] = loc;
    xdata[ref + 1] = (int) pc.p;
    ADDJMP (jmp, 0);
  }
  else
  {
    ADDJMP (jmp, xdata[loc]); /* create jump */
  }
}

void compile_gosub (void)
{
  int loc, ref;
  PC addr;
  datatype_t type = {DREFERENCE, 0};

  expect_label ();

  ADDCODE (B_CALL); /* create subroutine call */
  loc = find_token_label (); /* find destination */
  if (xdata[loc] == 0) /* not (yet) defined */
  {
    ref = newentry (type, (uchar*) ""); /* has to be linked later */
    xdata[ref] = loc;
    xdata[ref + 1] = (int) pc.c;
    ADDCODEP (nullpc); /* dummy destination */
  }
  else
  {
    addr.c = (codet*) xdata[loc];
    ADDCODEP (addr); /* call address */
  }
}

/*----------------------------------------------------*/

void compile_local (void)
{
  byte old_dr = declaration_required;

  lex ();
  if ((tok == NUMBER) && (datatype.type == DINTEGER))
  {
    declaration_required = lexint;
    return;
  }
  putback ();

  declaration_required = 0;

  while (lex () == IDENT)
  {
    if (lex_ref == 1)
      datatype.type |= KREFERENCE;
    findentry_token (&datatype);
    if (LEX_CHAR == ',')
      lex ();
  }

  declaration_required = old_dr;
}

void compile_dim (void)
{
  uchar name[IDENTLEN];
  byte old_dr = declaration_required;
  datatype_t dtype;
  int vaddr;
  long int size;
  int rowsize;
  bool is_var = FALSE;

  is_var = FALSE;
  declaration_required = 0;

  if (LEX_CHAR == '&')
  {
    is_var = TRUE;
    LEX_INCCHAR ();
  }
  while (lex () == IDENT)
  {
    dtype = datatype;
    tokcpy_ident (name);
    if (is_var)
      dtype.type |= KREFERENCE;
    vaddr = findentry_name (&dtype, name);
    expect ('[');
    if ((size = (expect_simple_constexp (DINTEGER) + 1)) < 1)
      serror (33);
    if (LEX_CHAR == ',') /* 2-dimensional array */
    {
      lex (); /* skip ',' */
      rowsize = expect_simple_constexp (DINTEGER) + 1;
      size *= rowsize;
    }
    else
      rowsize = 0;
    if (((dtype.type & KREFERENCE) != 0) && !is_var)  /* special case: procedure parameter
                                                    old style: array parameters are declared
                                                    with DIM within procedure */
    {
      xdata[vaddr + 1] = size;
      xdata[vaddr - 3] |= KARRAY;
    }
    /* normal case */
    else
    {
#ifdef LSI_C
      if (size > 0x7FFF)
        serror (33);
#endif
      expandentry (vaddr, size, rowsize);
    }
    expect (']');
    if (LEX_CHAR == ',')
      lex ();
    if (is_var = (LEX_CHAR == '&'))
      LEX_INCCHAR ();
  }

  declaration_required = old_dr;
}

void read_list (uchar* name, datatype_t* dtype)
{
  int vaddr;
  int i;
  int FAR *ic;
  FLOATT FAR *fc;
  int hi;
  FLOATT hf;
  type_t stype;
  int count = 0;
  datatype_t rtype = *dtype, utype = *dtype;

  stype = typesize (dtype->type);
  if (dtype->type & KREFERENCE)
    utype.type = (dtype->type & ~KREFERENCE) | KPOINTER;

  if (utype.type == DCHAR) /* char array initialized with string (MB,4.10) */
  {
    if (LEX_CHAR != '"')
    {
      if (LEX_CHAR == '\'')
        iconst (lex ());
      else
        serror (22);
    }
    else
    {
      while (LEX_PPCHAR () != '"')
      {
        if (LEX_CHAR == '\\')
          LEX_INCCHAR ();
        ++count;
        iconst (LEX_CHAR);
      }
      LEX_INCCHAR ();
    }
    expect (')');
  }
  else
  {
    lex ();
    while (tok != ')')
    {
      switch (stype)
      {
      case DINTEGER: /* and integer sized types */
        putback ();
        iconst (expect_constexp (&utype));
        break;
      case DSTRING: /* initialized string arrays (MB,4.25) */
        if (token_type != TSTRING)
          serror (0);
        iconst (comp_str ());
        i2push (firstconst - 2);
        break;
      case DFLOAT:
        putback ();
        expect_simple_constexp (DFLOAT);
        fconst (lexfloat);
        break;
      }
      ++count;
      if (LEX_CHAR == ',')
        lex ();
      if ((LEX_CHAR == 0x0A) || (LEX_CHAR == 0x0D))
        lex ();
      lex ();
    }
  }
  /* not nice, but works :
     invert order of constants
  */

  switch (stype)
  {
  case DSTRING:
    for (i = 0; i < count; ++i)
      iconst (i2pop ());
    break;
  case DINTEGER:
    ic = &xdata[firstconst];
    for (i = 0; i < count / 2; ++i)
    {
      hi = ic[i];
      ic[i] = ic[count - 1 - i];
      ic[count - 1 - i] = hi;
    }
    break;
  case DFLOAT:
    fc = (FLOATT FAR*) (&xdata[firstconst]);
    for (i = 0; i < count / 2; ++i)
    {
      hf = fc[i];
      fc[i] = fc[count - 1 - i];
      fc[count - 1 - i] = hf;
    }
    break;
  }

  rtype.type |= KCONST;
  if (searchentry_name (&rtype, name) >= 0)
    serror (28);
  vaddr = findentry_name (&rtype, name);
  xdata[vaddr + 1] = count;

  /* add array header */
  xdata[--firstconst] = count;
}


void compile_fptrdecl (void)
{
  uchar name[IDENTLEN + 1];
  type_t proctype;
  int addr, procaddr;
  datatype_t ctype;
  type_t FAR* paratype;
  int FAR* paraext = NULL;
  int i = 0;
  datatype_t etype = {DDATA, 0};
  datatype_t type = {DTYPEDEF, DFPTR};

  expect (IDENT);
  tokcpy_ident (name);
  if (searchentry_name (&type, name) >= 0)
    serror (12);
  addr = newentry (type, name);

  expect ('=');
  expect ('(');

  expect (IDENT);

  switch (look_up (commandtable))
  {
  case PROC:
    proctype = DPROCEDURE;
    break;
  case FUNC:
    proctype = DT (datatype.type);
    if (lex_ref)
      proctype |= KPOINTER;
    break;
  default:
    serror (0);
  }
  xdata[addr] = procaddr = new_procfunc ((uchar*) "", nullpc /* dummy */, type.ext);

  paratype = (type_t FAR*) (custom_data (procaddr));
  paratype[0] = DUNKNOWN;

  if (LEX_CHAR != ')')
    --i;
  while (LEX_CHAR != ')')
  {
    if (lex () == VAR)
    {
      lex ();
      ctype.type = KREFERENCE;
    }
    else
      ctype.type = DUNKNOWN;
    if (tok = IDENT) /* dummy identifer */
      ctype.type |= datatype.type;
    ctype.ext = datatype.ext;
    if (LEX_CHAR == '[')
    {
      LEX_INCCHAR ();
      ctype.type |= KARRAY;
      expect (']');
    }

    ++i;
    paratype[i] = ctype.type;
    if (ctype.ext)
    {
      if (!paraext)
        paraext = &xdata[xdata[procaddr + 3] = (newentry_custom (etype, (uchar*) "", MAXPARAM) + 5)];
      paraext[i] = ctype.ext;
    }

    if (LEX_CHAR == ',')
      lex ();
  }

  paratype[i] |= KLAST;
  paratype[i + 1] = DUNKNOWN;
  expect (')');
}


void compile_const (void)
{
  uchar name[IDENTLEN];
  int addr;
  datatype_t dtype;
  byte old_dr = declaration_required;
  bool following = TRUE;

  declaration_required = 0;


  while ((lex () == IDENT) && following)
  {
    tokcpy_ident (name);
    dtype.type = DT (datatype.type);
    dtype.ext = datatype.ext;
    if (lex_ref == 1)
      dtype.type |= KREFERENCE;

    expect ('=');

    if (LEX_CHAR == '(') /* const array */
    {
      lex ();
      read_list (name, &dtype);
    }
    else
    {
      if (dtype.type != DINTEGER)
        serror (27);
      dtype.type = DCONSTANT;
      addr = findentry_name (&dtype, name);
      xdata[addr] = expect_simple_constexp (DINTEGER);
      xdata[addr + 1] = DINTEGER;
      xdata[addr + 2] = 0;
    }
    if (following = (LEX_CHAR == ','))
      lex ();
  }
  if (following)
    serror (1000 + IDENT);

  declaration_required = old_dr;
}
/* -----------------*/

void compile_for (void)
{
  cstack_entry i;
  datatype_t dtype;
  PC cpc = pc;

  i.type = CFOR;
  i.step = 1;

  lex (); /* get FOR */
  if (!comp_lvalue (&dtype))
    serror (66);
  pc = cpc;

  if ((typesize (dtype.type) != DINTEGER) || dtype.ext)
    serror (22);

  i.var = varaddr;

  expect ('=');

  comp_simple_exp (dtype.type); /* initial value */
  ADDCODE (B_ASSIGN); ADDVAL (i.var); /* assign to loop variable */
  i.start = pc; /* entry point for closing loop */

  expect (TO);

  comp_simple_exp (dtype.type); /* loop end value */

  if (lex () != STEP)
    putback ();
  else /* optional STEP : only int constants allowed */
    if ((i.step = expect_simple_constexp (DINTEGER)) == 0)
      serror (38);

  if (i.step < 0)
  {
    ADDCODE (B_FORCHECKN);
  }
  else
  {
    ADDCODE (B_FORCHECK);
  }
  ADDVAL (i.var);
  i.ende = pc; ADDCODEP (nullpc);

  cpush (i);
}

void compile_next (void)
{
  cstack_entry i;

  i = cpopandtest (CFOR);
  cstack_nextlabel (&i, C_CONTINUE);

  if (i.step == 1)
  {
    /* short standard version */
    ADDCODE (B_NEXT);
    ADDVAL (i.var);
  }
  else
  {
    /* variable step width */
    ADDCODE (B_NEXT2);
    ADDVAL (i.var);
    ADDVAL (i.step);
  }
  ADDCODEP (i.start);
  *i.ende.p = pc.c;
  cstack_nextlabel (&i, C_BREAK);
}
/*---------------------------------------------------*/
void compile_if (void)
{
  /*int neg;
  int jmpcode;*/
  int cmd;
  cstack_entry cs;

  /*if (!(neg = (lex () == NOT)))
    putback ();*/ /* conflicting with expressions like "NOT a AND b" */

  comp_simple_exp (DBOOLEAN);

  lex ();
  if (tok == THEN)
  {
    cs.type = CIF;
    cs.ende = pc;
    ADDJMP (/*neg ? JMPIF :*/ JMPIFNOT, 0); /* destination unknown */
    cpush (cs);
  }
  else
  {
    /*jmpcode = neg ? JMPIFNOT : JMPIF;*/
    if (token_type == TIDENT)
    {
      cmd = look_up (commandtable);
      if (cmd == GOTO)
        compile_goto (/*jmpcode*/ JMPIF);
      else if ((cmd == BREAK) || (cmd == CONTINUE))
        compile_break (/*jmpcode*/ JMPIF, cmd - BREAK);
      else
        serror (0);
    }
    else
      serror (0);
  }
}

void compile_else (void)
{
  PC ad1;
  cstack_entry cs;

  cs = cpopandtest (CIF);
  ad1 = cs.ende;
  cs.ende = pc;
  cpush (cs);
  ADDJMP (JMP, 0);/* dummy, address unknown */
  LINKJMP (ad1.c, pc.c);
}

void compile_endif (void)
{
  PC ad;
  cstack_entry cs;
  cs = cpopandtest (CIF);
  ad = cs.ende;
  LINKJMP (ad.i, pc.c);
}

/* while .. wend */
void compile_while (void)
{
  cstack_entry cs;
  cs.start = pc;
  comp_simple_exp (DBOOLEAN);

  cs.type = CWHILE;
  cs.ende = pc;
  ADDJMP (JMPIFNOT, 0); /* destination unknown */
  cpush (cs);
}

void compile_endwhile (void)
{
  PC ad;
  cstack_entry cs;
  cs = cpopandtest (CWHILE);
  cstack_nextlabel (&cs, C_CONTINUE);
  ADDJMP (JMP,cs.start.c);
  ad = cs.ende;
  LINKJMP (ad.i, pc.c);
  cstack_nextlabel (&cs, C_BREAK);
}
/* DO .. LOOP - loop */
void compile_do (void)
{
  cstack_entry cs;
  cs.start = pc;
  lex ();
  if (tok == WHILE)
  {
    /* condition on entry WHILE */
    comp_simple_exp (DBOOLEAN);
    cs.type = CDOC;
    cs.ende = pc;
    ADDJMP (JMPIFNOT, 0); /* destination unknown */
  }
  else
    if (tok == UNTIL)
    {
      /* condition on entry UNTIL */
      comp_simple_exp (DBOOLEAN);
      cs.type = CDOC;
      cs.ende = pc;
      ADDJMP (JMPIF, 0); /* destination unknown */
    }
    else
    {
      /* condition on end of loop */
      putback ();
      cs.type = CDO;
      cs.start = pc;
    }
  cpush (cs);
}

void compile_loop (void)
{
  PC ad;
  cstack_entry cs;
  cs = cpop ();
  cstack_nextlabel (&cs, C_CONTINUE);
  if (cs.type == CDO)
    {
      /* condition must follow here */
      lex ();
      if (tok == WHILE)
      {
        comp_simple_exp (DBOOLEAN);
        ADDJMP (JMPIF, cs.start.c);
      }
      else
        if (tok == UNTIL)
        {
          comp_simple_exp (DBOOLEAN);
          ADDJMP (JMPIFNOT, cs.start.c);
        }
        else
          serror (0);
    }
  else
    if (cs.type == CDOC)
    {
      ADDJMP (JMP, cs.start.c);
      ad = cs.ende;
      LINKJMP (ad.i, pc.c);
    }
    else
      serror (11);
  cstack_nextlabel (&cs, C_BREAK);
}


/* case keyword */

const unsigned int FAR _case_ofs[] = {
  /*256 = LESS*/        B_OF_B,
                        B_OF_B /* should never be called */,
  /*258 = GREATER*/     B_OF_A,
  /*259 = NOT_LESS*/    B_OF_AE,
                        B_OF_B /* should never be called */,
  /*261 = NOT_GREATER*/ B_OF_BE,
};

void compile_case (void)
{
  datatype_t seltype;
  bool Else;
  int tok2;
  cstack_entry cs;

  lex ();
  if (tok == SELECT)
  { /* CASE SELECT */
    comp_exput (&seltype);
    cs.var = seltype.type;
    cs.brkcnt = seltype.ext;
    if ((seltype.type == DSTRING) || (seltype.type == DFLOAT))
      serror (22);
    ADDCODE (B_CASE);
    cs.type = CCASE;
    cs.start = pc;
    cs.step = 0;
    cpush (cs);
  }
  else
  {
    if (!(Else = (tok == ELSE)))
      putback ();

    cs = cpopandtest (CCASE);
    if (cs.step == 0) /* first CASE statement */
      if (cs.start.c != pc.c)
        serror (0); /* statements outside CASE */

    if (cs.step > 1)
      LINKJMP (cs.ende.i, pc.c); /* "close" previous case */
    if (cs.step > 0)
    {
      cs.ende = pc;
      ADDJMP (JMP, 0);
      *cs.start.p = pc.c;
    }
    if (!Else)
    { /* range checks in CASE (MB,4.10) */
      seltype.type = cs.var;
      seltype.ext = cs.brkcnt;
      lex ();
      switch (tok)
      {
      case GREATER:
      case LESS:
      case NOT_GREATER:
      case NOT_LESS:
        tok2 = tok;
        comp_exp (&seltype);
        ADDCODE (_case_ofs[tok2 - LESS]);
        break;
      case '=':
        comp_exp (&seltype);
        ADDCODE (B_OF_E);
        break;
      default:
        putback ();
        comp_exp (&seltype);
        lex ();
        if (tok == TO)
        {
          comp_exp (&seltype);
          ADDCODE (B_OF_BOUND);
        }
        else
        {
          putback ();
          ADDCODE (B_OF_E);
        }
      }
      cs.start = pc;
      ADDCODEP (nullpc);
    }
    else
      cs.start.i = NULL;
    ++cs.step;
    cpush (cs);
  }
}

void compile_endcase (void)
{
  cstack_entry cs;
  cs = cpopandtest (CCASE);
  if (cs.step == 0)
    serror (0);
  if (cs.step > 0)
    if (cs.start.i != NULL)
      *cs.start.p = pc.c;
  if (cs.step > 1)
    LINKJMP (cs.ende.i,pc.c);
}

/*----------------------------------------------*/
void compile_version (void)
{
  int version = expect_simple_constexp (DINTEGER);

  if (version > OWBVersion)
    serror (51);

  desired_version = version;
  if (version < 325)
    default_hash_mode = 2;
  if (version < 110)
  {
    intchar = '#';
    floatchar = '%';
  }
}
/*-------------------------------------------------*/
void compile_input (void)
{
  datatype_t vtyp;

  if (lex () == STRING)
  {
    /* a prompt is given */
    putback ();
    comp_simple_exp (DSTRING);
    ADDCODE (B_PRINT_STRING); /* print prompt */
    lex (); /* expect delimiter */
    if (tok != ',' && tok != ';')
      serror (14);
    lex (); /* get first identifier for input */
  }
  if (tok != IDENT)
    serror (4);

  comp_lvalue (&vtyp); /* compile variable */
  switch (vtyp.type)
  {
  case DCHAR:
  case DBOOLEAN:
  case DINTEGER:
    ADDCODE (B_INPUT);
    ADDCODE (B_ASSIGN_I);
    break;
  case DFLOAT:
    ADDCODE (B_INPUT_F);
    ADDCODE (B_ASSIGN_F);
    break;
  case DSTRING:
    ADDCODE (B_INPUT_S);
    ADDCODE (B_ASSIGN_S);
    break;
  default:
    serror (22);
  }
}
/*-------------------------------------------------*/
int compile_procedure (void)
{
  /* compile a procedure call */
  int found;
  uchar htoken[IDENTLEN];
  uchar version;

  KeywordPr FAR* hptr = proctable;
  /* all normal procedures in proctable */

  tokcpy_ident (htoken);
  /* find entry */
  found = FALSE;
  while ((!found) && (hptr->command != 0))
  {
    found = (strcmp_nf (htoken, hptr->command) == 0);
    if (!found)
      ++hptr;
  }
  if (!found)
    return (FALSE);

  /* compile parameters */
  comp_para (hptr->paratype, NULL, &version, DELIMITER, 0);
  ADDCODE (hptr->bincode[version]);
  return (TRUE);
}

bool cm_init = TRUE;
static void NEAR CompileMessage (uchar* memoname)
{
#ifndef DEBUGCOMP
  if (switches.quiet) /* do not show compiling (MB,4.20) */
    return;

  if (cm_init)
  {
    cm_init = FALSE;
    dbox (20, 64, 139, 94, NULL);
    Line (20, 74, 139, 74, 1);
    LibPutProStr (IB_PFONT2, 22, 66, (uchar*) Near (msg_mm_compiling), 117);
    LibPutProStr (IB_PFONT2, 22, 76, memoname, 117);
  }

  Box (21, 86, 137, 93, 10);
  LibPutProStr (IB_PFONT1, 22, 86, memoname ? memoname : msg_done, 117);
  LibPutDisp ();
#endif

#if FALSE
#ifndef DEBUGCOMP
  uchar temp[200];
  if (switches.quiet) /* do not show compiling (MB,4.20) */
    return;

  if (memoname == NULL)
    Box (5, 20, 155, 40, 10);
  else
  {
    sprintf ((char*) temp, Near (compiling), memoname);
    textbox (temp, 5, 20, 155, 40, 0); /* always font 0 (MB,4.10) */
  }
  LibPutDisp ();
#endif
#endif
}

bool loadprogmemo (void)
{
  uchar *sig;

  switch (ITOS.progtype)
  {
  case MMAIN:    sig = PROGRAMSIG; break;
  case MINCLUDE: sig = INCLUDESIG; break;
  case MAPPEND:  sig = APPENDSIG;  break;
  }

  /*LibPutMsgDlg ("sub:%d,fp:%d,sig:%s\rname:%s", ITOS.min, ITOS.fp, sig, ITOS.fname);*/

  if (ITOS.min != 0)
    return (load_memo_named (ITOS.min, sig, ITOS.fname, &ITOS.fp));
  else
    switch (ITOS.progtype)
    {
    case MMAIN:
      ITOS.min = switches.progsub;
      return (load_memo_named (switches.progsub, sig, ITOS.fname, &ITOS.fp));
    case MINCLUDE:
      ITOS.min = IncStack[0].min;
      if (load_memo_named (IncStack[0].min, sig, ITOS.fname, &ITOS.fp))
        return (TRUE);
      ITOS.min = switches.includesub;
      if (IncStack[0].min != switches.includesub)
        return (load_memo_named (switches.includesub, sig, ITOS.fname, &ITOS.fp));
      else
        return (FALSE);
    case MAPPEND:
      ITOS.min = IncStack[0].min;
      return (load_memo_named (IncStack[0].min, sig, ITOS.fname, &ITOS.fp));
    }
}

bool loadprog (void)
{
  ITOS.fp = 0;

#ifndef SMALL_NO_CACHE
  if (switches.caching && !switches.savetocache)
  { /* search for cached program (MB,4.22) */
    if (cache_read ())
    {
      switches.cached = TRUE;
      switches.statistics = FALSE;
      return (TRUE);
    }
  }
#endif

#ifdef REMOTE_CONTROL
  if (switches.remote)
  {
    ITOS.min = 0;
    if (rc_loadprog ())
      return (TRUE);
  }
#endif
  return (loadprogmemo ());
}

void set_global_variables (void)
{
  GlobalVar FAR* i;
  int j;
  datatype_t dtype;

  listreset ();

  dtype.ext = 0;
  for (i = vartable; i->name; ++i)
  {
    dtype.type = i->type & TYPEMASK;
    j = findentry_name (&dtype, (uchar*) (Near (i->name)));
    dtype.type = i->type;
    if (dtype.type & KCONST)
    {
      if (dtype.type & KREFERENCE)
        xdata[j] = *(i->u.index);
      else
        xdata[j] = i->u.val;
    }
    else
    {
      if (dtype.type & KPOINTER)
        *(i->u.ptr) = &xdata[xdata[j]];
      else
        *(i->u.index) = xdata[j];
      expandentry (lastaddr, i->size, 0);
    }
  }
}

void init_global_variables (void)
{
  xdata[owbversioni] = OWBVersion;
#ifdef __HITACHI__
  xdata[pvmodeli] = 1600;
#endif
#ifdef LSI_C
  xdata[pvmodeli] = 450;
#endif

  putvar (owbparameter, stringstack1);
}

int compile (int lastcode)
{
  /* compile process */
  int rc, ec = FALSE;

  ec = FALSE;
  if (setjmp (errorret))
    return (TRUE);
  prog_start = prog = aprog = FB;
  ITOS.line = 1;
  ctos = 0;            /* ForTo stackpointer reset */
  cpp = NULL;            /* no procedure declaration is running */
  last_proc = 0;
  a_overlay = FALSE;        /* not compiling overlay */

  default_datatype = DINTEGER; /* Default datatype for variables without suffix */
  pc.c = bin;                  /* codepointer to start of binary area */
  bin_start.c = bin;           /* start point for interpreter */
  lastpc.c = &bin[lastcode];   /* end of code */
  f_stack_ptr.f = f_stack;     /* for compile time evaluation */
  i_stack_ptr = i_stack - 1;   /* for compile time evaluation */

  IncStackPtr = IncStack;


  set_global_variables ();

  /*
    if (switches.printwhilecomp)
    {
    for (i=0;i<20;i++) OutTrm('\n');
    DisplayTrm();
    }
  */

  CompileMessage (ITOS.fname);
#ifdef DEBUGCOMP
 #ifdef __HITACHI__
  OutputTrm ((uchar*) "\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r");
 #endif
#endif
  rc = RC_CONTINUE;
  while (rc != RC_EXIT)
  {
    switch (rc = compile_command ())
    {
    case RC_INCLUDE:
      /* handle include */
      ITOS.prog = prog; /* store pointer of compiler */
      ++IncStackPtr;
      if (IncStackPtr - IncStack < INCSTACKSIZE) /* include stack not full? */
      {
        /* load new memo */
        ITOS.line = 1;
        ITOS.prog = prog = prog_start = aprog = FB;
        if (token_type == TIDENT)
          tokncpy (ITOS.fname, NAMELEN);
        else sconstncpy (ITOS.fname, NAMELEN);
        ITOS.progtype = MINCLUDE;
        ITOS.min = 0;
        CompileMessage (ITOS.fname);
        if (!loadprog ())
          serror1 (2043, ITOS.fname); /* not found */
      }
      else
        serror (42); /* include stack overflow */
      break;

    case RC_APPEND:
      /* load new memo from program category */
      ITOS.line = 1;
      /*  ITOS.min=0; */
      ITOS.progtype = MAPPEND;
      prog_start = aprog = prog = FB;
      if (token_type == TIDENT)
        tokncpy (ITOS.fname, NAMELEN);
      else
        sconstncpy (ITOS.fname, NAMELEN);
      CompileMessage (ITOS.fname);
      if (!loadprog ())
        serror1 (2044, ITOS.fname); /* not found */
      break;

    case RC_EXIT:
      /* end of memo */
      if (IncStackPtr > IncStack) /* entry in include stack ? */
      {
        --IncStackPtr; /* restore previous memo */
        aprog = prog = ITOS.prog;
        CompileMessage (ITOS.fname);
        if (ITOS.fp == 0)
          loadprog ();
        else
          fileread (3, ITOS.min, ITOS.fp);
        rc = RC_CONTINUE; /* continue compile */
      }
      break;
    }
  }
  /*  if (switches.printwhilecomp) OutTrm('\n');*/

  ADDCODE (B_END0); /* make sure the program doesn't run into undefined code */
  CompileMessage (NULL);
  check_labels ();  /* check if all used labels are defined */
  if (cpp)
    serror (63); /* open procedure definition */
  if (d_cnt != 1)
    serror (21); /* unclosed #IF */
  if (ctos != 0)
    notclosederror (cstack[ctos - 1].type);  /* unclosed control */
  create_variables ();        /* create variables for runtime
                                   (deleting symbols) */
  return (ec);
}
