#include <string.h>

#include "comp.h"
#include "owbasic.h"
#include "shared.h"
#include "graphic.h"
#include "ostring.h"
#include "file.h"

#ifdef tolower /* __HITACHI__ */
#undef tolower
#endif

static const byte FAR kind[16]={
  FILE_KIND_TEXT,FILE_KIND_TEXT,FILE_KIND_TEXT,FILE_KIND_TEXT,
  FILE_KIND_TEXT,FILE_KIND_BIN, FILE_KIND_BIN, FILE_KIND_TEXT,
  FILE_KIND_TEXT,FILE_KIND_TEXT,FILE_KIND_BIN, FILE_KIND_BIN,
  FILE_KIND_TEXT,FILE_KIND_TEXT,FILE_KIND_TEXT,FILE_KIND_TEXT};

FILE_INFO finf;

/* transform OWB submode -> PVOS submode + secretinfo */
#ifdef FAST_MACRO
#define Sub(sub)  (sub & 0x0F)
#define Info(sub) (((sub) > 15) ? FILE_SECRET_MODE : FILE_OPEN_MODE)
#else
int Sub(int sub) { return sub & 0x0F; }

int Info(int sub)
{
  if (sub>15) return FILE_SECRET_MODE;
  return FILE_OPEN_MODE;
}
#endif

/* Fill file datastructures */

static void NEAR SetFPLocal (int mode, int sub, word fp)
{
  fb.fsb_main_entry_ = mode << 4;	/* Main */
  fb.fsb_sub_entry_  = Sub (sub);	/* Sub */
  fb.fsb_scrt_info_  = Info (sub);	/* Open Mode */

  fb.fsb_todo_chek_ = 0xFF;
  fb.fsb_todo_rank_ = 0xFF;

  finf.fp      = fp;
  finf.kind    = kind[mode];
}

void SetFP (int mode, int sub, word fp)
{
  SetFPLocal (mode, sub, fp);
}

/*---------basic file operations -----------------*/
bool fileread(int mode,int sub,word fp)
{
  SetFPLocal (mode, sub, fp);
  return LibFileReadEx (&fb, &finf, FBSIZE / FILE_BLOCK_SIZE);
}

word filewrite(int mode,int sub, word fp)
{
  SetFPLocal(mode,sub,fp);
  if (LibFileWrite(&fb,&finf)!=TRUE) return 0xffff;
  return finf.fp;
}

bool fileremove(int mode,int sub,word fp)
{
  SetFPLocal(mode,sub,fp);
  return LibFileRemove(&fb,&finf);
}

bool filemove(int mode,int sub,word fp,word fpz)
{
  SetFPLocal(mode,sub,fp);
  return LibFileExch(&fb,&finf,fpz);
}

word filenext(int mode,int sub,word fp)
{
  SetFPLocal(mode,sub,fp);
  if (!LibFileFindNext(&fb,&finf,0x00)) return 0xffff;
  return finf.fp;
}

word fileprev(int mode,int sub,word fp)
{
  SetFPLocal(mode,sub,fp);
  if (!LibFileFindPrev(&fb,&finf,0x00)) return 0xffff;
  return finf.fp;
}

/* transform number to fp */
word get_file_fp(int mode,int sub, int nr)
{
  int found;
  if (nr==-1) return 0xffff;

  SetFPLocal(mode,sub,0xffff);

  while ((found=LibFileFindNext(&fb,&finf,0x00))&&(nr>0)) nr--;
  if (!found) return 0xffff;            /* failed */
  return finf.fp;
}

bool load_file(int mode,int sub,int nr)
{
  word fp;

  fp=get_file_fp(mode,sub,nr);
  if (fp==0xffff) return FALSE;

  return fileread(mode,sub,fp);
}

bool save_file(int mode,int sub, int nr)
{
  SetFPLocal(mode,sub,get_file_fp(mode,sub,nr));
  return LibFileWrite(&fb,&finf);
}

/* ------------------- MEMO  ----------------- */
static int NEAR fb2array(int addr,int size,int *total)
{
  /* transfer filebuffer to basic string array */
  int l;
  int rc;

  l=0;
  *total=0;
  fbinit();
  spush1();
  fbstring(STOS,&rc,1); /* read from buffer */
  while ((l<size)&&(rc==0))
    {
      (*total)++;        /* count line */
      spop(addr); addr++; /* store string to array element */
      spush1();
      fbstring(STOS,&rc,1); /* next string */
      l++;
    }
  spop1();

  while(l<size)
    {
      /* clear remaing array elements */
      spush1(); STOS[0]=0; spop(addr);
      addr++; l++;
    }

  spush1();
  while (rc==0)
    {
      (*total)++;
      fbstring(STOS,&rc,1);
    }
  spop1();
  return 0;
}

static int NEAR array2fb(int addr,int size,int anz)
{
  int i, rc;
  if (anz==-1) anz=size; /* store all elements */
  else
    {
      /* store anz elements */
      if (anz>size) return 38; /* valid number ?*/
      if (anz<1) return 38; /* valid number ?*/
    }
  fbinit();
  i=0;
  while (i<anz)
    {
      spush(addr); addr++;
      rc=fbstore(STOS,1);
      spop1();
      if (rc!=0) return rc;
      i++;
    }
  return 0;
}

int loadmemofp(int addr,int anz, int sub, word fp,int *total)
{
  *total=0;
  if (!fileread(3,sub,fp)) return 39;
  return fb2array(addr, anz, total);
}

int savememofp(int addr,int size, int sub, word *fp, int anz)
{
  int rc;
  rc=array2fb(addr,size,anz);
  if (rc!=0) return rc;
  *fp=filewrite(3,sub,*fp);
  if (*fp==0xffff) return 39;
  return 0;
}

int loadmemo(int addr,int size,int sub,int nr,int *total)
{
  word fp;

  fp=get_file_fp(3,sub,nr);
  if (fp==0xffff) return 61;
  return loadmemofp(addr,size,sub,fp,total);
}

int savememo(int addr,int size,int sub,int nr,int anz)
{
  word fp;
  if (nr==-1) { fp=0xffff; }
  else
    {
      fp=get_file_fp(3,sub,nr);
      if (fp==0xffff) return 61;
    }
  return savememofp(addr,size,sub,&fp,anz);
}

int filefind (int mode, int submode, const uchar* name, const uchar* sig, word* fp)
{
  uchar sname[MAXSTRLEN]; /* name to search for */

  /*uchar fname[MAXSTRLEN];*/

  bool found = FALSE;
  uchar* ip = sname;
  byte* bp;
  /*int i;*/
  register uchar bc, ic;
  bool is_bin = (kind[mode] == FILE_KIND_BIN);


  while (*(ip++) = tolower (*(name++)))
    ;
  *ip = 0x00;

  finf.fp = 0xFFFF; /* set file pointer for first search */
  do
  {
    SetFPLocal (mode, submode, finf.fp);
    LibFileFindNext (&fb, &finf, 0x00);

    if (LibFileReadEx (&fb, &finf , (MAXSTRLEN + FILE_BLOCK_SIZE - 1) / FILE_BLOCK_SIZE))
    {
      ip = sname - 1;
      bp = is_bin ? FBB : FB;
      found = TRUE;

      if (sig)
        if (*sig)
        {
          if ((*bp == *sig) && (bp[1] == sig[1]))
          {
            bp += 2;
            while (*bp == ' ')
              ++bp; /* skip whitespaces after signature */
          }
          else
            found = FALSE;
        }

      /*i = 0;
      while ((bp[i] != 0x00) && (bp[i] != 0x0A) && (bp[i] != 0x0D))
        { fname[i] = bp[i]; ++i; }
      fname[i] = 0;

      LibPutMsgDlg ("req:%s\rnme:%s\rfp:%d,sig:%c%c", sname, fname, finf.fp, *(bp - 2), *(bp - 1));*/

      --bp;

      if (found)
      {
        found = FALSE;
        do
        {
          bc = *(++bp);
          ic = *(++ip);
          if (((bc == 0x0A) || (bc == 0x0D) || (bc == 0x00)) && (ic == 0x00))
          {
            *fp = finf.fp;
            return (TRUE);
          }
          if (ic != tolower (bc))
            break;
        }
        while (ip < (sname + MAXSTRLEN - 1));
      }
    }
  } while (!found && (finf.fp != 0xFFFF));

  return (found);
}

bool load_memo_named (int sub, const uchar* sign, const uchar* name, word* fp)
{
  if (filefind (3, sub, name, sign, fp))
    return (LibFileReadEx (&fb, &finf, FBSIZE / FILE_BLOCK_SIZE));
}

int loadmemonamed (int addr, int size,
      int sub, const uchar* name, const uchar* sig,
      int* total, word* fp)
{
  if (!load_memo_named (sub, sig, name, fp))
    return (61);
  return fb2array (addr, size, total);
}

/*-------- Quickmemo -------------------------------------*/

int loadqmfp(int addr,int size,int sub,word fp)
{
  int i;
#ifdef __HITACHI__
  short val;
  int val1, val2;
#else
  int val;
#endif
  if (size<1301) return 30;
  xdata[addr]=0; /* icon invalid */
  if (fp==0xffff) return 61;
  if (!fileread(11,sub,fp)) return 39;
  xdata[addr]=156; xdata[addr+1]=130;
  fbbinit();
  for (i=0; i<1300; i++)
    {
#ifdef LSI_C
      fbbint(&val);
#endif
#ifdef __HITACHI__
      fbbbyte(&val1);
      fbbbyte(&val2);
      val=val1 + (val2 << 8);
#endif
      xdata[addr+2+i]=val;
    }
  return 0;
}

int loadqmhash (int addr,int size,int sub,int crc1,int crc2,word *fp,int mode)
{
  bool found=FALSE;
  int crc1a, crc2a;

  SetFPLocal(11,sub,0xffff);
  *fp=0xffff;
  while (!found)
    {
      if (!LibFileFindNext(&fb,&finf,0x00)) return 61; /* not found */
      if (!LibFileRead(&fb,&finf))
  return 39; /* should not happen */
      crc(FBB,2600,&crc1a,&crc2a,mode);
      found=(crc1==crc1a)&&(crc2=crc2a);
    }
  *fp=finf.fp;
  return loadqmfp(addr,size,sub,*fp);
}

int loadqm(int addr,int size,int sub,int nr)
{
  word fp;
  fp=get_file_fp(11,sub,nr);
  return loadqmfp(addr,size,sub,fp);
}

void icon2fb(byte FAR *iconbuf)
{
  int xi=*(word FAR*)iconbuf;
  int xbi=(xi+7)/8; /* number of bytes per line */
  int yi=*(word FAR*)(iconbuf+2);
  int xb,yb;

  byte FAR * bptr=iconbuf+4;

  if (yi>130) yi=130; /* no more then 130 line (quickmemo limit) */

  fbbinit();
  for (yb=0;yb<yi;yb++)
    {
      for (xb=0;xb<xbi;xb++)
  {
    fbbputbyte(*(bptr++));
  }
      for (;xb<20;xb++)
  {
    fbbputbyte(0);
  }
    }

  for (yb=yi;yb<130;yb++)
    {
      for (xb=0;xb<20;xb++)
  {
    fbbputbyte(0);
  }
    }
}

int saveqmfp (int addr, int size, int sub, word FAR* fp)
{
  char datestring[11];
  word year;
  byte month, day, hour, min, sec;
  int i;

  if (!checkicon(xdata[addr], xdata[addr + 1], size))
    return (59);

#ifdef LSI_C
  icon2fb ((byte FAR *) &(xdata[addr]));
#endif
#ifdef __HITACHI__
  toiconbuffer ((byte*) (xdata + addr));
  icon2fb (iconbuffer);
#endif

  LibGetDateTime2 (&year, &month, &day, &hour, &min, &sec);
  sprintf (datestring, "%4d/%2d/%2d", year, month, day);

  for (i = 0; i < 10; ++i)
    fbbputbyte (0);

  /*  ! write date  */
  for (i = 0;i < 10; ++i)
    fbbputbyte (datestring[i]);

  if ((*fp = filewrite (11, sub, *fp)) == 0xFFFF)
    return (39);
  return (0);
}

int saveqm(int addr,int size,int sub,int nr)
{
  word fp;
  fp=get_file_fp(11,sub,nr);
  return saveqmfp(addr,size,sub,&fp);
}

int iconhash (int addr, int size, int FAR* crc1, int FAR* crc2, int mode)
{
  if (!checkicon (xdata[addr], xdata[addr + 1], size))
    return (59);

#ifdef LSI_C
  icon2fb ((byte FAR *) &(xdata[addr]));
#endif
#ifdef __HITACHI__
  toiconbuffer ((byte*) (xdata + addr));
  icon2fb (iconbuffer);
#endif
  crc (FBB, 2600, crc1, crc2, mode);
  return (0);
}

char FAR prefsfilename[] = "OWBasic,prefs";
/* load / save settings */
void loadsettings (void)
{

  sfLoadRecord ((uchar*) (Near (prefsfilename)), sizeof (Pref), (byte*) &pref);

  if (pref.ver != SETTINGSVER)
  {
    msgbox ((uchar*) (Near (msg_wrongsettings)), NULL, 0.4, MBX_WAIT);
    default_prefs ();
    savesettings ();
  }

  stringstack1[0] = 0;
}

void savesettings (void)
{
  sfSaveRecord ((uchar*) Near (prefsfilename), sizeof (Pref), (byte*) &pref);
}

int namedfile (const uchar *name, int* imode, int* isubmode)
{
  byte submode, mode;
  /* find or create named file */
  if (!LibSubEntrySave ((byte*) name, &submode))
    return (39);

  /* get mode/submode */
  LibGetAllEntry ((byte*) name, &mode, &submode);
  *imode = mode >> 4; /* convert to OWB style */
  *isubmode = submode;
  return (0);
}

#define SIGLEN 80
/* Shared file access */
static void NEAR mksig(uchar *sig,const uchar *pname,const uchar *sub,int n)
{
  const uchar *p1;
  uchar *p2;
  p2=sig;
  p1=pname;
  while ((*p1!=0)&&(p2-sig<n)) *(p2++)=*(p1++);
  if (strlen((char*)sub)>0)
    {
      if (p2-sig<n) *(p2++)=',';
      p1=sub;
      while ((*p1!=0)&&(p2-sig<n)) *(p2++)=*(p1++);
    }
  if (p2-sig<n) *(p2++)=0;
}

bool loadshared(const uchar *sig,const uchar *pname,uchar FAR *buf,int n)
{
  /* load record from shared file */
  uchar fsig[SIGLEN];
  mksig(fsig,pname,sig,SIGLEN);
  return sfLoadRecord(fsig,n,buf);
}

bool saveshared(const uchar *sig,const uchar *pname,uchar FAR *buf,int n)
{
  /* save record to shared file */
  uchar fsig[SIGLEN];
  mksig(fsig,pname,sig,SIGLEN);
  return sfSaveRecord(fsig,n,buf);
}

bool removeshared(const uchar *sig,const uchar *pname)
{
  /* remove record from shared file */
  uchar fsig[SIGLEN];
  mksig(fsig,pname,sig,SIGLEN);
  return sfRemoveRecord(fsig);
}

bool findshared (const uchar* sig, const uchar* pname,
    uchar *rname, int namelen, word *fp)
{
  uchar fsig[SIGLEN];

  if (*pname) /* not an empty string */
  {
    mksig (fsig, pname, sig, SIGLEN);
    sig = fsig;
  }

  return (sfFindRecord (fsig, rname, namelen, fp));
}
/* -------------------------------------------- */

uchar recordname[] = PROGNAME ",NOverlay%02.2i,%s";

bool Overlay (int action, int nr, codet* addr, int n)
{
  uchar buf[20 + NAMELEN];
  int (*wfunc) (const uchar*, int, byte FAR*);
  bool cachemode = switches.cached || switches.savetocache;

  sprintf ((char*) buf, (char*) recordname, nr, cachemode ? (uchar*) "" : (uchar*) progname);
  if (cachemode)
    buf[8] = 'C';

  /*if (action == OV_DELETE)
    sfRemoveRecord (buf);
  else
  {*/
    wfunc = (action == OV_LOAD) ? sfLoadRecord : sfSaveRecord;
    wfunc (buf, n, (byte FAR*) addr);
  /*}*/
}
