#include "owbasic.h"

/* "pointer" for sequential read/write to buffer */
int fbptr;

#ifdef __HITACHI__
 #define FBBSIZE (32 * 1024)
#else
 #define FBBSIZE (3 * 1024)
#endif


const fbdatetime_t FAR fbdatep[] = {
  fb.fsb_date1_buf_,
  fb.fsb_date1_buf_,
  fb.fsb_date2_buf_,
  fb.fsb_date_aram_,
  fb.fsb_date_srch_
};

const fbdatetime_t FAR fbtimep[] = {
  fb.fsb_time1_buf_,
  fb.fsb_time2_buf_,
  fb.fsb_time3_buf_,
  fb.fsb_time_aram_,
  fb.fsb_time_srch_
};


#ifndef SMALL_FBDATETIME
int fbdate(int sel,uchar *date)
{
  int i;
  uchar *sp;
  uchar *dp;
  dp=date;
  sp=fbdp(sel);
  if (sp==NULL) return 38;
  for (i=0;i<8;i++) *(dp++)=*(sp++);
  *dp=0;
  return 0;
}

int fbtime(int sel,uchar *time)
{
  int i;
  uchar *sp;
  uchar *dp;
  dp=time;
  sp=fbtp(sel);
  if (sp==NULL) return 38;
  for (i=0;i<4;i++) *(dp++)=*(sp++);
  *dp=0;
  return 0;
}

int fbputdate(int sel,uchar *date)
{
  int i;
  uchar *sp;
  uchar *dp;
  sp=date;
  dp=fbdp(sel);
  if (dp==NULL) return 38;
  for (i=0;i<8;i++) *(dp++)=*(sp++);
  return 0;
}

int fbputtime(int sel,uchar *time)
{
  int i;
  uchar *sp;
  uchar *dp;
  sp=time;
  dp=fbtp(sel);
  if (dp==NULL) return 38;
  for (i=0;i<4;i++) *(dp++)=*(sp++);
  return 0;
}
#else

int fbdatetimei (int sel, uchar* datetime, int max, int put, const fbdatetime_t FAR* tbl)
{
  int i;
  uchar *hp;
  uchar *sp;

  if ((sel < 0) || (sel > 4))
    return (38);
  sp = tbl[sel];

  if (put)
  {
    hp = datetime;
    datetime = sp;
    sp = hp;
  }

  for (i = 0; i < max; ++i)
    *(datetime++) = *(sp++);
  if (!put)
    *datetime = 0;
  return (0);
}
#endif

/*
   write/read of string data to TEXT file buffer and move pointer
*/

#ifndef FAST_MACRO
int fbinit() { fbptr=0; return 0; }
#endif

int fbstore(uchar *s,int del)
{
  uchar *ptr=s;
  while ((*ptr!=0)&&(fbptr<2047)) FB[fbptr++]=*(ptr++);
  if (fbptr >= 2047)
    return 33;
  switch (del)
  {
  default:
  case 0: --fbptr;/*nothing*/ ; break;
  case 1: FB[fbptr]='\r'; break;
  case 2: FB[fbptr]=0xFE; break;
  case 3: FB[fbptr]='\r'; FB[++fbptr]=0xFE; break;
  }
  FB[++fbptr]=0;
  return 0;
}

int fbstring(uchar *s, int *rc,int del)
{
  uchar *ptr;
  uchar c;
  bool fini;
  ptr=s;
  *s=0; /* start with empty string */
  if (FB[fbptr]==0)
    {
      *rc=1; return 0; /* end of file */
    }
  else
    {
      do
  {
    c=FB[fbptr];
    if ((ptr-s)>=MAXSTRLEN) return 32;

    switch (del)
    {
    case 0: fini=(c==0); break;
    case 1: fini=(c=='\r') || (c==0); break;
    case 2: fini=(c==0xfe) || (c==0); break;
    case 3: fini=(c==0xfe) || (c=='\r') || c==0; break;
    }
    if (!fini)
      {
        *(ptr++)=c; fbptr++;
        *ptr=0;
      }
  } while (!fini);
      if (FB[fbptr]!=0) fbptr++;
      *rc=0;
    }
  return 0;
}

/*
  write and read BINARY filebuffer FBB
*/

#ifndef FAST_MACRO
int fbbinit()
{
  /* Initialise read/write process */
  fbptr=0;
  return 0;
}

int fbbtell (void)
{
  return (fbptr);
}
#endif

int fbbseek(int bnumber)
{
  /* set filepointer for binary read/write */
  fbptr=bnumber;
  if ((fbptr<0)||(fbptr>=FBBSIZE))
    {
      fbptr=0;
      return 30;
    }
  return 0;
}

int fbbputint(int val)
{
  /* write int to FBB, move pointer and set FBBNUM*/
  if (fbptr+sizeof(int)>FBBSIZE) return 33;

#ifdef LSI_C
  *(int*)(FBB+fbptr)=val;
#else
  memcpy(FBB+fbptr,&val,sizeof(int));
#endif
  fbptr+=sizeof(int);
  FBBNUM=fbptr;
  return 0;
}

int fbbputbyte(int val)
{
  /* write byte to FBB, move pointer and set FBBNUM*/
  if (fbptr+1>FBBSIZE) return 33;
  *(FBB+fbptr)=val;
  fbptr++;
  FBBNUM=fbptr;
  return 0;
}

int fbbputfloat(FLOATT val)
{
  /* write FLOAT to FBB, move pointer and set FBBNUM*/
  if (fbptr+sizeof(FLOATT)>FBBSIZE) return 33;
#ifdef LSI_C
  *(FLOATT*)(FBB+fbptr)=val;
#else
  memcpy(FBB+fbptr,&val,sizeof(FLOAT));
#endif
  fbptr+=sizeof(FLOATT);
  FBBNUM=fbptr;
  return 0;
}

int fbbputstring(uchar *s)
{
  /* write string to FBB, move pointer and set FBBNUM*/
  uchar *ptr=s;
  while (*ptr!=0)
    {
      if (fbptr+1>FBBSIZE) return 33;
      else *(FBB+fbptr++)=*ptr++;
    }
  *(FBB+fbptr++)=0;
  FBBNUM=fbptr;
  return 0;
}

int fbbint(int *val)
{
  /* read int from FBB, move pointer */
  if (fbptr+sizeof(int)>FBBNUM) return 60;

#ifdef LSI_C
  *val=*(int*)(FBB+fbptr);
#else
  memcpy(val,FBB+fbptr,sizeof(int));
#endif
  fbptr+=sizeof(int);
  return 0;
}

#ifdef __HITACHI__
int fbbshort(int *val)
{
  /* read short int from FBB, move pointer */
  if (fbptr+sizeof(short)>FBBNUM) return 60;

  /*  *val=*(short*)(FBB+fbptr);*/
  memcpy(val,FBB+fbptr,sizeof(short));

  fbptr+=sizeof(short);
  return 0;
}
#endif

int fbbbyte(int *val)
{
  /* read byte from FBB, move pointer */
  if (fbptr+1>FBBNUM) return 60;

  *val=*(FBB+fbptr);
  fbptr++;
  return 0;
}

int fbbfloat(FLOATT *val)
{
  /* read FLOATT from FBB, move pointer */
  if (fbptr+sizeof(FLOATT)>FBBNUM) return 60;

#ifdef LSI_C
  *val=*(FLOATT*)(FBB+fbptr);
#else
  memcpy(val,FBB+fbptr,sizeof(FLOATT));
#endif

  fbptr+=sizeof(FLOATT);
  return 0;
}

int fbbstring(uchar *s)
{
  /*
   *   read string from FBB, move pointer
   */
  uchar *ptr=s;
  *ptr=0;
  while ((*(FBB+fbptr)!=0)&&(ptr-s<MAXSTRLEN-1))
    {
      if (fbptr+1>FBBNUM) return 60;
      else *ptr++=*(FBB+(fbptr++));
    }
  if (fbptr<FBBNUM) fbptr++; /* set pointer after 0-Byte */
  *ptr=0;
  return 0;
}




