#include "math.h"
#include <owbasic.h>
#include <graphic.h>

#define GRAPHIC_SMALL
#define GRAPHIC_SMALL2

const byte FAR drawmode[MAXDRAWMODES]={
  IB_GDS_AND,  IB_GDS_OVER, IB_GDS_AND, IB_GDS_OR,
  IB_GDS_MESH, IB_GDS_XOR,  IB_GDS_AND, IB_GDS_AND
};

const setfuncp FAR psetfunc[MAXDRAWMODES]={
  LibGdsDotOff, LibGdsDotOn,  LibGdsDotOff, LibGdsDotOn,
  LibGdsDotOn,  LibGdsDotCmp, LibGdsDotOn,  LibGdsDotOn
};

const linfuncp FAR linefunc[MAXDRAWMODES]={
  LibGdsLineClr,  LibGdsLine,    LibGdsLineClr, LibGdsLine,
  LibGdsLineMesh, LibGdsLineCmp, LibGdsLine,    LibGdsLine
};

const boxfuncp FAR boxfunc[MAXDRAWMODES]={
  LibGdsBoxClr,  LibGdsBox,    LibGdsBoxClr, LibGdsBox,
  LibGdsBoxMesh, LibGdsBoxCmp, LibGdsBoxClr, LibGdsBoxClr
};

#ifdef GRAPHIC_SMALL
const linfuncp FAR boxfunc3[]={
  LibGdsLine, LibGdsLine, LibGdsLineClr, LibGdsLineMesh,
  LibGdsLine, LibGdsLineCmp
};

const boxfuncp FAR boxfunc2[]={
  LibGdsBox, LibGdsBox, LibGdsBoxClr, LibGdsBoxMesh,
  LibGdsBox, LibGdsBoxCmp
};

const byte FAR boxdm_area[]={
  0, IB_GDS_AND, IB_GDS_AND, IB_GDS_MESH,
  0, IB_GDS_REV
};
#endif

const byte FAR fillmode[MAXFILLMODES]={
  IB_GDS_MESH, IB_GDS_MESH, IB_GDS_AND, IB_GDS_REV,
  IB_GDS_MESH, IB_GDS_MESH, IB_GDS_AND, IB_GDS_AND
};

#ifdef GRAPHIC_SMALL
#ifndef GRAPHIC_SMALL2
#define BOX_SUBSTYLE_VALID(n)   ((n) ? (((n & 0x03) == n) ? TRUE : (n == 5)) : FALSE)
#else
static int NEAR box_substyle_valid (int style)
{
  return (style ? (((style & 0x03) == style) ? TRUE : (style == 5)) : FALSE);
}
#define BOX_SUBSTYLE_VALID(n)   box_substyle_valid (n)
#endif
#endif

#ifndef FAST_MACRO
int checkx(int x)
{
  if (x<MINX) return FALSE;
  return x<=MAXX;
}

int checky(int y)
{
  if (y<MINY) return FALSE;
  return y<=MAXY;
}

int checkxy(int x,int y)
{
  if (!checkx(x)) return FALSE;
  return checky(y);
}
#endif

int checkxy2(int x,int y,int x1,int y1)
{
  if (checkxy (x,y))
    return (checkxy (x1,y1));
  return (FALSE);
}

/*int inrange(int x,int x1,int x2)
{
  int h;
  if (x1>x2) { h=x1; x1=x2; x2=h;}
  if (x<x1) return FALSE;
  return x<=x2;
}*/

#ifndef LSI_C
int pset(int x,int y,int z)
{
  if (!checkxy(x,y)) return 65;
  psetfunc[z & DRAWMODEMASK](x,y);
  return 0;
}
#endif

int Box(int x1,int y1,int x2,int y2,int mode)
{
  int area, border, shadow;
  int x1a, y1a, x2a, y2a;
  if (checkxy2 (x1, y1, x2, y2))
  {
    if (mode < 1000)
    {
      if (mode < 8)
        boxfunc[mode & DRAWMODEMASK] (x1, y1, x2, y2);
      else
        LibCngeBoxSub (x1, y1, x2, y2, FILLMODE (mode));
    }
    else /* new modes */
    {
      area = mode % 10;
      border = (mode/10) % 10;
      shadow = (mode/100) % 10;
      x1a = x1 + 1;
      y1a = y1 + 1;
      x2a = x2 - 1;
      y2a = y2 - 1;

#ifdef GRAPHIC_SMALL
      if (BOX_SUBSTYLE_VALID (area))
        LibCngeBoxSub (x1a, y1a, x2a, y2a, boxdm_area[area]);
      if (area == 1)
        LibCngeBoxSub (x1a, y1a, x2a, y2a, IB_GDS_REV);

      if (BOX_SUBSTYLE_VALID (border))
        boxfunc2[border] (x1, y1, x2, y2);

      if (BOX_SUBSTYLE_VALID (shadow))
      {
        x1a = x1 + 1;
        y1a = y1 + 1;

        boxfunc3[shadow] (x1a, y2a, x2a, y2a);
        boxfunc3[shadow] (x2a, y1a, x2a, (shadow == 5) ? y2a : y2);
      }

#else
      switch (area)
        {
          /* case 0: do nothing */
        case 1:
          LibCngeBoxSub(x1a,y1a,x2a,y2a,IB_GDS_AND); /* Clear */
          LibCngeBoxSub(x1a,y1a,x2a,y2a,IB_GDS_REV); /* Reverse */
          break;
        case 2:
          LibCngeBoxSub(x1a,y1a,x2a,y2a,IB_GDS_AND); /* Clear */
          break;
        case 3:
          LibCngeBoxSub(x1a,y1a,x2a,y2a,IB_GDS_MESH);/* Shade */
          break;
        case 5:
          LibCngeBoxSub(x1a,y1a,x2a,y2a,IB_GDS_REV); /* Reverse */
          break;
        }
      switch (border)
        {
          /* case 0: do nothing */
        case 1:
          LibGdsBox(x1,y1,x2,y2); /* Overwrite */
          break;
        case 2:
          LibGdsBoxClr(x1,y1,x2,y2);  /* Clear */
          break;
        case 3:
          LibGdsBoxMesh(x1,y1,x2,y2); /* Dotted line */
          break;
        case 5:
          LibGdsBoxCmp(x1,y1,x2,y2);  /* inverse */
          break;
        }
      x1a=x1+1; y1a=y1+1;
      switch (shadow)
        {
          /* case 0: do nothing */
        case 1:
          LibGdsLine(x1a,y2a,x2a,y2a);
          LibGdsLine(x2a,y1a,x2a,y2);
          break;
        case 2:
          LibGdsLineClr(x1a,y2a,x2a,y2a);
          LibGdsLineClr(x2a,y1a,x2a,y2);
          break;
        case 3:
          LibGdsLineMesh(x1a,y2a,x2a,y2a);
          LibGdsLineMesh(x2a,y1a,x2a,y2);
          break;
        case 5:
          LibGdsLineCmp(x1a,y2a,x2a,y2a);
          LibGdsLineCmp(x2a,y1a,x2a,y2a);
          break;
        }
#endif
    }
  }
  else
    return (65);
  return (0);
}

#define round(f) ((int)floor((f)+0.5))

int Line (int x1, int y1, int x2, int y2, int color)
{
  FLOATT x1f, y1f, x2f, y2f;
  FLOATT f, lx, ly;
  if (checkxy2 (x1, y1, x2, y2))
  {
    if ((x1 == x2) && (y1 == y2))
      pset (x1, y1, color);
    else
      linefunc[color & DRAWMODEMASK] (x1, y1, x2, y2);
    return (0);
  }
  else
  {
    if (x1 < x2)
    { x1f = x1; x2f = x2; y1f = y1; y2f = y2; }
    else
    { x1f = x2; x2f = x1; y1f = y2; y2f = y1; }
    if ((x2f < MINX) || (x1f > MAXX))
      return (0);

    if ((x1f < MINX) || (x2f > MAXX))
    {
      lx = x2f - x1f;
      ly = y2f - y1f;
      if (x1f < MINX)
      {
        f = (MINX - x1f) / lx;
        x1f = MINX;
        y1f = y1f + f * ly;
      }
      else
      {
        f = (MAXX - x1f) / lx;
        x2f = MAXX;
        y2f = y1f + f * ly;

      }
    }
    /*if (x2f > MAXX)
    {
      lx = x2f - x1f;
      ly = y2f - y1f;
      f = (MAXX - x1f) / lx;
      x2f = MAXX;
      y2f = y1f + f * ly;
    }*/

    if (y1f > y2f)
    {
      lx = x1f; x1f = x2f; x2f = lx;
      ly = y1f; y1f = y2f; y2f = ly;
    }

    if ((y2f < MINY) || (y1f > MAXY))
      return (0);

    if ((y1f < MINY) || (y2f > MAXY))
    {
      lx = x2f - x1f;
      ly = y2f - y1f;
      if (y1f < MINY)
      {
        f = (MINY - y1f) / ly;
        y1f = MINY;
        x1f = x1f + f * lx;
      }
      else
      {
        f = (MAXY - y1f) / ly;
        y2f = MAXY;
        x2f = x1f + f * lx;
      }
    }

    /*if (y2f > MAXY)
    {
      lx = x2f - x1f;
      ly = y2f - y1f;
      f = (MAXY - y1f) / ly;
      y2f = MAXY;
      x2f = x1f + f * lx;
    }*/
    x1 = round (x1f); y1 = round (y1f);
    x2 = round (x2f); y2 = round (y2f);
    linefunc[color & DRAWMODEMASK] (x1, y1, x2, y2);
  }
  return (0);
}

int Ellipse (int center_x, int center_y, int rx, int ry, int co)
{
  /* intermediate terms to speed up loop */
  long rx2 = rx * rx, rx2_2 = rx2 << 1, rx2_4 = rx2_2 << 1;
  long ry2 = ry * ry, ry2_2 = ry2 << 1, ry2_4 = ry2_2 << 1;

  long rxry2 = rx * ry2_2;

  long t8 = rxry2 << 1, t9 = 0L;
  long d1 = rx2_2 - rxry2 + (ry2 >> 1);	/* error terms */
  long d2 = (rx2 >> 1) - t8 + ry2_2;

  register int x = rx, y = 0;	/* ellipse points */

  while (d2 < 0)			/* til slope = -1 */
  {
    /* draw 4 points using symmetry */
    pset (center_x + x, center_y + y, co);
    pset (center_x + x, center_y - y, co);
    pset (center_x - x, center_y + y, co);
    pset (center_x - x, center_y - y, co);
    ++y;		/* always move up here */
    t9 += rx2_4;
    if (d1 < 0)	/* move straight up */
    {
      d1 += t9 + rx2_2;
      d2 += t9;
    }
    else		/* move up and left */
    {
      --x;
      t8 -= ry2_4;
      d1 += t9 + rx2_2 - t8;
      d2 += t9 + ry2_2 - t8;
    }
  }

  do 				/* rest of top right quadrant */
  {
    /* draw 4 points using symmetry */
    pset (center_x + x, center_y + y, co);
    pset (center_x + x, center_y - y, co);
    pset (center_x - x, center_y + y, co);
    pset (center_x - x, center_y - y, co);
    --x;		/* always move left here */
    t8 -= ry2_4;
    if (d2 < 0)	/* move up and left */
    {
      ++y;
      t9 += rx2_4;
      d2 += t9 + ry2_2 - t8;
    }
    else		/* move straight left */
    d2 += ry2_2 - t8;
  } while (x >= 0);
  return (0);
}

