#include <string.h>
#include <math.h>
#include "owbasic.h"
#include "inter.h"
#include "graphic.h"

/* height of fonts */
#define FONTS 6
uchar fontheight[FONTS]={8,8,7,12,7,5};
/* gap between two textlines */
uchar fontgap[FONTS]={1,1,1,2,1,1};
#ifdef __HITACHI__
uchar libfonts[4]={IB_PFONT1,IB_PFONT2,IB_CG57FONT,IB_PFONT3};
#endif
int extstringsize (int font, unsigned char* text)
{
  int i, l;
  int ml;
  int sl;
  unsigned char a;
  l = strlen ((char*) text);
  sl = 0; ml = 0;
  for (i = 0; i < l; ++i)
  {
    a = text[i];
    if (a != NEWLINE)
    {
      if (font == 4)
        sl += fn57[a * 8 + 7];
      else
        sl += 4;
    }
    else
    {
      if (sl > ml)
        ml = sl;
      sl = 0;
    }
  }
  if (sl > ml)
    ml = sl;
  return (ml);
}

int stringsize (uchar* text, int font)
{
  uchar* hptr;
  uchar* hptr1;
  uchar hc;
  unsigned int ml, sl;
  if (font < 4)
  {
    ml=0;
    hptr = text;
    hptr1 = text;
    while (*hptr != 0)
    {
      while ((*hptr != 0) && (*hptr != NEWLINE))
        ++hptr;
      hc = *hptr; *hptr = 0;
      sl = LibGetProStrSize (LFONT(font), hptr1);
      *hptr = hc;
      if (hc != 0)
        ++hptr;
      hptr1 = hptr;
      if (sl > ml)
        ml = sl;
    }
    return (ml);
  }
  else
    return extstringsize (font,text);
}

int stringheight (uchar* text, int font)
{
  int c;
  c = countc (NEWLINE, text);
  return fontheight[font] * (c + 1) + c * fontgap[font];
}

void extdrawstring (int font, int x0, int y0, unsigned char *text, int xl)
{
  int i, l;
  unsigned char bm, pattern;
  int xp, yp;
  int x, y, xe;
  unsigned char a;
  l = strlen ((char*) text);
  x = x0; y = y0;
  for (i = 0; i < l; ++i)
  {
    a = text[i];
    xe = x + fn57[a * 8 + 7];
    if (xe <= x0 + xl)
      for (yp = 0; yp < 7; ++yp)
      {
        pattern = fn57[a * 8 + yp];
        bm = 0x20;
        for (xp = x; xp < xe; ++xp)
        {
          if ((pattern & bm) > 0)
            LibGdsDotOn (xp, yp + y);
          else
            LibGdsDotOff (xp, yp + y);
          bm >>= 1;
        }
      }
    x = xe;
  }
}

int drawstring (uchar *text, int x, int y, int font, int xl)
{
  unsigned char *hptr;
  unsigned char *hptr1;
  unsigned char hc;

  if (!text)
    return (0);
  if (xl < 0)
  {
    xl = x;
    x -= stringsize (text, font);
    if (x < 0)
      x = 0;
  }
  hptr=text;
  while (*hptr!=0)
  {
    if (!checkxy (x,y))
      return (65);
    hptr1 = hptr;
    while ((*hptr != 0) && (*hptr != NEWLINE))
      ++hptr;
    hc = *hptr;
    if (hptr != hptr1)
    {
      *hptr = 0;
      switch (font)
      {
      case 0:
      case 1:
      case 2:
      case 3:
        LibPutProStr (LFONT(font), x, y, hptr1, xl);
        break;
      case 4:
        extdrawstring (font, x, y, hptr1, xl);
        break;
      case 5:
        LibPut35Str (x, y, hptr1);
        break;
      }
      *hptr=hc;
    }
    if (hc!=0) hptr++;
    y+=fontheight[font]+fontgap[font];
  }
  return 0;
}

int drawtext (uchar* text, int x1, int y1, int x2, int y2, int font)
{
  int length,height;
  int xt,yt;
  int sx, sy;
  int fi;
  int ff[] = {3,0};

  if (!text)
    return (0);
  if (!checkxy2 (x1, y1, x2, y2))
    return (65);
  sy = y2 - y1 - 2;
  sx = x2 - x1 - 2;

  if (font == -1)
  {
    /* which font */
    fi = 0;
    while ((fi < 1) && (stringheight (text, ff[fi]) > sy))
      ++fi; /* biggest font */
    while ((fi < 1) && (stringsize (text, ff[fi]) > sx))
      ++fi; /* biggest font */
    font = ff[fi];
  }

  if (font < 0 || font > FONTS - 1)
    return (62);
  length = stringsize (text, font);
  height = stringheight (text, font);
  xt = max ((x1 + x2 + 1 - length) / 2, x1);
  yt = max ((y1 + y2 - height) / 2 + 1, y1);
  return (drawstring (text, xt, yt, font, x2-x1));
}

int textbox(uchar *text,int x1,int y1,int x2,int y2,int font)
{
  int rc;
  rc=Box(x1,y1,x2,y2,10);  /* clear box */
  if (rc!=0) return rc;
  rc=drawtext(text,x1, y1, x2, y2, font);
  if (rc!=0) return rc;
  return Box(x1,y1,x2,y2,1);   /* draw frame */
}

void floatinput (FLOATT* f)
{
  uchar formatstr[] = "%.12lf";
  TCHSTS tsts;
  CALWRAM calram;
  int i, n;
  int cpos;
  char nmcbuf[33];
  char* cp;
  FLOATT res;

  SCREEN_SAVE ();

  /* Touch settings */
  LibTchStackPush (NULL);
  LibTchStackPush (TchHardIcon);
  LibTchStackPush (TchModeIcon);

  /* Initializing calculator. */
  LibCalKeyInit (&calram); /* Initializes calculator keyboard. */

  /* initialize CALRAM (MB,4.40) */
  if (*f < 1.0)
    formatstr[3] = '1';
  sprintf (nmcbuf, formatstr, *f);
  cp = nmcbuf + strlen (nmcbuf) - 1;
  while (*cp == '0')
    --cp;
  calram.calxbuf[0] = 0xFF;
  calram.calxbuf[1] = (*nmcbuf == '-') ? 0x0C : 0x04;
  for (i = 2; i < CAL_BUFSIZE; ++i)
    if (cp >= nmcbuf)
    {
      if (*cp == '.')
      {
        calram.calxbuf[0] = CAL_BUFSIZE - 1 - i;
        --cp;
      }
      if (*cp != '-')
        calram.calxbuf[i] = *cp;
      --cp;
    }
  if (calram.calxbuf[0] == 0xFF)
    calram.calxbuf[0] = CAL_BUFSIZE - 1 - i;

  LibCalKeyDsp (&calram); /* Displays Calculator keyboard. */
  LibPutDisp (); /* Reflects to the screen. */
  do
    LibCalKeyTchWait (&calram, &tsts); /* Waits for touch of calculator */
  while ((tsts.obj != OBJ_CAL_NEXT)
        && (tsts.obj != OBJ_CAL_EQUAL)
        && (tsts.obj != OBJ_HIC_ESC));

  if (calram.calxbuf[0] != 0xFF)
  {
    res = 0;
    cpos = calram.calxbuf[0] - (CAL_BUFSIZE - 3);
    for (i = 2; i < CAL_BUFSIZE; ++i, ++cpos)
    {
      if ((n = (calram.calxbuf[i] & 0x0F)) != 0x00)
        res += n * pow (10, cpos);
    }
    if (calram.calxbuf[1] == 0x0C)
      res = -res;
    if (tsts.obj != OBJ_HIC_ESC)
      *f = res;
  }
  LibTchStackPop ();
  LibTchStackPop ();
  LibTchStackPop ();

  SCREEN_RESTORE ();
}

void strinpbkg (const uchar* topic)
{
  LibClrDisp ();
  LibGdsBox (0, 0, 159, 85);
  LibGdsDotOff (0, 0);
  LibGdsDotOff (159, 0);
  LibGdsDotOff (0, 85);
  LibGdsDotOff (159, 85);
  drawstring ((uchar*) topic, 2, 2, IB_PFONT2, 157);
}


static int dboxfont = 0;
void dbox (int x1, int y1, int x2, int y2, uchar* caption)
{
  Box (x1 + 1, y1 + 1, x2 + 1, y2 + 1, 1);
  /*Box (x1,     y1,     x2,     y2,     1012);*/
  textbox (caption, x1, y1, x2, y2, dboxfont);
  LibGdsDotOff (x1, y1);
  LibGdsDotOff (x2 + 1, y2 + 1);
}

int tchinside (int x1, int y1, int x2, int y2)
{
  check_touch ();
  return ((*tchxp >= x1) && (*tchxp <= x2)
       && (*tchyp >= y1) && (*tchyp <= y2));
}

int qbox (int x1, int y1, int x2, int y2)
{
  bool inside = FALSE;

  if (newtouched)
    if (tchinside (x1, y1, x2, y2))
    {
      Box (x1 + 1, y1 + 1, x2 - 1, y2 - 1, 11);
      LibPutDisp ();

      while (touched && (inside = tchinside (x1, y1, x2, y2)))
        ;

      Box (x1 + 1, y1 + 1, x2 - 1, y2 - 1, 11);
      LibPutDisp ();
      newtouched = FALSE;
    }
  return (inside);
}


void msgbox1 (uchar* text)
{
  msgbox (text, NULL, 0.0, MBX_OK);
}

int msgbox (uchar* text, uchar* caption, FLOATT time, int type)
{
  word save[1301];
  int x1, y1, x2, y2;
  int my1, my2;
  int xs, ys;
  int captsize = 4;
  int tmp = 0;
  int res = 0;

  if (caption)
    if (*caption)
    {
      captsize = 14;
      tmp = stringsize (caption, 1);
    }
  xs = stringsize (text, 0);
  xs = ((tmp > xs) ? tmp : xs) + 4;
  if ((type >= MBX_YESNO) && (xs < 88))
    xs = 88;
  if (xs > 156)
    xs = 156;
  ys = stringheight (text, 0) + captsize;
  if (type == MBX_YESNO)
    ys += 23;
  x1 = 79 - (xs >> 1);
  y1 = 79 - (ys >> 1);
  x2 = x1 + xs;
  y2 = y1 + ys;

  if (type > MBX_DRAW)
  {
    save[0] = xs + 2; save[1] = ys + 2;
    LibGetGraph (x1, y1, xs + 2, ys + 2, (byte FAR*) (save + 2)); /* save screen area */
  }

  dbox (x1, y1, x2, y2, NULL);

  if (captsize == 14)
  {
    LibPutProStr (IB_PFONT2, x1 + 2, y1 + 2, caption, xs);
    LibGdsLine (x1, y1 + 10, x2, y1 + 10);
  }
  drawstring (text, x1 + 2, y1 + captsize - 1, 0, xs);


  switch (type)
  {
  case MBX_DRAW:
    return (0);

  case MBX_WAIT:
    LibPutDisp ();

    PauseTimer ();
    InitSync (time);
    while (!Sync ());
    ContinueTimer ();
    break;

  case MBX_OK:
    LibPutDisp ();
    x_waittouch ();
    break;

  case MBX_YESNO:
    res = -1;
    dboxfont = 3;
    my1 = y2 - 18;
    my2 = y2 - 3;
    dbox (x2 - 64, my1, x2 - 35, my2, msg_yes);
    dbox (x2 - 32, my1, x2 - 3,  my2, msg_no);
    dboxfont = 0;
    LibPutDisp ();

    while (res < 0)
    {
      newtouched = FALSE;
      check_touch ();
      if (qbox (x2 - 64, my1, x2 - 35, my2))
        res = 1;
      if (qbox (x2 - 32, my1, x2 - 3,  my2))
        res = 0;
    };
    break;
  }

  LibPutGraph (x1, y1, (byte FAR*) save);
  LibPutDisp ();
  newtouched = FALSE;
  return (res);
}



