#include "modeldef.h"
#include <string.h>
#include "owbasic.h"
#include "shared.h"
#include "file.h"
#include "graphic.h"
#include "inter.h"

/* interpreter interface for filefunctions */

/* ----------- addin filename ---------------*/
void x_namedfile (void)
{
  int rc;
  int subaddr = ipop ();
  int modeaddr = ipop ();
  int mode, submode;

  rc = namedfile (STOS, &mode, &submode);

  if (rc != 0)
    rterror (FN_NAMEDFILE, 39);
  xdata[modeaddr] = mode;
  xdata[subaddr] = submode;
  spop1 ();
#ifdef LSI_C
  bk_counter = 0;
#endif
}

/* -------- MEMO ----------------------------*/
void x_loadmemo (void)
{
  int addr, size;
  int sub, nr;
  int rc;
  int total, totaladdr;

  totaladdr = ipop ();
  nr = ipop (); sub = ipop ();
  addr = ipop ();
  size = xdata[addr - 1];
  rc = loadmemo (addr, size, sub, nr, &total);

  if (rc != 0)
    rterror (FN_LOADMEMO, rc);
  if (totaladdr >= 0)
    xdata[totaladdr] = total;
}

void x_loadmemo3 (void)
{
  ipush (-1);
  x_loadmemo ();
}

void x_loadmemoname (void) /* LOADMEMONAME ARR$[], sub%, name$, sig$ [, LINES%] [, FP%] */
{
  int addr, size;
  int sub;
  int rc;
  int total, totaladdr;
  word fp;
  int fpaddr;

  fpaddr = ipop ();
  totaladdr = ipop ();
  sub = ipop ();
  addr = ipop ();
  size = xdata[addr - 1];
  rc = loadmemonamed (addr, size, sub, SST1, (STOS[0]) ? STOS : NULL, &total, &fp);

  if (rc)
    rterror (FN_NAMEDFILE, rc);
  if (totaladdr >= 0)
    xdata[totaladdr] = total;
  if (fpaddr >= 0)
  {
    if (fp != 0xFFFF)
      xdata[fpaddr] = fp;
    else
      xdata[fpaddr] = -1;
  }
  spop2 ();
}

void x_loadmemoname5 (void)
{
  ipush (-1);
  x_loadmemoname ();
}

void x_loadmemoname4 (void)
{
  ipush (-1);
  x_loadmemoname5 ();
}

void x_savememo (void)
{
  int addr, size;
  int sub, nr;
  int anz;
  int rc;

  anz = ipop ();
  nr = ipop (); sub = ipop ();
  addr = ipop ();
  size = xdata[addr - 1];
  rc=savememo (addr, size, sub, nr, anz);
  if (rc)
    rterror (FN_SAVEMEMO, rc);
#ifdef LSI_C
  bk_counter = 0;
#endif
}

void x_savememo3 (void)
{
  ipush (-1);
  x_savememo ();
}

void x_savememo2 (void)
{
  ipush (-1);
  x_savememo3 ();
}


void x_loadmemofp (void)
{
  int addr, size;
  int sub;
  word fp;
  int rc;
  int total, totaladdr;

  totaladdr = ipop ();
  fp = ipop (); sub = ipop ();
  addr = ipop ();
  size = xdata[addr - 1];
  rc = loadmemofp (addr, size, sub, fp, &total);

  if (rc)
    rterror (FN_LOADMEMO, rc);
  if (totaladdr >= 0)
    xdata[totaladdr] = total;
}

void x_loadmemofp3 (void)
{
  ipush (-1);
  x_loadmemofp ();
}

void x_savememofp (void)
{
  int addr, size;
  int sub;
  word fp;
  int fpaddr;
  int anz;
  int rc;

  anz = ipop ();
  fpaddr = ipop (); fp = xdata[fpaddr];
  sub = ipop ();
  addr = ipop ();
  size = xdata[addr - 1];
  rc = savememofp (addr, size, sub, &fp, anz);
  xdata[fpaddr] = fp;
  if (rc != 0)
    rterror (FN_SAVEMEMO, rc);
#ifdef LSI_C
  bk_counter = 0;
#endif
}

void x_savememofp3 (void)
{
  ipush (-1);
  x_savememofp ();
}

/* -------- general files -------------------*/
void x_savefile (void)
{
  int nr, min, mode;
  nr = ipop ();
  min = ipop ();
  mode = ipop ();
  if (!save_file (mode, min, nr))
    rterror (FN_SAVEFILE, 39);
#ifdef LSI_C
  bk_counter = 0;
#endif
}

void x_loadfile()
{
  int nr,min,mode;
  nr=ipop(); min=ipop(); mode=ipop();
  if (!load_file(mode,min,nr)) rterror(FN_LOADFILE,39);
}

void x_getfp (void) /* GETFP%(mode%, submode%, nr%) (MB,4.10) */
{
  int mode, submode, nr;

  nr      = ipop ();
  submode = ipop ();
  mode    = TOS;

  TOS = get_file_fp (mode, submode, nr);
  if (TOS == 0xFFFF)
    rterror (FN_GETFP, 39);
}

void x_fbbhash (void) /* FBBHASH CRC1%, CRC2% [max%=-1] [,mode%=1] (MB,4.10) */
{
  int crc2i, crc1i, max, mode;
  int crc1, crc2;

  mode  = ipop ();
  max   = ipop ();
  crc2i = ipop ();
  crc1i = ipop ();

  if (max == -1)
    max = FBBNUM;

  if (max < 0 || max >= 3072)
    rterror (FN_FBBHASH, 60);

  crc (FBB, max, &crc1, &crc2, mode);
  xdata[crc1i] = crc1;
  xdata[crc2i] = crc2;
}

void x_fbbhash3 (void)
{
  ipush (default_hash_mode);
  x_fbbhash ();
}

void x_fbbhash2 (void)
{
  ipush (-1);
  ipush (default_hash_mode);
  x_fbbhash ();
}

void x_fbhash (void) /* FBHASH CRC1%, CRC2% [,mode%=1] (MB,4.10) */
{
  int crc2i, crc1i, mode;
  int crc1, crc2, max;

  mode  = ipop ();
  crc2i = ipop ();
  crc1i = ipop ();

  for (max = 0; FB[max] != 0; ++max)
    if (max >= 3072)
      rterror (FN_FBHASH, 60);

  crc (FB, max, &crc1, &crc2, mode);
  xdata[crc1i] = crc1;
  xdata[crc2i] = crc2;
}

void x_fbhash2 (void)
{
  ipush (default_hash_mode);
  x_fbhash ();
}


/* ------------- Quickmemo -------------------------*/
void x_loadqmfp()
{
  int sub, addr, size;
  word fp;
  int rc;
  fp = ipop();  sub =ipop();
  addr=ipop();
  size=xdata[addr-1];

  rc=loadqmfp(addr,size,sub,fp);

  if (rc!=0) rterror(FN_LOADQMFP,rc);
}

void x_loadqm (void)
{
  int sub, addr, size, nr;
  int rc;
  nr=ipop();  sub =ipop();
  addr=ipop();
  size=xdata[addr-1];

  rc=loadqm(addr,size,sub,nr);
  if (rc!=0) rterror(FN_LOADQM,rc);
}

int default_hash_mode = 1;

void x_loadqmhash (void)
{
  int mode = ipop ();
  int fpaddr = ipop ();
  int crc2 = ipop ();
  int crc1 = ipop ();
  int sub = ipop ();
  int addr = ipop ();
  int size = xdata[addr-1];
  word fp;
  int rc = loadqmhash (addr, size, sub, crc1, crc2, &fp, mode);
  if (rc != 0)
    rterror (FN_LOADQMHASH, rc);
  if (fpaddr >= 0)
    xdata[fpaddr] = (fp == 0xFFFF) ? -1 : fp;
#ifdef LSI_C
  bk_counter = 0;
#endif
}

void x_loadqmhash5 (void)
{
  ipush (default_hash_mode); /* use default mode */
  x_loadqmhash ();
}

void x_loadqmhash4 (void)
{
  ipush (-1); /* default: don't store filepointer */
  x_loadqmhash5 ();
}

void x_iconhash (void)
{
  /*int crc1,crc2;
  int mode=ipop();
  int crc2addr=ipop();
  int crc1addr=ipop();
  int addr=ipop();
  int size=xdata[addr-1];
  int rc=iconhash(addr,size,&crc1,&crc2,mode);
  if (rc!=0) rterror(FN_ICONHASH,rc);
  xdata[crc1addr]=crc1;
  xdata[crc2addr]=crc2;*/

  int mode = ipop ();
  int crc2addr = ipop ();
  int crc1addr = ipop ();
  int addr = ipop ();
  int size = xdata[addr - 1];
  int rc = iconhash (addr, size, &xdata[crc1addr], &xdata[crc2addr], mode);
  if (rc != 0)
    rterror (FN_ICONHASH, rc);
}

void x_iconhash3 (void)
{
  ipush (default_hash_mode);
  x_iconhash ();
}

void x_saveqm (void)
{
  int addr, size;
  int nr;
  int sub;
  int rc;
  nr = ipop (); sub = ipop ();
  addr = ipop ();
  size = xdata[addr - 1];

  rc = saveqm (addr, size, sub, nr);

  if (rc != 0)
    rterror (FN_SAVEQM, rc);
#ifdef LSI_C
  bk_counter = 0;
#endif
}

void x_saveqm2 (void)
{
  ipush (-1);
  x_saveqm ();
}

void x_saveqmfp (void)
{
  int addr, size;
  int fpaddr;
#ifdef __HITACHI__
  word fp;
#endif
  int sub;
  int rc;

  fpaddr = ipop (); sub = ipop ();
  addr = ipop ();
  size = xdata[addr - 1];
#ifdef __HITACHI__
  fp = xdata[fpaddr];
  rc = saveqmfp (addr, size, sub, &fp);
  xdata[fpaddr] = fp;
#else
  rc = saveqmfp (addr, size, sub, &xdata[fpaddr]);
#endif
if (rc != 0)
    rterror (FN_SAVEQMFP, rc);
  /*xdata[fpaddr] = fp;*/
#ifdef LSI_C
  bk_counter = 0;
#endif
}

/*
Sequential file handling
*/

void x_filefind (void) /* FILEFIND%(mode%, submode%, name$) (MB,4.24) */
{
  int submode;

  submode = ipop ();
  if (!filefind (TOS, submode, STOS, NULL, (word*) (&TOS)))
    TOS = -1;
#ifdef __HITACHI__
  else
    TOS = *((word*) (&TOS));
#endif
  spop1 ();
}

void x_filenext (void)
{
  int fpp, sub, main;
  fpp = ipop(); sub = ipop (); main = ipop ();
  xdata[fpp] = filenext (main, sub, xdata[fpp]);
#ifdef __HITACHI__
  if (xdata[fpp] == 0xFFFF)
    xdata[fpp] = -1;
#endif
}

void x_fileprev (void)
{
  int fpp,sub, main;
  fpp = ipop (); sub = ipop (); main = ipop ();
  xdata[fpp] = fileprev (main, sub, xdata[fpp]);
#ifdef __HITACHI__
  if (xdata[fpp] == 0xFFFF)
    xdata[fpp] = -1;
#endif
}

void x_filewrite (void)
{
  int fpp, sub, main;
  fpp = ipop (); sub = ipop (); main = ipop ();
  xdata[fpp] = filewrite (main, sub, xdata[fpp]);
  if (xdata[fpp] == 0xFFFF)
  {
    rterror (FN_FILEWRITE, 39);
#ifdef __HITACHI__
    xdata[fpp] = -1;
#endif
  }
#ifdef LSI_C
  bk_counter = 0;
#endif
}

void x_fileread (void) /* FILEREAD mode%, submode%, fp% [, blocks%] (MB,4.50) */
{
  int fp, sub, main, blocks;

  blocks = ipop (); fp = ipop (); sub = ipop (); main = ipop ();
  SetFP (main, sub, fp);
  if (!LibFileReadEx (&fb, &finf, blocks))
    rterror (FN_FILEREAD, 39);
}

void x_fileread3 (void)
{
  ipush (FBSIZE / FILE_BLOCK_SIZE);
  x_fileread ();
}

void x_fileremove (void)
{
  int fp, sub, main;
  fp = ipop (); sub = ipop (); main=ipop ();
  if (!fileremove (main, sub, fp))
    rterror (FN_FILEREMOVE, 39);
#ifdef LSI_C
  bk_counter = 0;
#endif
}

void x_filexchg (void)
{
  int fp, sub, main, fpz;
  fpz = ipop (); fp = ipop (); sub = ipop (); main = ipop ();
  if (filemove (main, sub, fp, fpz) != TRUE)
    rterror (FN_FILEXCHG, 39);
}

void x_flashremake (void)
{
  LibFileRemake ();
#ifdef LSI_C
  bk_counter = 0;
#endif
}

/* read date and time */
void x_fbdate (void)
{
  int sel;
  int rc;

  spush1 ();
  sel = ipop ();
  rc = fbdate (sel, STOS);
  if (rc != 0)
    rterror (FN_FBDATE, rc);
}

void x_fbtime (void)
{
  int sel;
  int rc;

  spush1 ();
  sel = ipop ();
  rc = fbtime (sel, STOS);
  if (rc != 0)
    rterror (FN_FBTIME, rc);
}

void x_fbputdate (void)
{
  int sel;
  int rc;
  sel = ipop();
  rc = fbputdate (sel, STOS);
  if (rc != 0)
    rterror (FN_FBDATE, rc);
  spop1 ();
}

void x_fbputtime (void)
{
  int sel;
  int rc;
  sel = ipop ();
  rc = fbputtime (sel, STOS);
  if (rc != 0)
    rterror (FN_FBTIME, rc);
  spop1 ();
}

/*
 Read and write data to FILEBUFFER (bytewise)
*/

/* direct read from filebuffer as byte */
void x_fbget (void)
{
  int byteoffset = ipop ();
  ipush (FB[byteoffset]);
}

/* direct write to filebuffer as byte */
void x_fbput (void)
{
  int val = ipop ();
  int byteoffset = ipop ();
  FB[byteoffset] = val;
}

void x_getfbbsize (void)
{
  ipush (FBBNUM);
}

void x_setfbbsize (void)
{
  FBBNUM = ipop ();
}

void x_cbput (void) /* direct write to clipboard as byte (MB,4.20) */
{
  int val = ipop ();
  int byteoffset = ipop();
  SYS_INP_BUF[byteoffset] = val;
}

void x_cbget (void) /* direct read from clipboard as byte (MB,4.20) */
{
  int byteoffset = ipop();
  ipush (SYS_INP_BUF[byteoffset]);
}

/* write/read of string data to text file buffer and move pointer */
void x_fbinit (void) { fbinit (); }

void x_fbstore (void)
{
  int del = ipop ();
  int rc = fbstore (STOS, del);
  if (rc != 0)
    rterror (FN_FBSTORE, rc);
  spop1 ();
}

void x_fbstore1 (void)
{
  ipush (1);
  x_fbstore ();
}

void x_fbstring (void)
{
  int del = ipop ();
  int rcaddr = ipop (); /* addr of returncode variable */
  int rc, rcf;
  spush1 ();
  rc = fbstring (STOS, &rcf, del);
  if (rc!=0) rterror (FN_FBSTRING, rc);
  if (rcaddr >= 0)
    xdata[rcaddr] = rcf;
}

void x_fbstring1 (void)
{
  ipush (1);
  x_fbstring ();
}

void x_fbstring0 (void)
{
  ipush(-1);
  x_fbstring1 ();
}

/*
  write and read  binary filebuffer FBB
*/

#ifdef NEVER_DEFINED
void x_fbbinit (void)  /* Initialise read/write process */
{
  int rc = fbbinit ();
  if (rc != 0)
  rterror (FN_FBBINIT, rc);
}
#endif

void x_fbbtell (void) /* FBBTELL%() (MB,4.24) */
{
  ipush (fbbtell ());
}

void x_fbbseek (void)
{
  int seekval = ipop ();
  int rc = fbbseek (seekval);
  if (rc != 0)
    rterror (FN_FBBSEEK, rc);
}

void x_fbbputint (void)
{
  /* write int to FBB, move pointer and set FBBNUM*/
  int val = ipop ();
  int rc;
  rc = fbbputint (val);
  if (rc != 0)
    rterror (FN_FBBPUTINT, rc);
}

void x_fbbputbyte (void)
{
  /* write byte to FBB, move pointer and set FBBNUM*/
  int val = ipop ();
  int rc;
  rc = fbbputbyte (val);
  if (rc != 0)
    rterror (FN_FBBPUTBYTE, rc);
}

void x_fbbputfloat()
{
  /* write FLOAT to FBB, move pointer and set FBBNUM*/
  FLOATT val = fpop ();
  int rc;
  rc = fbbputfloat (val);
  if (rc != 0)
    rterror (FN_FBBPUTFLOAT, rc);
}

void x_fbbputstring (void)
{
  /* write string to FBB, move pointer and set FBBNUM*/
  int rc;
  rc = fbbputstring (STOS);
  if (rc != 0)
    rterror (FN_FBBPUTSTRING, rc);
  spop1 ();
}

void x_fbbint (void)
{
  /* read int from FBB, move pointer */
  int rc, val;
  rc = fbbint (&val);
  if (rc != 0)
    rterror (FN_FBBINT, rc);
  ipush (val);
}

void x_fbbbyte (void)
{
  /* read byte from FBB, move pointer */
  int val, rc;
  rc = fbbbyte (&val);
  if (rc != 0)
    rterror (FN_FBBBYTE, rc);
  ipush (val);
}

void x_fbbfloat (void)
{
  /* read FLOATT from FBB, move pointer */
  int rc;
  FLOATT val;
  rc = fbbfloat (&val);
  if (rc != 0)
    rterror (FN_FBBFLOAT, rc);
  fpush (val);
}

void x_fbbstring (void)
{
  /* read string from FBB, move pointer */
  int rc;
  spush1 ();
  rc = fbbstring (STOS);
  if (rc != 0)
    rterror (FN_FBBSTRING, rc);
}
