#include "owbasic.h"
#include "inter.h"
#include "graphic.h"
#include "form.h"
#include "comp.h"

/* interpreter interface for forms */

FElement FAR Form[MAXFORMELEMENTS];
int   lastelement;
uchar defaultstatus;
int   formfont;


void x_forminit (void) /* FORMINIT [defstat%] [,deffont%] */
{
  ipush (formstrings);
  x_clears (); /* free string space of former forms */

  lastelement = -1;
  formfont = ipop ();
  defaultstatus = ipop ();
}

void x_forminit1 (void)
{
  ipush (-1);
  x_forminit ();
}

void x_forminit0 (void)
{
  ipush (0);
  x_forminit1 ();
}

static int NEAR findid (int fid)
{
  int i = 0;
  int id = fid / 10;
  int subid = fid % 10;

  for (; i <= lastelement; ++i)
    if ((Form[i].id == id) && (Form[i].subid == subid))
      return (i);
  return (-1);
}

static void NEAR formtext (uchar* text, int x1, int y1, int x2, int y2)
{
  int gnr;

  if (!checkxy2 (x1, y1, x2, y2))
  {
    rterror (FN_FORMDISPLAY, 65);
    return;
  }
  if (text[0] == '#')
  {
    sscanf ((char*) text, "#%d", &gnr);
    LibPutFarData (x1 + 3, y1 + 3, gnr);
  }
  else
    drawtext (text, x1, y1, x2, y2, formfont);
}

const static char FAR formdisp_gnr[] = {16, 9, 13, 12};

void x_formdisplay (void)
{
  int i;
  int x1, y1, x2, y2;
  uchar stt;
  PFElement cf;

  if (lastelement < 0)
  {
    rterror (FN_FORMDISPLAY, 67);
    return;
  }

  for (i = 0; i <= lastelement; ++i)
  {
    cf = &Form[i];
    stt = cf->status;

    if (!(stt & 0x10)) /* hidden element? */
    {
      x1 = cf->x1; y1 = cf->y1; x2 = cf->x2; y2 = cf->y2;

      switch (cf->typ)
      {
      case 0:
        spush (formstrings + i);
        formtext (STOS, x1, y1, x2, y2);
        spop1 ();
        break;
      case 1:
      case 2:
        spush (formstrings + i);
        formtext (STOS, x1 + 15, y1, x2, y2);
        spop1 ();
        LibPutFarData (Form[i].x1 + 7, (Form[i].y1 + Form[i].y2) / 2 - 4,
         formdisp_gnr[(stt & 0x01) + ((cf->typ - 1) << 1)]);
        break;
      }

      if ((stt & 0x02) == 0) /* no frame ? */
        Box (x1, y1, x2, y2, 1);

      if ((stt & 0x04) > 0)  /* double frame */
        Box (x1 - 2, y1 - 2, x2 + 2, y2 + 2, 1);

      if ((stt & 0x08) > 0)  /* inactive */
        Box (x1, y1, x2, y2, 12);
    }
  }
  LibPutDisp ();

#ifdef LSI_C
  bk_counter = 0;
#endif
}

void x_formentry (void)
{
  PFElement cf;

  int x1, y1, x2, y2;
  int typ, id;


  id = ipop (); typ = ipop ();
  y2 = ipop (); x2  = ipop ();
  y1 = ipop (); x1  = ipop ();

  if ((typ < 0) || (typ > 2))
    typ = 0;
  if ((++lastelement) >= MAXFORMELEMENTS)
  {
    rterror (FN_FORMENTRY, 33);
    return;
  }
  if (!checkxy2 (x1, y1, x2, y2))
  {
    rterror (FN_FORMENTRY, 65);
    return;
  }

  cf = &Form[lastelement];
  cf->x1 = x1;
  cf->y1 = y1;
  cf->x2 = x2;
  cf->y2 = y2;
  cf->typ = typ;
  cf->id = id / 10;
  cf->subid = id % 10;
  cf->status = defaultstatus;

  putvar (formstrings + lastelement, STOS);
  spop1 ();
}

void x_formtouch (void)
{
  PFElement cf;
  int touchx, touchy;
  int x1, y1, x2, y2;
  int n, i;
  int sel, addr, nowait;

  nowait = ipop ();
  addr = ipop ();

  sel = -1;
  if (lastelement < 0)
  {
    rterror (FN_FORMTOUCH, 67);
    return;
  }

  LibPutDisp ();

  do
  {
    if (nowait ? check_touch () : (x_waittouch (), TRUE))
    {
      if (ex)
        return;
      touchx = *tchxp; touchy = *tchyp;

      for (n = lastelement; n >= 0; --n)
      {
        cf = &Form[n];
        x1 = cf->x1; y1 = cf->y1; x2 = cf->x2; y2 = cf->y2;

        if (((cf->status & 0x08) == 0)
          && (cf->id != 0)
          && (x1 <= touchx)
          && (y1 <= touchy)
          && (x2 >= touchx)
          && (y2 >= touchy))
        {
          switch (cf->typ)
          {
          case 1:
            cf->status ^= 0x01;
            cf->subid = cf->status & 0x01;
            break;
          case 2:
            if ((cf->status & 0x01) == 0)
            {
              for (i = 0; i <= lastelement; ++i)
                if (Form[i].id == cf->id)
                  Form[i].status &= 0xFE;
              cf->status |= 0x01;
            }
            break;
          }
          sel = cf->id * 10 + cf->subid;
          break;
        }
      }

      x_formdisplay ();

      if (sel >= 0)
      {
        if (!(cf->status & 0x10)) /* not hidden */
        {
          Box (x1, y1, x2, y2, 11);
          WaitNoTouch ();
          Box (x1, y1, x2, y2, 10);
          x_formdisplay ();
        }
        else
          WaitNoTouch ();
      }
      newtouched = FALSE;
    }
  } while ((sel < 0) && (!nowait));

  xdata[addr] = sel;

#ifdef LSI_C
  bk_counter = 0;
#endif
}

void x_formtouch1 (void)
{
  ipush (FALSE);
  x_formtouch ();
}

void x_formsetstatus (void)
{
  int sel;
  int n;

  sel = ipop ();
  n = findid (ipop ());

  if (n == -1)
    rterror (FN_FORMSETSTATUS, 61);
  else
    Form[n].status |= sel;
}

void x_formresetstatus (void)
{
  int sel;
  int n;

  sel = ipop ();
  n = findid (ipop ());

  if (n == -1)
    rterror (FN_FORMRESETSTATUS, 61);
  else
    Form[n].status &= 0xFF ^ sel;
}



void x_inputf (void) /* INPUTF F# (MB,4.40) */
{
  int addr = ipop ();
  FLOATT f = fdata (addr);
  floatinput (&f);
  fstore (addr, f);
}

void x_calcinput (void)
{
  fpush (0.0);
  floatinput (&FTOS);

  /*FLOATT f = 0.0;
  floatinput (&f);
  fpush (f);*/

#ifdef NEVER_DEFINED
  TCHSTS tsts;
  CALWRAM calram;
  bool done = FALSE;
  int i;
  /*char nmcbuf[33];
  char* cp;*/
  FLOATT res;

  /******************/
  /* Touch settings */
  /******************/
  LibTchStackClr ();
  LibTchStackPush (TchHardIcon);
  LibTchStackPush (TchModeIcon);
  LibTchInit ();

  /****************************/
  /* Initializing calculator. */
  /****************************/
  LibCalKeyInit (&calram); /* Initializes calculator keyboard. */

  /* initialize CALRAM (MB,4.40) */
  sprintf (nmcbuf, "%lg", fpop ());
  cp = nmcbuf + strlen (nmcbuf) - 1;
  calram.calxbuf[0] = 0xFF;
  calram.calxbuf[1] = (*nmcbuf == '-') ? 0x0C : 0x04;
  for (i = 2; i < CAL_BUFSIZE; ++i)
  {
    if (cp >= nmcbuf)
    {
      if (*cp == '.')
      {
        calram.calxbuf[0] = CAL_BUFSIZE - 1 - i;
        --cp;
      }
      if (*cp != '-')
        calram.calxbuf[i] = *cp;
      --cp;
    }
    else
      calram.calxbuf[i] = 0x00;
  }
  if (calram.calxbuf[0] == 0xFF)
    calram.calxbuf[0] = CAL_BUFSIZE - 1 - i;

  LibCalKeyDsp (&calram); /* Displays Calculator keyboard. */

  LibPutDisp (); /* Reflects to the screen. */
  while (!done)
  {
    LibCalKeyTchWait (&calram, &tsts); /* Waits for touch of calculator */
    switch (tsts.obj)
    {
    case OBJ_CAL_NEXT: /* "Next" */
      /***** Writing and item move processes, etc. *****/
      /*if (calram.calerror == 0x00)
        done = TRUE;
      break;*/
    case OBJ_CAL_EQUAL: /* "=" */
      /***** Writing process and others... *****/
      if (calram.calerror == 0x00)
        done = TRUE;
      break;
    case OBJ_HIC_ESC: /* ESC */
      LibCalKeyInit (&calram);
      done = TRUE;
      break;
    default:
      break;
    }
  }

  res = 0;
  for (i = 2; i < 14; ++i)
    res = res / 10.0 + (calram.calxbuf[i] & 0x0F);
  if (calram.calxbuf[1] == 0x0C)
    res = -res;
  for (i = 0; i < calram.calxbuf[0]; ++i)
    res *= 10.0;
  fpush (res);
#endif
}
