#include "owbasic.h"
#include "mymath.h"
#include "inter.h"
/*
 * Interpreter for mathematical functions
 *
 */

static PC addr;


void x_dropf (void) { fpopn (1); }

void x_sin_f() { FTOS=Sin(FTOS); }
void x_cos_f() { FTOS=Cos(FTOS); }
void x_tan_f() { FTOS=Tan(FTOS); }

void x_asin_f() { FTOS=ASin(FTOS); }
void x_acos_f() { FTOS=ACos(FTOS); }
void x_atan_f() { FTOS=ATan(FTOS); }
void x_atan2_f() { FLOATT arg2=fpop(); FTOS=ATan2(FTOS,arg2); }

void x_sinh_f() { FTOS=Sinh(FTOS); }
void x_cosh_f() { FTOS=Cosh(FTOS); }
void x_tanh_f() { FTOS=Tanh(FTOS); }

void x_asinh_f() { FTOS=ASinh(FTOS); }
void x_acosh_f() { FTOS=ACosh(FTOS); }
void x_atanh_f() { FTOS=ATanh(FTOS); }

void x_log_f() { FTOS=Log(FTOS);}
void x_ln_f() { FTOS=Ln(FTOS);}
void x_log2_f() { FTOS=Log2(FTOS);}

void x_exp_f() { FTOS=ALn(FTOS);}
void x_exp10_f() { FTOS=ALog(FTOS);}
void x_exp2_f() { FTOS=ALog2(FTOS);}

void x_abs_f() { FTOS=fabs(FTOS); }
void x_sqare_f() { FTOS=FTOS*FTOS; }
void x_sqr_f() { FTOS=Sqrt(FTOS); }

void x_xcart_f() { FLOATT fi=fpop(); FTOS=XCart(FTOS,fi); }
void x_ycart_f() { FLOATT fi=fpop(); FTOS=YCart(FTOS,fi); }
void x_radius_f() { FLOATT y=fpop(); FTOS=Radius(FTOS,y); }
void x_phi_f() { FLOATT y=fpop(); FTOS=Phi(FTOS,y); }

void x_deg_f() { FTOS=Deg(FTOS); }
void x_rad_f() { FTOS=Rad(FTOS); }
void x_grad_f() { FTOS=Grad(FTOS); }

void x_conv_fi() { FLOATT vf=fpop(); ipush((int)floor(vf)); }
void x_round() { FLOATT vf=fpop(); ipush((int)floor(vf+0.5)); }
void x_roundf() { FTOS=floor(FTOS+0.5); }
void x_floor() { FLOATT vf=fpop(); ipush((int)floor(vf)); }
void x_floorf() { FTOS=floor(FTOS); }
void x_ceil() { FLOATT vf=fpop(); ipush((int)ceil(vf)); }
void x_ceilf() { FTOS=ceil(FTOS); }

/*void x_inc_f() { int addr; GETIPC(addr); *(FLOATT FAR*)&(xdata[addr])+=1.0; }*/


void x_inc_fl1 (void) /* incf/decf procedures (MB,4.10) */
{
  *(FLOATT FAR*) &(xdata[ipop ()]) += 1.0;
}
void x_dec_fl1 (void)
{
  *(FLOATT FAR*) &(xdata[ipop ()]) -= 1.0;
}

void x_inc_fl (void)
{
  *(FLOATT FAR*) &(xdata[ipop ()]) += fpop ();
}
void x_dec_fl (void)
{
  *(FLOATT FAR*) &(xdata[ipop ()]) -= fpop ();
}


void x_add_ff() { FLOATT vf=fpop(); FTOS+=vf; }
void x_sub_ff() { FLOATT vf=fpop(); FTOS-=vf; }
void x_mul_ff() { FLOATT vf=fpop(); FTOS*=vf; }
void x_div_ff() { FLOATT vf=fpop(); FTOS/=vf; }

void x_add_fi() { FTOS+=ipop(); }
void x_sub_fi() { FTOS-=ipop(); }
void x_mul_fi() { FTOS*=ipop(); }
void x_div_fi() { FTOS/=ipop(); }

void x_add_if() { FTOS+=ipop(); }
void x_sub_if() { FTOS=ipop()-FTOS; }
void x_mul_if() { FTOS*=ipop(); }
void x_div_if() { FTOS=ipop()/FTOS; }

void x_fract() { FTOS=FTOS-floor(FTOS); }

void x_pow_fi()
{
  int vi=ipop();
  FLOATT vf=FTOS;
  int k;
  for (k=1;k<vi;k++) vf*=FTOS;
  FTOS=vf;
}

/*void x_powi_f()
{
  int vi=ipop();
  FLOATT vf=FTOS;
  int k;
  for (k=1;k<vi;k++) vf*=FTOS;
  FTOS=vf;
}*/

void x_pow_f()
{
  FLOATT v2=fpop(); FTOS=Pow(FTOS,v2);
}

void x_nrt_f()
{
 FLOATT v2=fpop(); FTOS=Nrt(FTOS,v2);
}

void x_neg_f() { FTOS=-FTOS; }
void x_min_f() { FLOATT vf=fpop(); if (vf<FTOS) FTOS=vf; }
void x_max_f() { FLOATT vf=fpop(); if (vf>FTOS) FTOS=vf; }
void x_imm_f() { FLOATT f; GETFLOATPC(f); fpush(f); }

#define compare_ff(rel) FLOATT f2=fpop(); FLOATT f1=fpop(); ipush(f1 rel f2);
void x_less_ff() { compare_ff(<); }
void x_equal_ff()  { compare_ff(==); }
void x_greater_ff()  { compare_ff(>); }
void x_not_less_ff()  { compare_ff(>=); }
void x_not_equal_ff()  { compare_ff(!=); }
void x_not_greater_ff()  { compare_ff(<=); }

#define compare_if(rel) FLOATT f2=fpop(); int i1=ipop(); ipush(i1 rel f2);
void x_less_if() { compare_if(<); }
void x_equal_if() { compare_if(==); }
void x_greater_if()  { compare_if(>); }
void x_not_less_if()  { compare_if(>=); }
void x_not_equal_if()  { compare_if(!=); }
void x_not_greater_if()  { compare_if(<=); }

#define compare_fi(rel) int i2=ipop(); FLOATT f1=fpop(); ipush(f1 rel i2);
void x_less_fi() { compare_fi(<); }
void x_equal_fi() { compare_fi(==); }
void x_greater_fi()  { compare_fi(>); }
void x_not_less_fi()  { compare_fi(>=); }
void x_not_equal_fi()  { compare_fi(!=); }
void x_not_greater_fi()  { compare_fi(<=); }


void x_setf (void) /* Syntax: SETF ARR#[], from%, to%, val# (MB,4.10) */
{
  FLOATT val;
  int from;
  int to;
  int addr;

  val  = fpop ();
  to   = ipop ();
  from = ipop ();
  addr = ipop ();

  if (to < from)
    return;
  if (to >= xdata[addr - 1])
    rterror (FN_SETF, 30);

  for (; from <= to; ++from)
    *((FLOATT FAR*) (&(xdata[addr])) + from) = val;
}

void x_setf3 (void)
{
  fpush (0.0);
  x_setf ();
}

void x_clearf (void)
{
  ipush (0);
  ipush (xdata[ST1 - 1] - 1);
  x_setf3 ();
}


void x_trigmode (void)
{
  switch (ipop ())
  {
  case 1:
    cf = 1.0;
    cf2 = 1.0;
    break;
  case 2:
    cf = 180.0 / PI;
    cf2 = 1.0 / (180.0 / PI);
    break;
  case 3:
    cf = 200.0 / PI;
    cf2 = 1.0 / (200.0 / PI);
    break;
  default:
    rterror (FN_TRIGMODE, 38);
  }
}

/*void x_popf()
{
  int addr;
  addr=ipop();
  fstore(addr,fpop());
}

void x_pushf()
{
  int addr;
  addr=ipop();
  fpush(fdata(addr));
}*/

void x_popif()
{
  int addr;
  GETIPC(addr);
  fstore(addr,fpop());
}

void x_pushif()
{
  int addr;
  GETIPC(addr);
  fpush(fdata(addr));
}
