#include <string.h>

#include "owbasic.h"
#include "comp.h"

#undef tolower

/*
 * LEX.C - lexical analysis for BASIC-Compiler
 * int lex ();
 *   analyses source code starting at char *prog, splits one token and moves
 *   prog to start of next token.
 *   returns an identifier for the current token
 *   also stored in int tok.
 *   Single characters are returned with codes<256
 *   Defined identifiers (keywords) are returned with there key>255
 *   General identifiers return IDENT
 *   Numbers (integer, float) return NUMBER and set lexint or lexfloat to
 *   the corresponding value
 */

uchar intchar='%'; /* suffix for ints */
uchar floatchar='#'; /* suffix for floats */

int lexint;
FLOATT lexfloat;
/*#if FALSE*/
uchar lex_ref = 0;
/*#endif*/

#if FALSE
#define IS_WHITE 1
#define IS_DELIM 2
#define IS_EDELIM 4
#define IS_OPERATOR 8
#define IS_MDIGIT 16
#define IS_HDIGIT 32
#define IS_MALPHA 64
#endif

uchar FAR chartype[]={
  /*0*/ 0|IS_DELIM,/*1*/ 0,/*2*/ 0,/*3*/ 0,
  /*4*/ 0,/*5*/ 0,/*6*/ 0,/*7*/ 0,
  /*8*/ 0,/*9*/ 0|IS_WHITE,/*10*/ 0|IS_DELIM,/*11*/ 0,
  /*12*/ 0,/*13*/ 0|IS_DELIM,/*14*/ 0,/*15*/ 0,
  /*16*/ 0,/*17*/ 0,/*18*/ 0,/*19*/ 0,
  /*20*/ 0,/*21*/ 0,/*22*/ 0,/*23*/ 0,
  /*24*/ 0,/*25*/ 0,/*26*/ 0,/*27*/ 0,
  /*28*/ 0,/*29*/ 0,/*30*/ 0,/*31*/ 0,
  /*32=' '*/ 0|IS_WHITE,/*33='!'*/ 0,/*34='"'*/ 0,/*35='#'*/ 0,
  /*36='$'*/ 0,/*37='%'*/ 0|IS_OPERATOR,/*38='&'*/ 0/*|IS_OPERATOR*/,/*39='''*/ 0,
  /*40='('*/ 0|IS_OPERATOR,/*41=')'*/ 0|IS_OPERATOR,/*42='*'*/ 0|IS_OPERATOR,/*43='+'*/ 0|IS_OPERATOR,
  /*44=','*/ 0|IS_EDELIM,/*45='-'*/ 0|IS_OPERATOR,/*46='.'*/ 0,/*47='/'*/ 0|IS_OPERATOR,
  /*48='0'*/ 0|IS_MDIGIT|IS_HDIGIT,/*49='1'*/ 0|IS_MDIGIT|IS_HDIGIT,/*50='2'*/ 0|IS_MDIGIT|IS_HDIGIT,/*51='3'*/ 0|IS_MDIGIT|IS_HDIGIT,
  /*52='4'*/ 0|IS_MDIGIT|IS_HDIGIT,/*53='5'*/ 0|IS_MDIGIT|IS_HDIGIT,/*54='6'*/ 0|IS_MDIGIT|IS_HDIGIT,/*55='7'*/ 0|IS_MDIGIT|IS_HDIGIT,
  /*56='8'*/ 0|IS_MDIGIT|IS_HDIGIT,/*57='9'*/ 0|IS_MDIGIT|IS_HDIGIT,/*58=':'*/ 0|IS_DELIM,/*59=';'*/ 0|IS_EDELIM,
  /*60='<'*/ 0,/*61='='*/ 0|IS_OPERATOR,/*62='>'*/ 0,/*63='?'*/ 0,
  /*64='@'*/ 0,/*65='A'*/ 0|IS_HDIGIT|IS_MALPHA,/*66='B'*/ 0|IS_HDIGIT|IS_MALPHA,/*67='C'*/ 0|IS_HDIGIT|IS_MALPHA,
  /*68='D'*/ 0|IS_HDIGIT|IS_MALPHA,/*69='E'*/ 0|IS_HDIGIT|IS_MALPHA,/*70='F'*/ 0|IS_HDIGIT|IS_MALPHA,/*71='G'*/ 0|IS_MALPHA,
  /*72='H'*/ 0|IS_MALPHA,/*73='I'*/ 0|IS_MALPHA,/*74='J'*/ 0|IS_MALPHA,/*75='K'*/ 0|IS_MALPHA,
  /*76='L'*/ 0|IS_MALPHA,/*77='M'*/ 0|IS_MALPHA,/*78='N'*/ 0|IS_MALPHA,/*79='O'*/ 0|IS_MALPHA,
  /*80='P'*/ 0|IS_MALPHA,/*81='Q'*/ 0|IS_MALPHA,/*82='R'*/ 0|IS_MALPHA,/*83='S'*/ 0|IS_MALPHA,
  /*84='T'*/ 0|IS_MALPHA,/*85='U'*/ 0|IS_MALPHA,/*86='V'*/ 0|IS_MALPHA,/*87='W'*/ 0|IS_MALPHA,
  /*88='X'*/ 0|IS_MALPHA,/*89='Y'*/ 0|IS_MALPHA,/*90='Z'*/ 0|IS_MALPHA,/*91='['*/ 0|IS_OPERATOR,
  /*92='\'*/ 0,/*93=']'*/ 0|IS_OPERATOR,/*94='^'*/ 0|IS_OPERATOR,/*95='_'*/ 0|IS_MALPHA,
  /*96='`'*/ 0,/*97='a'*/ 0|IS_HDIGIT|IS_MALPHA,/*98='b'*/ 0|IS_HDIGIT|IS_MALPHA,/*99='c'*/ 0|IS_HDIGIT|IS_MALPHA,
  /*100='d'*/ 0|IS_HDIGIT|IS_MALPHA,/*101='e'*/ 0|IS_HDIGIT|IS_MALPHA,/*102='f'*/ 0|IS_HDIGIT|IS_MALPHA,/*103='g'*/ 0|IS_MALPHA,
  /*104='h'*/ 0|IS_MALPHA,/*105='i'*/ 0|IS_MALPHA,/*106='j'*/ 0|IS_MALPHA,/*107='k'*/ 0|IS_MALPHA,
  /*108='l'*/ 0|IS_MALPHA,/*109='m'*/ 0|IS_MALPHA,/*110='n'*/ 0|IS_MALPHA,/*111='o'*/ 0|IS_MALPHA,
  /*112='p'*/ 0|IS_MALPHA,/*113='q'*/ 0|IS_MALPHA,/*114='r'*/ 0|IS_MALPHA,/*115='s'*/ 0|IS_MALPHA,
  /*116='t'*/ 0|IS_MALPHA,/*117='u'*/ 0|IS_MALPHA,/*118='v'*/ 0|IS_MALPHA,/*119='w'*/ 0|IS_MALPHA,
  /*120='x'*/ 0|IS_MALPHA,/*121='y'*/ 0|IS_MALPHA,/*122='z'*/ 0|IS_MALPHA,/*123='{'*/ 0|IS_OPERATOR,
  /*124='|'*/ 0,/*125='}'*/ 0|IS_OPERATOR,/*126='~'*/ 0/*|IS_OPERATOR*/,/*127*/ 0,
#if FALSE
  /*128*/ 0,/*129*/ 0,/*130*/ 0,/*131*/ 0,
  /*132*/ 0,/*133*/ 0,/*134*/ 0,/*135*/ 0,
  /*136*/ 0,/*137*/ 0,/*138*/ 0,/*139*/ 0,
  /*140*/ 0,/*141*/ 0,/*142*/ 0,/*143*/ 0,
  /*144*/ 0,/*145*/ 0,/*146*/ 0,/*147*/ 0,
  /*148*/ 0,/*149*/ 0,/*150*/ 0,/*151*/ 0,
  /*152*/ 0,/*153*/ 0,/*154*/ 0,/*155*/ 0,
  /*156*/ 0,/*157*/ 0,/*158*/ 0,/*159*/ 0,
  /*160*/ 0,/*161*/ 0,/*162*/ 0,/*163*/ 0,
  /*164*/ 0,/*165*/ 0,/*166*/ 0,/*167*/ 0,
  /*168*/ 0,/*169*/ 0,/*170*/ 0,/*171*/ 0,
  /*172*/ 0,/*173*/ 0,/*174*/ 0,/*175*/ 0,
  /*176*/ 0,/*177*/ 0,/*178*/ 0,/*179*/ 0,
  /*180*/ 0,/*181*/ 0,/*182*/ 0,/*183*/ 0,
  /*184*/ 0,/*185*/ 0,/*186*/ 0,/*187*/ 0,
  /*188*/ 0,/*189*/ 0,/*190*/ 0,/*191*/ 0,
  /*192*/ 0,/*193*/ 0,/*194*/ 0,/*195*/ 0,
  /*196*/ 0,/*197*/ 0,/*198*/ 0,/*199*/ 0,
  /*200*/ 0,/*201*/ 0,/*202*/ 0,/*203*/ 0,
  /*204*/ 0,/*205*/ 0,/*206*/ 0,/*207*/ 0,
  /*208*/ 0,/*209*/ 0,/*210*/ 0,/*211*/ 0,
  /*212*/ 0,/*213*/ 0,/*214*/ 0,/*215*/ 0,
  /*216*/ 0,/*217*/ 0,/*218*/ 0,/*219*/ 0,
  /*220*/ 0,/*221*/ 0,/*222*/ 0,/*223*/ 0,
  /*224*/ 0,/*225*/ 0,/*226*/ 0,/*227*/ 0,
  /*228*/ 0,/*229*/ 0,/*230*/ 0,/*231*/ 0,
  /*232*/ 0,/*233*/ 0,/*234*/ 0,/*235*/ 0,
  /*236*/ 0,/*237*/ 0,/*238*/ 0,/*239*/ 0,
  /*240*/ 0,/*241*/ 0,/*242*/ 0,/*243*/ 0,
  /*244*/ 0,/*245*/ 0,/*246*/ 0,/*247*/ 0,
  /*248*/ 0,/*249*/ 0,/*250*/ 0,/*251*/ 0,
  /*252*/ 0,/*253*/ 0,/*254*/ 0,/*255*/ 0,
#endif
};

/*int isdelim (char c)
{
  return (chartype[c] & IS_DELIM);
}*/
#if FALSE
#define isdelim(c)    (chartype[c] & IS_DELIM)
/*  int iswhite(char c); */
#define iswhite(c)    (chartype[c] & IS_WHITE)
/*  int isedelim(char c); */
#define isedelim(c)   (chartype[c] & IS_EDELIM)
/*  int isoperator(char c); */
#define isoperator(c) (chartype[c] & IS_OPERATOR)
/*int ismalpha (char c)
{
  return (chartype[c] & IS_MALPHA);
}*/
#define ismalpha(c)   (chartype[c] & IS_MALPHA)
/*int ismdigit(char c);*/
#define ismdigit(c)   (chartype[c] & IS_MDIGIT)
/*  int ishdigit(char c); */
#define ishdigit(c)   (chartype[c] & IS_HDIGIT)
#endif

int source_line_old;

void progeol (void)
{
  while (!isdelim (LEX_CHAR))
    LEX_INCCHAR ();
}

void skipws (void)
{
  while (iswhite (LEX_CHAR))
    LEX_INCCHAR ();
}

static void NEAR progsw (void)
{
  /* skip white space at end of token*/
  token_end = prog; /* if whitespace at end of token -> ignore*/
  while (iswhite (LEX_CHAR))
    LEX_INCCHAR ();
}

#ifndef FAST_MACRO
static void NEAR progpp (void)
{
  /* to next char skipping white space */
  LEX_INCCHAR ();
  progsw ();
}
#else
#define progpp()  LEX_INCCHAR (); progsw ()
#endif

type_t suffix_type (void)
{
  uchar name[IDENTLEN];
  uchar* nptr = name;
  datatype_t type = {DTYPEDEF, 0};
  register char c = LEX_CHAR;

  datatype.ext = 0;
  if (c == '$')
    return (DSTRING);
  if (c == '!')
    return (DBOOLEAN);
  if (c == '&')
    return (DCHAR);
  if (c == intchar)
    return (DINTEGER);
  if (c == floatchar)
    return (DFLOAT);
  if (c == '\\')
  { /* extended datatype */
    LEX_INCCHAR ();
    if (!ismalpha (LEX_CHAR))
      serror (1000 + IDENT);
    while (ismalpha (LEX_CHAR) || ismdigit (LEX_CHAR))
      *(nptr++) = LEX_CHARPP ();
    *nptr = 0x00;

    if ((datatype.ext = searchentry_name (&type, name)) < 0)
      serror (16);
    return (xdata[datatype.ext - 2]);
  }
  return (DUNKNOWN);
}

int lex1 (void)
{
  uchar hchar;
  datatype_t ctype;
  uchar *hptr;
  uchar cchar;
  int addr;
  uchar name[NAMELEN];

reparse:
  datatype.ext = 0;
  tok = 0;
  lex_ref = 0;

  progsw ();

  token_start = prog;
  source_line_old = ITOS.line;
  token_type = TIDENT;
  cchar = LEX_CHAR;

  if (cchar > 0x7F) /* code >127 */
    serror (0);

  if (cchar == 0)
  {
    token_type = TDELIM;
    token_end = prog;
    return tok = END;
  }

  if (cchar == '!') /* Comment */
  {
    while ((LEX_CHAR != 0x0A) && (LEX_CHAR != 0x0D) && (LEX_CHAR != 0))
      LEX_INCCHAR ();
    ++ITOS.line;
    token_type = TKEY;
    if (LEX_CHAR != 0)
      LEX_INCCHAR ();
    token_end = prog;
    return (tok = REM);
  }

  if (cchar == '?')    /* print */
  {
    progpp ();
    token_type = TKEY;
    return tok = PRINT;
  }

  if (isdelim (cchar)) /* delimiter */
  {
    if ((cchar == 0x0A) || (cchar == 0x0D))
      ++ITOS.line;
    progpp ();
    token_type = TDELIM;
    return (tok = DELIMITER);
  }

#if FALSE
if (cchar == '&') /* pointer operators (MB,4.20) */
  {
    if (*(prog + 1) == '&')
    {
      prog += 2;
      skipws ();
      token_type = TOPERATOR;
      return (tok = OP_PTRADDR); /* && : pointer -> pointer (to pointer) */
    }
    else if (*(prog + 1) == '~')
    {
      prog += 2;
      skipws ();
      token_type = TOPERATOR;
      return (tok = OP_PTRARRAYADDR); /* &~ : pointer array -> pointer (array indexing) */
    }
  }
#endif

  if (isoperator (cchar)) /* single character operator */
  {
    if ((cchar == '/') && (*(prog + 1) == '*')) /* multiline comments (MB,4.40) */
    {
      prog += 2;
      while ((LEX_CHAR != '*') || (*(prog + 1) != '/'))
      {
        if ((LEX_CHAR == 0x0A) || (LEX_CHAR == 0x0D))
          ++ITOS.line;
        if (LEX_CHAR == 0)
          serror (0);
        LEX_INCCHAR ();
      }
      prog += 2;
      goto reparse;
    }
    tok = LEX_CHAR;
    progpp ();
    token_type = TOPERATOR;
    return (tok);
  }

  if (cchar == '<') /* comparism */
  {
    LEX_INCCHAR ();
    if (LEX_CHAR == '=')
    {
      progpp ();
      return (tok = NOT_GREATER);
    }
    if (LEX_CHAR == '>')
    {
      progpp ();
      return (tok = NOT_EQUAL);
    }
    return (tok = LESS);
  }

  if (cchar == '>')
  {
    LEX_INCCHAR ();
    if (LEX_CHAR == '=')
    {
      progpp ();
      return (tok = NOT_LESS);
    }
    progsw ();
    return (tok = GREATER);
  }

  if (isedelim (cchar)) /* delimiter of expressions , ; */
  {
    tok = LEX_CHAR;
    progpp ();
    token_type = TEDELIM;
    return (tok);
  }

  if (cchar == '"')    /* string constants */
  {
    LEX_INCCHAR ();
    while ((LEX_CHAR != '"') &&
     (LEX_CHAR != NEWLINE) &&
     (LEX_CHAR != 0))
    {
      if (LEX_CHAR == '\\')
      {
        LEX_INCCHAR ();
        if ((LEX_CHAR != NEWLINE) && (LEX_CHAR != 0))
          LEX_INCCHAR ();
      }
      else
        LEX_INCCHAR ();
    }

    if (LEX_CHAR != '"')
      serror (1);
    progpp ();
    token_type = TSTRING;
    datatype.type = DSTRING;
    return (tok = STRING);
  }

  if (cchar == '\'')    /* char constants */
  {
    lexint = tok = LEX_PPCHAR ();
    LEX_INCCHAR ();
    if (tok == NEWLINESYMBOL)
      tok = NEWLINE;
    if (tok == '\\')
      tok = LEX_CHARPP ();
    if (LEX_CHAR != '\'')
      serror (1);
    progpp ();
    token_type = TCHAR;
    datatype.type = DCHAR;
    return (tok);
  }

  if (ismdigit (cchar)) /* numbers */
  {
    datatype.type = DINTEGER;
    if (cchar == '0')
    {
      if (LEX_PPCHAR () != 'x')
        goto nohex;
      LEX_INCCHAR ();
      while (ishdigit (LEX_CHAR))
        LEX_INCCHAR ();
      hchar = LEX_CHAR; LEX_CHAR = 0;
      sscanf ((char*) token_start, "%x", &lexint);
      LEX_CHAR = hchar;
      lexfloat = lexint;
    }
    else
    {
    nohex:
      while ((ismdigit (LEX_CHAR)) ||
       (LEX_CHAR == '.') ||
       (LEX_CHAR == 'e') || (LEX_CHAR == 'E')
       )
      {
        if ((LEX_CHAR == '.') || (LEX_CHAR == 'e') || (LEX_CHAR == 'E'))
          datatype.type = DFLOAT;

        LEX_INCCHAR ();
      }

      hchar = LEX_CHAR;
      LEX_CHAR = 0;
      switch (datatype.type)
      {
      case DINTEGER:
        sscanf ((char*) token_start, "%d", &lexint); /* convert integer constant */
        lexfloat = lexint; /* int const can be used as float const too */
        break;
      case DFLOAT:
        sscanf ((char*) token_start, FFORMATI, &lexfloat); /* convert to float */
        break;
      }
      LEX_CHAR = hchar;
    }
    token_type = TNUMBER;
    progsw ();
    return (tok = NUMBER);
  }

  if (cchar == '$') /* labels */
  {
    LEX_INCCHAR ();
    while (ismalpha(LEX_CHAR)||
     ismdigit(LEX_CHAR)
     ) LEX_INCCHAR ();
    token_type = TLABEL;
    datatype.type = DUNKNOWN;
    progsw ();
    return (tok = LABEL);
  }

  if (cchar == '#') /* constants */
  {
    LEX_INCCHAR ();
    hptr = name;
    while (ismalpha (LEX_CHAR)
        || ismdigit (LEX_CHAR))
      *(hptr++) = LEX_CHARPP ();
    *hptr = 0;
    progsw ();
    ctype.type = DCONSTANT;
    ctype.ext = 0;
    if ((addr = searchentry_name (&ctype, name)) < 0)
      serror1 (2020, name);
    lexfloat = lexint = xdata[addr];

    datatype.type = xdata[addr + 1];
    datatype.ext = xdata[addr + 2];
    token_type = TNUMBER;
    return (tok = NUMBER);
  }


  if (ismalpha (cchar) || (cchar == '&') || (cchar == '~')) /* identifiers */
  {
/*#if FALSE*/
    /*   __________________________________________
        | prefix | description             | value |
        ||||
        | (-)    | variable                | 0     |
        | &      | address                 | 1     |
        | ~      | pointer array values    | 2     |
        | &~     | pointer array addresses | 3     |
        | &&     | pointer address         | 4     |
         
    */
    if (cchar == '&') /* pointer treatment suffixes recognized in lex (MB,4.20) */
    {
      lex_ref = 1; /* & suffix; returns address of variable */
      LEX_INCCHAR ();
      if (LEX_CHAR == '&')
      {
        lex_ref = 4; /* && suffix; returns address of pointer itself */
        LEX_INCCHAR ();
      }
      else if (LEX_CHAR == '~')
      {
        lex_ref = 3; /* &~ suffix; access to pointer array addresses */
        LEX_INCCHAR ();
      }
    }
    else if (cchar == '~')
    {
      lex_ref = 2; /* ~ suffix; access to pointer array values */
      LEX_INCCHAR ();
    }
/*#endif*/
    token_start = prog;

    while (ismalpha (LEX_CHAR)
        || ismdigit (LEX_CHAR))
      LEX_INCCHAR ();
    if (token_start == prog)
      serror (0);

    hptr = prog;

    if ((datatype.type = suffix_type ()) != DUNKNOWN)
      if (datatype.ext == 0)
        LEX_INCCHAR ();

    progsw ();
    token_end = hptr;
    token_type = TIDENT;

    tok = look_up (keytable); /* keyword ? */

    if (tok != UNKNOWN)
      token_type = TKEY;
    else
    {
      tok = IDENT;
#if FALSE
      ctype.type = DMACRO;
      ctype.ext = 0;
      if ((addr = searchentry_token (&ctype)) >= 0) /* macro existing */
      {
        /* TODO: insert macro */
      }
#endif
    }

    return (tok);
  }

  serror (0);
  return 0;
}

#ifdef DEBUGLEX
int lex(void)
{
  int rc;
  rc=lex1 ();
  /*LibPutMsgDlg("Token: >>%s<<\r tok:%d", token,tok);*/
  /* if (switches.printwhilecomp)*/
    {
      while (aprog<prog)
  OutTrm(*(aprog++));
      DisplayTrm();
    }

  return rc;
}
#endif

void expect (int k)
{
  if (lex () != k)
    serror (1000 + k);
}


void expect_label (void)
{
  tok = lex (); /* read destination */
  if ((tok != NUMBER) && (tok != IDENT) && (tok != LABEL))
    serror (37);
}

int look_up (Keyword FAR *list)
{
  int i;
  uchar name[IDENTLEN];

  tokcpy_ident (name);
  for (i = 0; list[i].command != NULL; ++i)
    if (strcmp_nf (name, list[i].command) == 0)
      return (list[i].tok);
  return (UNKNOWN);
}

void putback (void)
{
  prog = token_start;
  switch (lex_ref)
  {
  case 3:
  case 4:
    --prog;
  case 1:
  case 2:
    --prog;
  }
  ITOS.line = source_line_old;
}

void tokncpy (uchar* dest, int n)
{
  uchar* sptr = token_start;
  uchar* dptr = dest;

  while ((sptr != token_end) && (dptr - dest < n - 1))
    *(dptr++) = tolower (*(sptr++));
  *dptr = 0;
}

void tokcpy_ident (uchar* dest)
{
  tokncpy (dest, IDENTLEN);
}

void sconstncpy (uchar *dest, int n)
{
  uchar* sptr = token_start + 1;
  uchar* dptr = dest;

  while ((sptr != token_end - 1) && (dptr - dest < n - 1))
    *(dptr++) = *(sptr++);
  *dptr = 0;
}
