/*
 * Mathematik fr OW-Basic.
 * (c) Bernd Huebler, Wolfgang Ortmann
 */
#ifdef LSI_C
#include <stdrom.h>
#endif
#ifdef __HITACHI__
#include <hitachi.h>
#endif
#include <math.h>
#include "owbasic.h"

#ifdef LSI_C
#ifndef DUMMY_H
#define DUMMY_H
char _osmajor='1';
void *sbrk(size_t s) { return(NULL); };
#endif
#endif

FLOATT cf =  1.0; /* 1.0=rad,     180.0/PI=degree,    (200.0/PI = grad */
FLOATT cf2 = 1.0; /* 1.0=rad, 1/(180.0/PI)=degree, 1/(200.0/PI) = grad (MB,4.10) */

#ifndef FAST_MATH

#define LN10 (/*log(10)*/ 1)
#define LN2 (/*log(2)*/ 0.30102999566398119521373889472449)
#define NAN (1.0e301)

#ifndef NEVER_DEFINED
#define TORAD(value) ((value) * cf2)
#define FROMRAD(value) ((value) * cf)
#else
double torad(FLOATT v) { return (v * cf2); }
double fromrad(FLOATT v) {return (v * cf); }
#define TORAD(value) torad(value)
#define FROMRAD(value) fromrad(value)
#endif

#endif

#ifndef FAST_MATH
FLOATT Sin(FLOATT value)
{
   return sin(TORAD(value));
}

FLOATT Cos(FLOATT value)
{
   return cos(TORAD(value));
}

FLOATT Tan(FLOATT value)
{
   return tan(TORAD(value));
}

FLOATT Log(FLOATT value)
{
   return log(value)/LN10;
}

FLOATT Log2(FLOATT value)
{
  return  log(value)/LN2;
}

FLOATT Ln(FLOATT value)
{
  return log(value);
}

FLOATT Sqrt(FLOATT value)
{
  return sqrt(value);
}

FLOATT ASin(FLOATT value)
{
  return FROMRAD(asin(value));
}

FLOATT ACos(FLOATT value)
{
  return FROMRAD(acos(value));
}

FLOATT ATan(FLOATT value)
{
  return FROMRAD(atan(value));
}

FLOATT ATan2(FLOATT value1,FLOATT value2)
{
  return FROMRAD(atan2(value1,value2));
}

FLOATT Sinh(FLOATT value)
{
  return sinh(value);
}

FLOATT Cosh(FLOATT value)
{
  return cosh(value);
}

FLOATT Tanh(FLOATT value)
{
  return tanh(value);
}

FLOATT ASinh(FLOATT value)
{
  return log(value+sqrt(value*value+1));
}

FLOATT ACosh(FLOATT value)
{
  return log((value>.0)?value+sqrt(value*value-1):value-sqrt(value*value-1));
}

FLOATT ATanh(FLOATT value)
{
  return .5*log((1+value)/(1-value));
}

FLOATT ALog(FLOATT value)
{
  return exp(value * LN10);
}

FLOATT ALog2(FLOATT value)
{
  return exp(value * LN2);
}

FLOATT ALn(FLOATT value)
{
  return exp(value);
}

FLOATT Sqr(FLOATT value)
{
  return value*value;
}

FLOATT Pow(FLOATT value, FLOATT val)
{
  return pow(value,val);
}

FLOATT Nrt(FLOATT val, FLOATT value)
{
  return pow(val,1.0/value);
}

/*
 * Koordinatenwandlungen
 */

FLOATT XCart(FLOATT r,FLOATT fi)
{
  /* x=r*cos(fi) */
  return cos(TORAD(fi))*r;
}

FLOATT YCart(FLOATT r,FLOATT fi)
{
  /* y=r*sin(w) */
  return sin(TORAD(fi))*r;
}

FLOATT Radius(FLOATT x,FLOATT y)
{
  return Sqrt(x*x+y*y);
}

FLOATT Phi(FLOATT x,FLOATT y)
{
  return ATan2(y,x);
}

/* Conversion from current mode (rad/degree/grad) to specified mode */

FLOATT Deg(FLOATT value)
{
  return value*cf2*(180/PI);
}

FLOATT Rad(FLOATT value)
{
  return TORAD(value);
}

FLOATT Grad(FLOATT value)
{
  return value*cf2*(200/PI);
}
#else

#include "mymath.h"

#endif
