#include "modeldef.h"
#include <shared.h>
#include "owbasic.h"
#include "cstack.h"
#include "comp.h"
#include "rc.h"
#include "graphic.h"
#include "modecall.h"
#include "inter.h"
#include "file.h"
#include "cache.h"
#ifndef __HITACHI__
 #include "gui.h"
#endif

static void NEAR welcome (void);

Pref pref;
Switches switches;

PC pc; /* code pointer for compiler and interpreter */
PC lastpc;

bool isAutorun;

bool isFindRTE;
bool overlays_used = FALSE;
byte declaration_required = 0;

#if FALSE
/* pointers to xdata - array for user variables */
int tchxi = 0; int tchyi = 1;
int tchobji = 2; int tchstsi = 3; /* values used before start of compiler */
int pvmodeli = 4; int owbversioni = 5;
int FAR *tchxp = &(xdata[0]);
int FAR *tchyp = &(xdata[1]);
int FAR *tchobjp = &(xdata[2]);
int FAR *tchstsp = &(xdata[3]);
int FAR *pvmodelp = &(xdata[4]);
int FAR *owbversionp = &(xdata[5]);
#endif

int error, inputrc;
/* binary code */
codet bin[PROG_SIZE];

/* the one and only filebuffer */
FILE_BUF fb;
/* name of program = name of main memo */
uchar progname[NAMELEN];
/*uchar recallprogname[NAMELEN];*/
int   progcat;


int pvmodel;

/* how we were called */
word mymode;
word mystatus;

/* how we return */
byte oret = O_NORMAL;

void Restart (byte status)
{
  StopTimer ();
  LibModeJump (mymode, status);
  LibJumpMenu (); /* if modejump fails - should never happen */
}

static byte FAR prefs_default[] = {
  5,      /* progsub */
  5,      /* includesub */
  FALSE,  /* statistics */
  M_FILE, /* startmenu */
  1,      /* finalwait */
  TRUE,   /* askabort */
  TRUE,   /* clearoverlay */
  FALSE,  /* autorun */
  FALSE,  /* restart */
  FALSE,  /* caching */
  FALSE,  /* searcherror */
  TRUE,   /* call_cache */
};

void default_prefs (void)
{
  int i = 0;

  pref.ver = SETTINGSVER;
  for (; i <= PREF_SETTINGS; ++i)
    ((byte*) (&pref.progsub))[i] = prefs_default[i];

  strcpy ((char*) pref.autorun_name, "autorun");
}

#if FALSE
static void NEAR welcome (void)
{
  uchar versions_string[10];
  /* welcome screen */
  /*LibClrDisp ();*/
  LibPutBoxSub (5, 135, 83, 155, (byte) 1);
  /*sprintf ((char*) versions_string, "%d", OWBVersion);*/
  sprintf ((char*) versions_string, "%0.2lf", (double) (OWBVersion / 100.0));
  LibStringDsp ((byte*) PROGNAME,        (word) 10, (word) 140, (word) 125, LFONT ((byte) 3));
  LibStringDsp ((byte*) versions_string, (word) 60, (word) 146, (word) 125, LFONT ((byte) 1));
  LibPutDisp ();
}
#endif

static void NEAR Select (void)
{
#ifndef NEVER_DEFINED
  mainmenu (pref.startmenu);

#else
  int menu = pref.startmenu;

  while (menu != M_READY)
    switch (menu)
    {
    case M_FILE:
      menu = filemenu ();
      break;
    case M_PROGRAM:
      menu = programmenu ();
      break;
    case M_PREFS:
      menu = prefsmenu ();
      if (menu == 0)
        menu = pref.startmenu;
      break;
    default:
      menu = M_FILE; /* repair wrong assignments */
      break;
    }
#endif
}


void ClearOverlays (void)
{
  sffindfile (&fb);
  recordname[15] = 0;
  finf.fp = 0xFFFF;

  while (LibFileFindNext (&fb, &finf, 0x00))
  {
    LibFileReadEx (&fb, &finf, FILE_BLOCK_SIZE);
    FBB[15] = 0;
    if (strcmp ((char*) (FBB), (char*) recordname) == 0)
    {
      LibFileRemove (&fb, &finf);
      finf.fp = 0xFFFF;
    }
  }
}

void main (void)
{
  int /*i,*/ j, callstat;
#ifndef SMALL_NO_BUILTIN
  int k;
#endif
  int submode;
#ifdef LSI_C
  word segm, offs;
#endif
#ifdef __HITACHI__
  void *seg;
#endif
  byte *hptr;
  /*codet* cptr;*/

  InitTrm ();

#ifdef LSI_C
  LibGetMode (&mymode, &mystatus, &segm, &offs); /* how were we called */
#endif
#ifdef __HITACHI__
  LibGetMode (&mymode, &mystatus, &seg);         /* how were we called */
#endif

#ifdef LSI_C
  pvmodel = getmodel ();
#endif
#ifdef __HITACHI__
  pvmodel = 5;
#endif

  /*recallprogname[0] = 0;*/

  /*switches.recall = 0;*/
  switches.last = PROG_SIZE - 1 /* 10 */;
  switches.remote = FALSE;
  switches.calllevel = 0;
  switches.quiet = FALSE;
  switches.cached = FALSE;
  switches.rterror = FALSE;

  callstat = mystatus & STATUS_CALLMASK;
  if (mc_isCall ())
  {
    switches.calllevel = mc_CallLevel ();
    if (callstat == STATUS_QUIET)
    {
      switches.calllevel = 0;
      switches.quiet = TRUE; /* do not show compiler messages */
    }
  }

#if FALSE
  if ((mystatus & ~STATUS_NO_AUTORUN) == STATUS_NORMAL) /* normal mode (called from menu) */
    welcome ();  /* welcome screen */
#endif

#ifndef SMALL_NO_BUILTIN
  if (mystatus==STATUS_BUILTIN) {
    /* initialize program, variables and constants */
    for (i=0; i<initcodesize; i++) bin[i]=ibin[i];
    i=0; j=0;
    while (i<datanext)
      {
  /*xdata[i++]=xdata_var[j++];*/
  xdata[i++] = XDATA_VAR(j);
  ++j;
  for (k=0;k</*xdata_var[j]*/ XDATA_VAR(j);k++,i++) xdata[i]=0;
  j++;
      }
    for (j=0,i=firstconst; i<DATASIZE; i++,j++) xdata[i]=xdata_const[j];
  }
#endif

#ifdef LSI_C
  /* set int vector for division by zero */
  setintvecdiv0 ();
#endif

  InitTimer ();
#ifdef __HITACHI__
  InitBreakTimer ();
#endif

  /* load setting and initialise switches and variables */

  loadsettings ();

  switches.statistics = pref.statistics;
  switches.progsub = pref.progsub;
  switches.includesub = pref.includesub;
  switches.caching = pref.caching; /* caching allowed? */
  switches.savetocache = (callstat == STATUS_CACHE);

  isAutorun = FALSE;

  IncStackPtr = IncStack;
  ITOS.min = 0;
  ITOS.progtype = MMAIN;

  if (isFindRTE = (callstat == STATUS_FIND_RTE))
  {
    switches.caching = FALSE;
    mystatus &= ~STATUS_CALLMASK;
    /* find runtime error */
    FBBNUM = 3072;
    loadshared ((byte*) "rterror", (byte*) PROGNAME, FBB, 3072);
    fbbinit ();
    fbbstring (progname);
    fbbint (&progcat);
    fbbint (&(switches.last));
    strcpy ((char*) ITOS.fname, (char*) progname);
  }

  if (mystatus == STATUS_RC)
    switches.remote = (byte) TRUE; /* remote mode */

#ifndef SMALL_NO_BUILTIN
  if (mystatus == STATUS_BUILTIN)
    goto start;
#endif

  /* selection of program */

  prog = FB;

#ifdef REMOTE_CONTROL
  if (switches.remote) /* are we in remote control mode? */
  {
    switches.caching = switches.statistics = FALSE;
    /* selection from pc */
    if (!rc_selprog ())
    {
      /* verify abort */
      if (verify_abort ())
        LibJumpMenu (); /* Aborted */
      Restart (STATUS_RC);
    }
  }
  else
#endif
    if (mc_isCall ())
    {
      if ((callstat == STATUS_IEDIT) && !pref.call_cache) /* called from IEdit */
        switches.caching = FALSE; /* do not search for cached program */
      switches.statistics = FALSE;

      if (!mc_Init ())
        goto end;
      mc_ReadParam (&submode, ITOS.fname,  /* category and name */
             &j);              /* dummy parameters (cursor) */

      if (mystatus & STATUS_PARAM)
      {
        mc_ReadString (stringstack1);
        /*mc_Skip ();
        mc_ReadInt (&j);
        if (j & 0x01)
          mc_ReadString (stringstack1);
        if (j & (0x02 | 0x04))
        {
          switches.recall = j;
          if (j & 0x02)
            mc_ReadString (recallprogname);
          mc_ReadInt (&switches.binstart);
        }*/
      }

      if (submode < 0)
        ITOS.min = 0;     /* not given, use default */
      else
        ITOS.min = submode;

      if ((ITOS.fname[0] == '!') && (ITOS.fname[1] == ' ')) /* skip "! " prefix */
        strcpy ((char*) ITOS.fname, (char*) &(ITOS.fname[2]));
    }
    else
    {
      /* autorun or find runtime error or interactive selection mode */
      if (!isFindRTE)
      {
        if (pref.autorun && (mystatus != STATUS_NO_AUTORUN))
        {
          isAutorun = TRUE;
          ITOS.min = switches.progsub;
          strcpy ((char*) progname, (char*) pref.autorun_name);
          strcpy ((char*) ITOS.fname, (char*) progname);
          if (!loadprog ())
            Select ();
        }
        else
          Select ();
      }
    }
  progcat = ITOS.min;
  strcpy ((char*) progname, (char*) ITOS.fname);


  if (!loadprog ())
  {
#ifdef REMOTE_CONTROL
    if (switches.remote)
    {
      rc_finish (FALSE, (uchar*) Near (msg_cantload));
      Restart (STATUS_RC);
    }
#endif
    msgbox1 ((byte*) Near (msg_cantload));
    goto end;
  }

  if (!switches.cached)
    /* compile process */
    if (compile (switches.last) != 0)
    {
      /* compiler error */
      oret = O_COMPILE_ERROR;
      goto end;
    }

  /*if (switches.recall = 0x04)
    bin_start.c = bin + switches.binstart;*/

  /*DEBUG*/
  /*for (cptr = bin; cptr < pc.c; ++cptr)
    OutputITrm (*cptr);

  x_waittouch ();*/

#ifndef SMALL_NO_CACHE
  if (switches.savetocache)
  {
    cache_write ();
    overlays_used = FALSE;
  }
#endif


#ifdef REMOTE_CONTROL
  if (SendBinary)
    rc_send_binary ();
#endif

  if (switches.statistics) /* print statistics */
  {
    sprintf ((char*) stringstack2, Near (msg_stattext),
      pc.c - bin,                       PROG_SIZE,
      listnext,                         DATASIZE,
      datanext + DATASIZE - firstconst, DATASIZE,
      /*DATASIZE - firstconst, DATASIZE,*/
      bin_start.c - bin,
      last_proc + 1);
    msgbox (stringstack2, msg_statistics, 0.0, MBX_OK);
  }

  /* prepare program execution */
#ifndef SMALL_NO_BUILTIN
 start:
#endif
  LibTchStackClr ();
  LibTchStackPush (NULL);
  LibTchStackPush (TchAllDsp);
  LibTchStackPush (TchHardIcon);
#ifdef LSI_C
  InitTimer ();
#endif
  /* execute the program */

  if (!Exec (bin_start))
    oret = O_RUNTIME_ERROR;

end:
  if (pref.clearoverlay && overlays_used)
    ClearOverlays ();

#ifdef REMOTE_CONTROL
  if (switches.remote)
  {
    if (oret == O_NORMAL)
      rc_finish (TRUE, (uchar*) "");
    Restart (STATUS_RC);
  }
#endif

  /* normal call */
  if (oret == O_RUNTIME_ERROR)
    if (switches.rterror)
      Restart (mystatus);

  if (switches.calllevel > 0)
  { /* return to caller */
    if ((oret == O_COMPILE_ERROR)
     || ((oret == O_RUNTIME_ERROR) && callstat && switches.rterror))
    { /* prepare filename (memoname) of error */
      hptr = prog_start;
      while ((*hptr != 0x0D) && (*hptr != 0))
        ++hptr;
      *hptr = 0;

      mc_Init (); mc_Skip (); mc_Skip ();
      mc_WriteParam (progcat, prog_start, prog - prog_start);
    }
    StopTimer ();
    LibModeJump (mc_CallerMode (), MC_RETURNVALUE + switches.calllevel - 1);
  }

  /* finish OWBasic */
  if (pref.restart)
    Restart (STATUS_NORMAL); /* restart */

  StopTimer ();
  LibJumpMenu ();
}
