/******************************************************************************
*	Based on an idea from Alessandro Dorigatti
*       Sound for OWBasic by Wolfgang Ortmann
******************************************************************************/
#include "owbasic.h"
#include "sound.h"

/* Functions definition  -------------------------------------------*/

/* Play a SOUND: freq  :  frequency ( or phase shift per cycle )   *
 * ------------  dur   :  duration  ( nearly time in milli seconds *
 *               vol   :  waveform/volume (volume from 1 to 100)   */
#ifdef LSI_C
/*#define fullcycle 0x100000*/
#define fullcycle 221000

int sound (int freq, unsigned int dur, unsigned int vol)
{
  unsigned long tl;   /* Total note time (in counter cycles)*/
  unsigned long durl; /* duration (in counter cycles)*/
  unsigned long cycl; /* Internal waveform counter */
  unsigned long voll; /* length of positive phase (in counter cycles)*/
  byte quiet;

  /* Valid duration range: [1..10000] */
  if (dur > 10000 || dur < 1)
    return (38);
  /* Valid volume range: [1..100] */
  if (vol > 100 || vol < 1)
    return (38);
  /* Valid frequency range: [1..30000] */
  if (freq > 30000 || freq < 0)
    return (38);

  if (quiet = freq == 0)
    freq = 10000;
  durl = dur * (unsigned long) 221;
  cycl = 0;
  voll = fullcycle / (unsigned long) 200 * vol;
  if (!quiet)
    outpsound (0x84);
  tl = 0;
  while (tl < durl)
  {
    while (cycl < voll)
    {
      ++tl;
      cycl += freq;
    }

    outpsound(0x00);
    while (cycl < fullcycle)
    {
      ++tl;
      cycl += freq;
    }

    if (!quiet)
      outpsound (0x84);
    cycl -= fullcycle;
  }

  outpsound (0x00);
  return 0;
}
#endif

#ifdef __HITACHI__
word oldPECR;               /* Saves old value of control register */

/* Port E Control Register */
#define PECR ((word*)0xA4000108)
/* Port E Data Register */
#define PEDR ((byte*)0xA4000128)

/* Init Buzzer... */
void initBuzzer() {
  oldPECR=*PECR; /* Save value of Port E Control Register */
  *PECR=(oldPECR & 0x0FFF) | 0x5000;  /* Set PTE7 and PTE6 to output-mode */
}

void setBuzzer(char v);

/* release Buzzer... */
void releaseBuzzer() {
  setBuzzer(0);   /* Turn off the Buzzer */
  *PECR=oldPECR;  /* Restore Port E Control Register */
}

/*
 * Buzzer is connected to PTE6 and PTE7
 * Values: -1, 0 or 1
 */

void setBuzzer(char v)
{
  byte value=(*PEDR) & 63;
  switch(v) {
  case -1: value=value | 64;     break;
  case 1:  value=value | 128;    break;
  default:  break;
  }
  *PEDR=value;
}

#define fullcycle 57540

#define counter (*(int*)0xFFFFFEB0 & 0xFFFF)

int sound(int freq, unsigned int dur, unsigned int vol)
{
  unsigned long tl;   /* Total note time (in counter cycles)*/
  unsigned long durl; /* duration (in counter cycles)*/
  unsigned long cycl; /* Internal waveform counter */
  unsigned long voll; /* length of positive phase (in counter cycles)*/

  int lt;

  byte quiet;

  /* Valid duration range: [1..10000] */
  if (dur>10000 || dur<1) return 38;
  /* Valid volume range: [1..100] */
  if (vol>100 || vol<1) return 38;
  /* Valid frequency range: [1..3000] */
  if ((freq!=0)&&(freq>30000 || freq<1)) return 38;

  quiet=(freq==0);
  if (quiet) freq=10000;
  durl = dur*(unsigned long)58;
  cycl = 0;
  voll = fullcycle/(unsigned long)200*vol;
  initBuzzer();
  tl=0;
  lt=counter;
  while (tl<durl)
    {
      while (cycl<voll)
  {
    while (lt==counter);
    while (lt!=counter)
      {
        lt=(lt-1) & 0xffff;
        tl++;
        cycl=cycl+freq;
      }
  }

      if (!quiet) setBuzzer(-1);

      while (cycl<fullcycle)
  {
    while (lt==counter);
    while (lt!=counter)
      {
        lt=(lt-1) & 0xffff;
        tl++;
        cycl=cycl+freq;
      }
  }

      if (!quiet) setBuzzer(1);
      cycl=cycl-fullcycle;
    }

  releaseBuzzer(0);
  return 0;
}

#endif

/* Play a string: *n  :  string like "a4+#a2d"                     *
 * -----------    dur :  duration                                  *
 *                vol : volume/waveform                            */

char FAR nval[8]={9,10,0,2,4,5,7,11};
int FAR nfreq[]=
{
   500, 529, 556, 594, 630, 667, 707, 749, 793, 841, 891, 944,
  1000,1059,1122,1189,1260,1335,1414,1498,1587,1682,1782,1888,
  2000,2119,2245,2387,2520,2670,2828,2997,3175,3364,3564,3775,
  4000,4238,4490,4757,5040,5339,5657,5993,6350,6727,7127,7551,
  8000,8476,8980,9514,10079,10679,11314,11986,12699,13454,14254,15102,
};

static int octave;  /* current octave */
static int ndur;    /* note duration */

/* 20 millisecond pause between notes */
#define TPAUSE 20

int play(uchar *n, unsigned int dur, unsigned int vol)
{
  uchar *p;     /* current character */
  uchar c;
  byte is;     /* cis, dis .. */
  int note;

  is = FALSE;
  octave = 2;
  ndur = dur / 4; /* start value 1/4 note */

  p = n;
  while (*p)
  {
    c = *(p++);
    c = tolower (c);	 /* convert letters to lower case */

    if ((c >= 'a') && (c <= 'h')) /* play note */
    {
      note = octave * 12 + nval[c - 'a'];
      if (is)
        ++note;
#if TPAUSE>0
      sound (nfreq[note], ndur - TPAUSE, vol);
      sound (0, TPAUSE, vol);
#else
      sound (nfreq[note], ndur, vol);
#endif
      is = FALSE;
    }
    /* pause */
    if ((c == 'p') || (c == ' '))
      sound (0, ndur, vol);

    /* note length */
    if (c == '1') ndur = dur;
    if (c == '2') ndur = dur / 2;
    if (c == '4') ndur = dur / 4;
    if (c == '8') ndur = dur / 8;
    if (c == '6') ndur = dur / 16;


    if (c == 'r')
    {
      /* reset */
      octave = 2;
      ndur = dur / 4; /* start value 1/4 note */
    }

    /* change octave */
    if (c == '+')
      if (octave < 4)
        ++octave;
    if (c == '-')
      if (octave > 0)
        --octave;

    /* increment next note cis, dis,.. */
    if (c == '#')
      is = TRUE;

  }
  return 0;
}
