#include <string.h>

#include "owbasic.h"
#include "ostring.h"

#undef tolower

/* tools for uchar[] */
int strcmp_nf (const uchar* sn, const uchar FAR* sf)
{
  /* compare near uchar[] and FAR uchar[]*/
  while ((*sn != 0) && (*sn == *sf))
  {
    ++sn;
    ++sf;
  }
  if (*sn > *sf)
    return 1;
  if (*sn < *sf)
    return -1;
  return 0;
}

void strcpy_nf (uchar* sn, const uchar FAR* sf)
{
  /* copy FAR_string sf to near_string sn*/
  register uchar cch;
  do
    cch = *(sn++) = *(sf++);
  while (cch);
  /*while (*sf)
    *(sn++) = *(sf++);
  *sn = 0;*/
}

#ifndef NEVER_DEFINED
/* copy "special" strings (with given stringlength)*/
void strcpy_ident(uchar *d,const uchar *s)
{
  strncpy((char*)d,(char*)s,IDENTLEN);
}

void strcpy_string(uchar *d,const uchar *s)
{
  strncpy((char*)d,(char*)s,MAXSTRLEN);
}

void strcpy_name(uchar *d,const uchar *s)
{
  strncpy((char*)d,(char*)s,NAMELEN);
}
#endif

void strlower(uchar* s)
{
  /* convert uchar[] to lower case letters */
  while (*s)
  {
    *s = tolower (*s);
    ++s;
  }
}

/* string normalization */
void strupper (uchar* s)
{
  /* convert uchar[] to upper case letters */
  while (*s)
  {
    *s = toupper (*s);
    ++s;
  }
}

void trim (uchar* s)
{
  /* remove leading and trailing white space */
  uchar *ptr1, *ptr2;
  ptr1 = ptr2 = s;
  while ((*ptr1 == ' ') || (*ptr1 == '\t'))
    ++ptr1;
  while (*ptr1)
    *(ptr2++) = *(ptr1++);
  --ptr2;
  while ((ptr2 > s) && ((*ptr2 == ' ') || (*ptr2 == '\t')))
    --ptr2;
  ++ptr2;
  *ptr2 = 0;
}

/* count appearances of a character in string */
int countc (uchar c, const uchar* s)
{
  int count = 0;
  while (*s)
  {
    if (*s == c)
      ++count;
    ++s;
  }
  return (count);
}

static int NEAR strcmp_ncs (const uchar* lhs, const uchar* rhs, int len)
{
  while (len > 0)
  {
    if (tolower (*(lhs++)) != tolower (*(rhs++)))
      break;
    --len;
  }
  return (len);
}

int pos (const uchar* needle, const uchar* haystack, bool case_sensitive)
{
  int plen = strlen ((const char*) needle);
  int apos = 0;
  while ((haystack[apos] != 0)
      && (case_sensitive
       ? strncmp ((const char*) needle, (const char*) haystack + apos, plen)
       : strcmp_ncs (needle, haystack + apos, plen)))
    ++apos;

  if (haystack[apos] != 0)
    return (apos);
  return (-1);
}

#ifdef LSI_C
char nbuffer[NEARBUFFERSIZE];
char* nbptr = (char*) stringstack3;

char* Near (char FAR* st)
{
  /* copy FAR string to NEAR nbuffer */
  register char *np = nbptr;
  do
    *(np++) = *(st++);
  while ((*(np - 1) != 0x00) && (np <= (nbptr + MAXSTRLEN)));
  return (nbptr);
}
#endif

