#ifndef COMP_H
#define COMP_H

#include "modeldef.h"
#include "owbasic.h"

/* lexical analysis */
void putback(void);
#ifdef DEBUGLEX
int lex (void);
#else
int lex1 (void);
#define lex() lex1()
#endif

#define IS_WHITE 1
#define IS_DELIM 2
#define IS_EDELIM 4
#define IS_OPERATOR 8
#define IS_MDIGIT 16
#define IS_HDIGIT 32
#define IS_MALPHA 64

#define isdelim(c)    (chartype[c] & IS_DELIM)
#define iswhite(c)    (chartype[c] & IS_WHITE)
#define isedelim(c)   (chartype[c] & IS_EDELIM)
#define isoperator(c) (chartype[c] & IS_OPERATOR)
#define ismalpha(c)   (chartype[c] & IS_MALPHA)
#define ismdigit(c)   (chartype[c] & IS_MDIGIT)
#define ishdigit(c)   (chartype[c] & IS_HDIGIT)

extern uchar FAR chartype[];

#define LEX_CHAR      (*prog)
#define LEX_INCCHAR() ++prog
#define LEX_CHARPP()  (*(prog++))
#define LEX_PPCHAR()  (*(++prog))

void progeol (void);
void skipws (void);
type_t suffix_type (void);
#define prognl()  progeol (); ++prog
void expect (int k);
void expect_label (void);
void tokncpy (uchar* dest, int n);
void tokcpy_ident (uchar *dest);
void sconstncpy (uchar *dest, int n);

/* compiler (parser) functions */
int  compile (int lastcode);

void compile_overlay (void);
void compile_procfunc (bool func);
void compile_endp (bool func);
bool compile_userprocedure (type_t* ltype, bool is_fptrcall);
int  compile_function (datatype_t* ltype);
void compile_fptrdecl (void);
int  compile_command (void);
bool compile_include (bool);
void compile_assignment (void);
void compile_print (void);
void compile_next (void);
void compile_input (void);
void compile_for (void);
void compile_goto (int jmp);
void compile_if (void);
void compile_else (void);
void compile_endif (void);
void compile_gosub (void);
void compile_return (void);
void compile_exit (void);
void compile_dim (void);
void compile_const (void);
void compile_system (void);
void compile_wait (void);
void compile_case (void);
void compile_endcase (void);
void compile_endwhile (void);
void compile_local (void);
void compile_break (int jmp, int n);
void compile_type (void);
void compile_macro (void);
void compile_enum (void);
void compile_version (void);
/*
  void compile_repeat (void);
  void compile_until (void);
*/
void compile_while (void);
void compile_do (void);
void compile_loop (void);

#if FALSE
#ifdef LSI_C
 #define MACRO_EXPANSION_SPACE (((uchar*) (FB + 2048 + 1)))
 #define MACRO_EXPANSION_SIZE  1024
#else
 #define MACRO_EXPANSION_SPACE _macro_expansion_space
 #define MACRO_EXPNASION_SIZE  1024
 extern uchar _macro_expansion_space[MACRO_EXPANSION_SIZE + 8];
#endif
#endif

/*void comp_var (type_t* dtyp, int* use_addr);
int  comp_varaddr (type_t* dtyp);
void comp_varindex (type_t* ltype);
void comp_array (type_t *dtyp);*/


void   comp_drop (type_t ltype);
int    comp_str (void);
void   comp_number (void);
bool   comp_lvalue (datatype_t* ltype);
void   comp_para (void FAR* para, int FAR* proc, uchar* version, int delim, uchar parawidth);
type_t comp_fptrdeclaration (bool is_typedef);

/*int store_label (void);*/


void serror (int error);
void serror1 (int error, const uchar* s);
void serrortoken (int error);

extern uchar* rterrormsg;

bool comp_simple_exp (type_t wtype);
bool comp_exp (datatype_t* wtype);
bool comp_exput (datatype_t* wtype);

int  expect_simple_constexp (type_t type);
int  expect_constexp (const datatype_t* type);
void expect_constexput (datatype_t* ltype);

int getint (PC p);
#define getfloat(p) *((FLOATT*) (p.c + 1)) /* float literals as code */
/*#define getfloat(p) *((FLOATT FAR*) (&xdata[*((int*) (p.c + 1))]))*/ /* float literals as fconst() */

int FAR* custom_data (int addr);
int newentry (datatype_t dtype, const uchar *n);
int newentry_custom (datatype_t dtype, const uchar* name, int adsize);
int findentry_name (datatype_t* dtypep, const uchar* n);
int findentry_token (datatype_t* dtypep);
int searchentry_name (datatype_t* dtypep,const uchar* n);
int searchentry_token (datatype_t* dtypep);
int expandentry (int addr, int el, int el_per_line);

void iconst (int i);
void fconst (FLOATT f);

void listreset (void);
void listhide (int pos);
int find_label (uchar* s);
int find_token_label (void);
void check_labels (void);

void create_variables (void);

int new_procfunc (uchar* s,PC addr, type_t typ);
int search_procfunc (uchar* s, type_t typ);

type_t typesize (type_t type);

extern int lastaddr;
extern int listnext;

/*extern int found_procfunc;
extern int found_variable;*/

void set_global_variables (void);
void init_global_variables (void);


/*
extern int datanext;  first free "address" after variables (runtime)
extern int firstconst; start of constants (compile and runtime)

defined in owbasic.h
*/

type_t DT (type_t dtyp);

extern type_t     default_datatype;

extern int        token_type;
extern datatype_t datatype;

extern int    lexint;
extern FLOATT lexfloat;
extern int    tok;
extern uchar* token_start;
extern uchar* token_end;
extern uchar* quote;
extern PC     apc;
extern int    varaddr;

extern uchar lex_ref;


typedef struct
{
  uchar FAR *command;
  int tok;
} Keyword;

typedef struct
{
  uchar FAR*    name;
  type_t        type;
  unsigned char size;
  union
  {
    int*      index;
    int FAR** ptr;
    int       val;
  } u;
} GlobalVar;

typedef void (*CompileFunction) (void);
typedef struct
{
  int             tok;
  CompileFunction func;
} CompileFunctionTable;


#define F_MAXPARAM 7 /* 10 before, 7 minimum (MB,4.22) */
#define F_MAXVERSIONS 3 /* 10 before, 3 minimum (MB,4.22) */

#define P_MAXPARAM 9 /* 10 before, 9 minimum (MB,4.22) */
#define P_MAXVERSIONS 4 /* 10 before, 4 minimum (MB,4.22) */

typedef struct keywordpr
{
  uchar FAR*  command;
  smalltype_t paratype[P_MAXPARAM];
  int         bincode[P_MAXVERSIONS];
} KeywordPr; /* LSI C: 26 bytes, __HITACHI__: 38 bytes */
        /* old: LSI C: 21 bytes, __HITACHI__: 29 bytes */

typedef struct keywordfn
{
  uchar FAR*  command;
  int         rtype;
  smalltype_t paratype[F_MAXPARAM];
  int         bincode[F_MAXVERSIONS];
} KeywordFn; /* LSI C: 19 bytes, __HITACHI__: 27 bytes */

#define MAXPARAM 10


extern CompileFunctionTable FAR comptable[];
extern GlobalVar FAR vartable[];
extern Keyword FAR keyfunctable[];
extern Keyword FAR keytable[];
extern Keyword FAR commandtable[];
extern KeywordPr FAR proctable[];
extern KeywordFn FAR functable[];

int look_up (Keyword FAR* list);

extern int last_proc;
/*--- include-stack ---*/

typedef struct {
  int    line;
  uchar* prog;
  word   fp;
  byte   min;
  byte   progtype;
  uchar  fname[NAMELEN];
} SourcePtr; /* LSI_C: 40 bytes, __HITACHI__: 44 bytes */

#define INCSTACKSIZE 6 /* LSI_C: 240 bytes, __HITACHI__: 264 bytes */
extern SourcePtr IncStack[INCSTACKSIZE];
extern SourcePtr *IncStackPtr;
#define ITOS (*IncStackPtr)

#endif

