#ifndef _OBASIC_H
#define _OBASIC_H

#include "modeldef.h"
#include "ostring.h"
#include "msg.h"

#define FAST
#define FAST_MACRO
#define FAST_MATH

#define SMALL_FBDATETIME
#define SMALL_NO_BUILTIN
#define SMALL_NO_AAS
#define SMALL_NO_FORMFUNCS
#define SMALL_INTERFM

#define SMALL_NO_SCREEN_SAVE
#define SMALL_NO_OLDCRCMODE
/*#define SMALL_NO_CACHE*/
#define SMALL_NO_OPTIONS

#define SMALL_INPUTI_AS_INPUTF

#ifdef LSI_C
 #define REMOTE_CONTROL
#endif

#ifdef __DEBUG__
 #define OWBVersion 503
#endif

/* length of different strings */
/* identifiers */
#define IDENTLEN    30 /* 17 before */
/* memo names, Shared record names */
#define NAMELEN     30


/*#define DEBUGCOMP*/
/*#define DEBUGLEX*/
/*#define DEBUGLIST*/
/*#define DEBUGINTER*/

#define PROGNAME "OWBasic"

/* special characters */
#define NEWLINESYMBOL 0xAE
#define NEWLINE '\r'


/* datatype */
typedef word  type_t;
typedef uchar smalltype_t;
typedef struct
{
  type_t  type; /* basic datatype */
  int   ext;  /* extended datatype information (e.g. type definition, ...)*/
} datatype_t;

/* UNKNOWN must be set to 0 */
#define DUNKNOWN         0x0000
#define DINTEGER         0x0001
#define DFLOAT           0x0002
#define DBOOLEAN         0x0003
#define DSTRING          0x0004
#define DCHAR            0x0005
#define DFPTR            0x0006
#define DENUM            0x0007

#define DCONSTANT        0x000A
#define DPROCFUNC        0x000B
#define DTYPEDEF         0x000C
#define DMACRO           0x000D
#define DLABEL           0x000E
#define DREFERENCE       0x000F
#define DPOINTER         DREFERENCE
#define DDATA            DCONSTANT
#define DPROCEDURE       DCONSTANT

#define TYPEMASK         0x000F
#define EVALMASK         (TYPEMASK | KREFERENCE | KARRAY | KPOINTER)
#define LASTVARTYPE      0x0009
#define LASTCVARTYPE     0x000A


/* special kind of data (bit mask) */
#define KREFERENCE       0x0010
#define KLAST            0x0020
#define KARRAY           0x0040
#define KPOINTER         0x0080

#define KCONST           0x0100
/*#define KLOCAL           0x0200*/

/* token_type */
#define TCHAR            1
#define TKEY             2
#define TSTRING          3
#define TEDELIM          4
#define TOPERATOR        5
#define TNUMBER          6
#define TIDENT           7
#define TDELIM           8
#define TLABEL           9


/*
 * values for tok ( lex() )
 * for single characters their ASCII value is used
 */

#define LESS            256
#define GREATER         258
#define NOT_LESS        259
#define NOT_EQUAL       260
#define NOT_GREATER     261
#define DELIMITER       262
#define NUMBER          263
#define IDENT           264
#define LABEL           265
#define STRING          266
#define INT             267
#define FLOAT           268
#define CONST           269
/*#define OP_PTRADDR      281
#define OP_PTRARRAYADDR 283*/
/* -------------------------- */
#define PRINT           301
#define INPUT           302
#define IF              303
#define THEN            304
#define ELSE            314
#define ENDIF           322
#define CASE            352
#define SELECT          353
#define ENDCASE         354
#define UNTIL           356
#define WHILE           357
#define WEND            358
#define DO              359
#define LOOP            360
#define FOR             305
#define NEXT            306
#define TO              307
#define GOTO            308
#define GOSUB           311
#define RETURN          312
#define END             313
#define REM             318
#define DIM             325
#define DEFAULT         326
#define WAIT            328
#define INCLUDE         337
#define APPEND          338
#define OVERLAY         339
#define STEP            340
#define VERSION         341
#define PROC            342
#define ENDP            343
#define AND             345
#define OR              346
#define VAR             347
#define FILEBUFFER      348
#define XOR             349
#define True            350
#define False           351
#define FUNC            361
#define CLIPBOARD       362
#define LOCAL           363
#define CALL            364
#define BREAK           365
#define CONTINUE        366
#define FRETURN         367
#define NOT             368
#define PROP_FBBSIZE    369
#define TYPE            370
#define MACRO           371
#define KEY_ENUM        372
#define FPTR            373

#define D_VARDEF        401
#define D_PROCFUNCDEF   402

#define F_INT           1
#define F_BOOL          2
#define F_FLOAT         3
#define F_STRING        4
#define F_CHAR          5
#define F_ENUM          6
#define F_POINTER       7
#define F_ASC           10
#define F_ARRAYSIZE     11
#define F_CALL          20
#define F_PROC          21
#define F_FUNC          22


#define UNKNOWN         999

/*#ifdef SMALL_NO_AAS
 #define B_ARRAY_ADDR_S   B_ARRAY_ADDR
 #define B_ARRAY2_ADDR_S  B_ARRAY2_ADDR
#endif*/

/*
 * constants for ReturnCode of compile_command
 */

#define RC_CONTINUE 0x01
#define RC_INCLUDE  0x02
#define RC_APPEND   0x03
#define RC_EXIT     0x04

/*
 * constants for status of mode
 */
#define STATUS_NORMAL     0x00
#define STATUS_RC         0x01
#define STATUS_FIND_RTE   0x02
#define STATUS_BUILTIN    0x03
#define STATUS_NO_AUTORUN 0x04
#define STATUS_CACHE      0x05
#define STATUS_PARAM      0x08

#define STATUS_CALLMASK   0x07
#define STATUS_IEDIT      0x01
#define STATUS_QUIET      0x03

extern word mymode;
extern word mystatus;

void Restart (byte status);

/*
 * internal status, return code
 */
#define O_NORMAL        0
#define O_COMPILE_ERROR 1
#define O_RUNTIME_ERROR 2

/*extern byte oret;*/

extern byte intchar;
extern byte floatchar;

/* top of stack for control structures (compiler) */
extern int ctos;

/*
 * Limits
 */
#ifdef LSI_C
#define FOR_NEST        19 /* 15 originally */
#define NUM_PROC        250

#ifndef DEBUGCOMP
#define PROG_SIZE       0x1600 /* 0x1680 max., 0x1410 originally */
#else
#define PROG_SIZE       0x800
#endif

#define G_STACK_SIZE    20
#define DATASIZE        0x3F00 /* 0x3F00 max., 0x3000 originally */
#define MAXSTRLEN       220 /* 150 originally */
#define S_STACK_SIZE    5
#endif

#ifdef __HITACHI__
#define FOR_NEST        160
#define NUM_PROC        1000
#define PROG_SIZE       90000
#define G_STACK_SIZE    50
#define DATASIZE        0x30000
#define MAXSTRLEN       1500
#define S_STACK_SIZE    5
#endif


#include "trm.h"
#include "type.h"
#include "stacks.h"

#include "file.h"
#include "code.h"
#include "inter.h"
#include "fnames.h"

extern int progcat;

extern PC g_stack[G_STACK_SIZE];
extern PC* g_stack_ptr;

PC gpop(void);
void gpush(PC pc);

extern uchar *prog;	/* current pos. to be compiled */
extern uchar *aprog;	/* pointer for debug output lex() */
extern uchar *prog_start; /* the source */

typedef void (FAR * interfunc)(void);
#ifdef DEBUGINTER
#define execute executec
#else
#define execute executea
#endif

extern byte declaration_required;
extern bool overlays_used;
/*extern bool show_compiling;*/
/*extern bool save_to_cache;*/
extern int datanext; /* first free "address" after variables (runtime) */
extern int firstconst; /* start of constants (compile and runtime)*/

extern interfunc FAR ftable[];
extern codet bin[]; /* binary code */
extern PC bin_start;
extern PC pc;
extern PC lastpc;

/*void _asm_ii ();
#define GETIPC(pp) _asm_ii("\nMOV BX,[pc_].W\nMOV AX,[BX].W\nADD BX,2\nMOV [pc_].W,BX\nMOV BX," #pp "_\nMOV [BX],AX")*/

#define GETIPC(ii) ii = *pc.i; ++pc.i


#ifdef LSI_C
 void _asm_pp ();
 #define GETPCPC(pp) _asm_pp("\nMOV BX,[pc_].W\nMOV AX,[BX].W\nADD BX,2\nMOV [pc_].W,BX\nMOV BX," #pp "_\nMOV [BX],AX")
#else
 #define GETPCPC(pp)  pp = *pc.pc; ++pc.p
#endif

#define GETCODEPC(code) code = *pc.c; ++pc.c

#define GETFLOATPC(fl)  fl = *pc.f; ++pc.f

extern int FAR xdata[DATASIZE];
#define XDATA_BYTE ((byte*) (xdata))

#ifndef SMALL_NO_BUILTIN
extern codet FAR ibin[]; /* initialization data for binary code */
extern unsigned int initcodesize;
#ifndef SMALL_ARRAY
extern const int FAR xdata_var[];
#define XDATA_VAR(n)  (xdata_var[n])
#else
extern const unsigned char FAR xdata_var[];
#define XDATA_VAR(n)  ((xdata_var[n] == 255) ? 1302 : xdata_var[n])
#endif
extern const int FAR xdata_const[];
#endif


#ifdef SMALL_NO_SCREEN_SAVE
 #define SCREEN_SAVE()
 #define SCREEN_RESTORE()
#else
 #ifdef LSI_C
  #define SCREEN_SAVE()    int _screen[1602];              \
                           _screen[0] = _screen[1] = 160;  \
                           LibGetGraph (0, 0, 160, 160, (byte FAR*) (&_screen[2]))
  #define SCREEN_RESTORE() LibPutGraph (0, 0, (byte FAR*) _screen)
 #endif
 #ifdef __HITACHI__
  #define SCREEN_SAVE()    byte _screen[3204];             \
                           ((short*) _screen)[0] = 160;    \
                           ((short*) _screen)[1] = 160;    \
                           LibGetGraph (0, 0, 160, 160, &_screen[4])
  #define SCREEN_RESTORE() LibPutGraph (0, 0, _screen)
 #endif
#endif

extern int StringStackPointer;
extern uchar *SStack[S_STACK_SIZE];

#ifdef __HITACHI__
/*template <class T> class Pointer
{
  T* ptr;
public:
  Pointer ()      {}
  Pointer (T* p)  {ptr = p;}
  ~Pointer ()     {}

  T& operator * (void);
};

inline int Pointer <int>::operator * (void)
{
  if ((ptr & 0x01) == 0)
    return (*ptr);
  return (((*((byte*) ptr))      << 24)
      || ((*(((byte*) ptr) + 1)) << 16)
      || ((*(((byte*) ptr) + 2)) << 8)
      ||  (*(((byte*) ptr) + 3)));
}*/
#endif

#define STOS stos
#define SST1 sst1
#define SST2 sst2
extern uchar *stos;
extern uchar *sst1;
extern uchar *sst2;

void spop(int vadr);
void spop1(void);
void spop2(void);
void spopn(int n);
void spush(int vadr);
void spush1(void);

extern char FAR sin_tab[];

extern jmp_buf errorret;

/*void NewInp(byte *c,int maxchr);*/

/* Touch variables */
/*extern int tchxi, tchyi, tchobji, tchstsi;*/ /* indices of touch variables */
extern int FAR *tchxp;
extern int FAR *tchyp;
extern int FAR *tchstsp;
extern int FAR *tchobjp; /* addresses of touch variables */

extern int pvmodeli, owbversioni; /* indices of version informations */

extern uchar block_errors;
extern int error;
extern int inputrc;

int execute(void);

void loadsettings(void);
void savesettings(void);
void editsettings(void);

/* runtime error */
void rterror(const uchar FAR *func,int enr);


#define PREF_SETTINGS 12
typedef struct pref_ {
  /* persistent user settings */
  int  ver;          /* 0 */
  byte progsub;      /* 1 */    /* program submode/category */
  byte includesub;   /* 2 */    /* include submode/category */
  byte statistics;   /* 3 */    /* show statistics */
  byte startmenu;    /* 4 */    /* program menu or file menu */
  byte finalwait;    /* 5 */    /* behaviour after end */
  byte askabort;     /* 6 */    /* ask before abort */
  byte clearoverlay; /* 7 */    /* clear overlays after end */
  byte autorun;      /* 8 */    /* implicitly start program specified by autorun_name[] */
  byte restart;      /* 9 */    /* restart OWBasic after end */
  byte caching;     /* 10 */    /* load programs from cache if available */
  byte searcherror; /* 11 */    /* ask for searching errors */
  byte call_cache;  /* 12 */    /* load programs from cache when called from external application */

  /* temporary switches */
  uchar autorun_name[NAMELEN];  /* name of autorun program */
} Pref;

typedef struct switch_
{
  /* temporary settings */
  /*  byte printwhilecomp;*/
  byte calllevel;
  byte statistics;
  byte progsub;
  byte includesub;
  byte caching;     /* caching allowed */
  byte remote;      /* is remote session */
  byte cached;      /* current program cached */
  byte savetocache; /* write current program to cache */
  byte quiet;       /* show compiler messages */
  byte rterror;     /* search for error */
  /*byte recall;*/  /* recall */
  /*int binstart;*/
  int last;
} Switches;

#define SETTINGSVER (0x10 + sizeof (Pref))

extern Pref pref;
extern Switches switches;

extern bool isFindRTE;

extern int desired_version;
/*extern int current_version;*/
extern int owbparameter;

extern int errorhandler;
extern int exithandler;

extern int pvmodel;

#ifdef LSI_C
int getmodel (void);
#endif


/* organization of programs */

#define PROGRAMSIG  (uchar *) "! "
#define INCLUDESIG  (uchar *) "!*"
#define APPENDSIG   (uchar *) "!+"

#define MMAIN     1
#define MINCLUDE  2
#define MAPPEND   3

bool loadprogmemo ();
bool loadprog ();

void default_prefs (void);

#define M_FILE    1
#define M_PREFS   2
#define M_PROGRAM 3
#define M_READY   99

int filemenu();
int prefsmenu(void);
int programmenu();

/* dyn string storage */
extern int firststring;
void resetstrings(void);
void garbagecollection(void);
void getvar(int vadr,uchar *where);
void putvar(int vadr,const uchar *value);

/* fast timer ...*/
#ifdef LSI_C
extern unsigned int counter;
extern void setintvec(void);
extern void setintvecdiv0(void);
extern void resetintvec(void);
extern int timerruns;
#endif

extern char TimerRuns;

/*void InitTimer (void);*/ /* at the moment the same as StartTimer () */
#define InitTimer() StartTimer ()
FLOATT Timer (void);
void StopTimer (void);
void StartTimer (void);
void PauseTimer (void);
void ContinueTimer (void);

int InitSync (FLOATT dt);
bool Sync (void);
#endif
