#include "ibuild.h"
#include <ctype.h>

KeywordFn ft[MAXFUNC];
BCodeEntry bt[MAXCODE];

int funcnumber;
int maxparnumber;
int bcodenumber;
int linenumber;

/* tools for strings */
bool whitespace (char c)
{
  return ((c == ' ') || (c == '\t'));
  /*if (c==' ') return true;
  if (c=='\t') return true;
  return false;*/
}

void trim (string &s)
{
  string hs = "";
  unsigned int i = 0;
  while ((i < s.length ()) && whitespace(s[i]))
    ++i;
  while ((i < s.length ()) && !whitespace(s[i]))
  {
    hs += s[i];
    ++i;
  }
  s = hs;
}

//bool split(string &s,const string &del,string part)
//{
//  if (s.find(del)<0) return false;
//}

/*
   Functions for reading the description file
*/
bool getname (ifstream &df, int i)
{ // name of basic procedure
  char c;

  ft[i].name = "";
  ft[i].kind = E_NAME | E_CODE | E_TABLE | E_INTER | E_RT;
  c = df.get ();
  while (!df.eof () && (c != ';') && (c != '='))
  {
    ft[i].name += tolower (c);
    ft[i].uname += toupper (c);
    c = df.get ();
  }
  if (ft[i].name == "")
  {
    while (!df.eof () && (c != '\n'))
      c = df.get ();
    return false;
  }
  trim (ft[i].name);
  trim (ft[i].uname);

  if (ft[i].name == "#")
  {
    ft[i].name = "";
    ft[i].uname = "";
    ft[i].kind &= ~(E_NAME | E_TABLE);
  }
  if (df.eof ())
    return false;
  if (c == '=')
  {
    // user defined attributes
    ft[i].kind = 0;
    while (!df.eof() && (c != ';'))
    {
      switch (c)
      {
      case 'n':
        ft[i].kind |= E_NAME;
        break;

      case 't':
        ft[i].kind |= E_TABLE;
        break;

      case 'c':
        ft[i].kind |= E_CODE;
        break;

      case 'i':
        ft[i].kind |= E_INTER | E_RT;
        break;
      }
      c = df.get ();
    }
  }
  return true;
}

bool getreturn (ifstream &df,int i)
{
  // return value ("void" for procedures)
  char c;

  c = df.get ();
  ft[i].rtype = DUNKNOWN; // no return = void
  while (c != ';')
  {
    switch (c)
    {
    case 'i': ft[i].rtype = DINTEGER; break;
    case 'f': ft[i].rtype = DFLOAT; break;
    case 's': ft[i].rtype = DSTRING; break;
    case 'b': ft[i].rtype = DBOOLEAN; break;
    case 'c': ft[i].rtype = DCHAR; break;
    case 'v': /* void */ break;
    }
    c = df.get ();
  }
  return (!df.eof ());
}

bool getparameter (ifstream &df,int i)
{
  // read parameter list
  char c;
  int j;
  c=df.get();
  j=0;
  ft[i].parnumber=0;
  ft[i].versions=1;
  while ((!df.eof())&&(c!=';')&&(j<MAXPARAM))
    {
      ft[i].paratype[j]=0;
      ft[i].def[j]="";
      while ((!df.eof())&&(c!=',')&&(c!=';')&&(c!='='))
  {
    // read all attributes of parameter
    switch (c)
      {
      case 'i': ft[i].paratype[j]|=DINTEGER; break;
      case 'f': ft[i].paratype[j]|=DFLOAT; break;
      case 's': ft[i].paratype[j]|=DSTRING; break;
      case 'c': ft[i].paratype[j]|=DCHAR; break;
      case 'b': ft[i].paratype[j]|=DBOOLEAN; break;
      case 'v': ft[i].paratype[j]|=KREFERENCE; break;
      case 'a': ft[i].paratype[j]|=KARRAY; break;
      case 'l': ft[i].paratype[j]|=KLAST; break;
      default:
        cerr <<"Error " << "owb.dsc " << linenumber << ":" ;
        cout << "Unknown Typeletter " << c << endl; break;
      }
    c=df.get();
  }
      if (c=='=')
  {
    // read default value
    c=df.get();
    while ((!df.eof())&&(c!=',')&&(c!=';'))
      {
        ft[i].def[j]+=c;
        c=df.get();
      }
  }
      if ((ft[i].paratype[j] & KLAST)!=0) ft[i].versions++;
      j++;
      ft[i].parnumber=j;
      if (c==',') c=df.get();
    }
  if (df.eof()) return false;
  return true;
}

bool getiname(ifstream &df,int i)
{
  // read name of interpreter function
  char c;
  ft[i].cname="";
  ft[i].cuname="";
  ft[i].clname="";
  c=df.get();
  while ((!df.eof())&&(c!=';')&&(c!='\n'))
    {
      ft[i].cname+=c;
      ft[i].cuname+=toupper(c);
      ft[i].clname+=tolower(c);
      c=df.get();
    }
  if (c=='\n') linenumber++;
  if (ft[i].cname=="")
    {
      ft[i].cname=ft[i].name;
      ft[i].cuname=ft[i].uname;
      ft[i].clname=ft[i].name;
    }
  trim(ft[i].cname);
  trim(ft[i].cuname);
  trim(ft[i].clname);
  return true;
}

bool getcondition(ifstream &df,int i)
{
  // read name of condition for conditional compilation
  char c;
  ft[i].condition="";
  c=df.get();
  while ((!df.eof())&&(c!=';')&&(c!='\n'))
    {
      ft[i].condition+=c;
      c=df.get();
    }
  if (c=='\n') linenumber++;
  trim(ft[i].condition);
  return true;
}

bool readdescription()
{
  // read description file and stores values is function table
  string iline;
  int i;
  ifstream df("owb.dsc");
  linenumber=1;
  if (! df.is_open()) return false;
  i=0;
  while (!df.eof()&&(i<MAXFUNC))
    {
      if (getname(df,i))
  {
    //	  cout << ft[i].name << endl;
    if (!getreturn(df,i)) return false;
    if (!getparameter(df,i)) return false;
    if (!getiname(df,i)) return false;
    if (!getcondition(df,i)) return false;
    i++;
    funcnumber=i; // number of defined functions
  }
    }
  if (i>=MAXFUNC) return false;
  return true;
}
//--------------------------------
// tools to create outputfiles
//--------------------------------
void header(ofstream &os,string content="",
      string comment1="/*",string comment2="*/")
{
  // add comment header to generated files
  os << comment1 << " This file is created by ibuild. " << comment2 << endl;
  if (content!="")
    {
      os << comment1 << " It contains " << content << ". " << comment2 << endl;
    }
  os << comment1 << " This is an automatically created file. " << comment2 << endl;
  os << comment1 << " Don't change this file, but change owb.dsc or the ibuild sources! " << comment2 << endl;
}

string hs(int i)
{
  int z;
  string ret="";
  if (i==0) return "0";
  while (i>0)
    {
      z=i % 10;
      i=i/10;
      ret=(char)(z+48)+ret;
    }
  return ret;
}

//--------------------------------------------------------
bool preparedata()
{
  // prepares data from function list ft
  // creates binary code list bt
  int i,j;
  int code;
  KeywordFn h;
  // sort functions by name
  // simple bubble sort
  for (j=0;j<funcnumber-1;j++)
    {
      for (i=0;i<funcnumber-1;i++)
  {
    if (ft[i].name>ft[i+1].name)
      {
        h=ft[i];
        ft[i]=ft[i+1];
        ft[i+1]=h;
      }
  }
    }

  // write codes into functionlist
  code=0;
  for (i=0;i<funcnumber;i++)
    {
      if (CODE(ft[i]))
  {
    ft[i].bincode=code;
    code+=ft[i].versions;
  }
      else ft[i].bincode=-1;
    }

  // create binary code list (one entry per calling version of function)
  maxparnumber=0;
  code=0;
  for (i=0; i<funcnumber; i++)
    {
      if (CODE(ft[i]))
  {
    // call with all parameters
    bt[code].fnr=i;
    bt[code].parnumber=ft[i].parnumber;
    if (ft[i].parnumber>maxparnumber)
      maxparnumber=ft[i].parnumber;
    bt[code].full=true;
    code++;
    // call with less parameters (if allowed)
    for (j=ft[i].parnumber-1;j>=0;j--)
      {
        if ((ft[i].paratype[j] & KLAST)!=0) // call allowed ?
    {
      bt[code].fnr=i;
      bt[code].parnumber=j;
      bt[code].full=false;
      code++;
    }
      }
    bcodenumber=code;
  }
    }
  return true;
}
//-------------------------------------------------------
// tools to create names
char TypeLetter(char typ)
{
  switch (typ & TYPEMASK)
    {
    case DUNKNOWN: return 'x';
    case DINTEGER: return 'i';
    case DFLOAT: return 'f';
    case DSTRING: return 's';
    case DBOOLEAN: return 'b';
    case DCHAR: return 'c';
    }
  return '_';
}

string TypeString(char typ)
{
  switch (typ & TYPEMASK)
    {
    case DUNKNOWN: return "DUNKNOWN";
    case DINTEGER: return "DINTEGER";
    case DFLOAT: return "DFLOAT";
    case DSTRING: return "DSTRING";
    case DBOOLEAN: return "DBOOLEAN";
    case DCHAR: return "DCHAR";
    }
  return "Undefined type";
}

string BCodeName(int i)
{
  string s;
  int fnr;
  fnr=bt[i].fnr;
  s="B_"+ft[fnr].cuname;
  if (!bt[i].full) s+=('0'+bt[i].parnumber);
  return s;
}

string IFuncName(int i)
{
  string s;
  int fnr;
  fnr=bt[i].fnr;
  s="x_"+ft[fnr].clname;
  if (!bt[i].full) s+=('0'+bt[i].parnumber);
  return s;
}

string FunctionName(int i)
{
  string s;
  int fnr;
  fnr=bt[i].fnr;
  s=ft[fnr].clname;
  return s;
}

string FName(int i)
{
  return "FN_"+ft[i].uname;
}


void writetype(ofstream &df,char typ)
{
  df << TypeString(typ); // write basetype
  if ((typ & KREFERENCE)!=0) df << "|KREFERENCE"; // write modifiers
  if ((typ & KARRAY)!=0) df << "|KARRAY";
  if ((typ & KLAST)!=0) df << "|KLAST";
}

bool Createfnames()
{
  // creates tables of function names
  // have to add removing of double names
  ofstream nhf("../h/fnames.h");
  int i;
  string olds,news;
  if (!nhf.is_open()) return false;
  header(nhf,"the declarations of function name strings");

  nhf << "#ifndef FNAMES_H" << endl;
  nhf << "#define FNAMES_H" << endl;

  olds="";
  for (i=0;i<funcnumber;i++)
    {
      if (NAME(ft[i]))
  {
    news=FName(i);
    if (olds!=news)
      {
        if (ft[i].condition!="")
    nhf << "#ifdef " << ft[i].condition << endl;
        nhf << "extern uchar FAR " << news << "[];" << endl;
        if (ft[i].condition!="")
    nhf << "#endif " << endl;
        olds=news;
      }
  }
    }

  nhf << "#endif" << endl;

  ofstream ncf("../c/fnames.c");
  if (!ncf.is_open()) return false;
  header(ncf,"the definition os function name strings");

  ncf << "#include <owbasic.h>" << endl;
  ncf << "#include <fnames.h>" << endl;

  olds="";
  for (i=0;i<funcnumber;i++)
    {
      if (NAME(ft[i]))
  {
    news=FName(i);
    if (olds!=news)
      {
        if (ft[i].condition!="")
    ncf << "#ifdef " << ft[i].condition << endl;
        ncf << "uchar FAR " << news << "[]={\"" ;
        ncf << ft[i].name << "\"};"<< endl;
        if (ft[i].condition!="")
    ncf << "#endif " << endl;
        olds=news;
      }
  }
    }
  return true;
}

bool Createftables()
{
  // create compiler tables (procedures and functions)
  // create procedure table
  ofstream ptf("../c/fptable.c");
  int i,j,code;
  if (!ptf.is_open()) return false;
  header(ptf,"the compiler table of procedures");

  // FN_NAMEDFILE,{DSTRING,DINT|KREFERENCE,DINT|KREFERENCE,DUNKNOWN},{B_NAMEDFILE},
  // FN_LOADSHARED, {DSTRING,DUNKNOWN}, {B_LOADSHARED},
  // FN_SAVESHARED, {DSTRING,DUNKNOWN}, {B_SAVESHARED},

  for (i=0;i<funcnumber;i++)
    {
      if (CODE(ft[i]))
  if (TABLE(ft[i]))
    {
      if (ft[i].rtype==DUNKNOWN) // only procedures
        {
    if (ft[i].condition!="")
      ptf << "#ifdef " << ft[i].condition << endl;
    // Name of procedure
    ptf << " " << FName(i) << "," ;
    // parameter types
    ptf << "{" ;
    for (j=0;j<ft[i].parnumber;j++)
      {
        if (j>0) ptf << ",";
        writetype(ptf,ft[i].paratype[j]);
      }
    if (j>0) ptf << ",";
    ptf << "DUNKNOWN},";
    // Binrcodes
    ptf << "{" << BCodeName(ft[i].bincode) ;

    for (j=ft[i].bincode+1; j<=ft[i].bincode+ft[i].versions-1; j++)
      {
        ptf << "," << BCodeName(j) ;
      }
    ptf << "},";


    ptf << endl;
    if (ft[i].condition!="")
      ptf << "#endif" << endl;
        }
    }
    }

  // create function table
  ofstream ftf("../c/fftable.c");
  if (!ftf.is_open()) return false;
  header(ftf,"the compiler table of functions");

  for (i=0; i<funcnumber; i++)
    {
      if (CODE(ft[i]))
  if (TABLE(ft[i]))
    {
      if (ft[i].rtype!=0) // only functions
        {
    if (ft[i].condition!="")
      ftf << "#ifdef " << ft[i].condition << endl;
    // Name of function
    ftf << " " << FName(i) << "," ;
    // return type
    writetype(ftf,ft[i].rtype);
    ftf << "," ;
    // parameter types
    ftf << "{" ;
    for (j=0;j<ft[i].parnumber;j++)
      {
        if (j>0) ftf << ",";
        writetype(ftf,ft[i].paratype[j]);
      }
    if (j>0) ftf << ",";
    ftf << "DUNKNOWN},";
    // Binrcodes
    ftf << "{" << BCodeName(ft[i].bincode) ;

    for (j=ft[i].bincode+1; j<=ft[i].bincode+ft[i].versions-1; j++)
      {
        ftf << "," << BCodeName(j) ;
      }
    ftf << "},";
    ftf << endl;
    if (ft[i].condition!="")
      ftf << "#endif" << endl;
        }
    }
    }

  return true;
}

//-------------------------------------------------------------------
bool Createbcodeheader()
{
  // create header file with binary codes
  ofstream bhf("../h/bcodes.h");
  int i;

  if (!bhf.is_open()) return false;
  header(bhf,"the binary code definitions");

  for (i=0;i<bcodenumber;i++)
    {
      bhf << "#ifndef " << BCodeName(i) << endl;
      bhf << "#define " << BCodeName(i) << " " << i << endl;
      bhf << "#endif" << endl;
    }
  return true;
}

//-------------------------------------------------------------------
bool Createjumptable()
{
  // create jumptable for interpreter
  // pv3
  ofstream jtf("../c/ftable.a86");
  int code;

  if (!jtf.is_open()) return false;
  header(jtf,"the jumptable of the interpreter",";","");

  //	preamble
  jtf << "DGROUP	GROUP	DATA,BSS" << endl ;
  jtf << "DATA	DSEG" << endl ;
  jtf << "BSS	DSEG" << endl ;
  jtf << "ftable_DATA	DSEG	PARA 'FAR_DATA'" << endl ;
  jtf << "ftable_::" << endl ;
  // interpreter atoms
  for (code=0;code<bcodenumber;code++)
    {
      jtf << "DW " << IFuncName(code) << "_,SEG " ;
      jtf << IFuncName(code) << "_ ; " << code << endl;
      jtf << "EXTRN " << IFuncName(code) << "_" << endl;
    }
  // pv4

  ofstream jtcf("../c/ftable4.c");
  if (!jtcf.is_open()) return false;
  header(jtcf,"the jumptable of the interpreter");

  //	preamble
  jtcf << "#include \"owbasic.h\"" << endl;
  jtcf << "interfunc FAR ftable[]={" << endl;
  // interpreter atoms
  for (code=0;code<bcodenumber;code++)
    {
      jtcf << IFuncName(code) << ", /* " << code << "*/" << endl ;
    }
  jtcf << "};";
  return true;
}
//-------------------------------------------------------------------
bool CreateIFuncHeader()
{
  int code;
  // create declaration of interpreter functions
  ofstream ifh("../h/ifunc.h");

  if (!ifh.is_open()) return false;
  header(ifh,"the declarations of interpreter functions");

  //	preamble
  ifh << "#ifndef _IFUNC_H" << endl;
  ifh << "#define _IFUNC_H" << endl;
  // interpreter functions
  for (code=0; code<bcodenumber; code++)
    {
      ifh << "void " << IFuncName(code) << "(void);" << endl;
    }
  ifh << "#endif" << endl;
  return true;
}
//-------------------------------------------------------------------
bool CreateInterpreterFunctions()
{
  int code;
  int fnr;
  int ic,fc,sc; // counter for stack usage (integer, float, string)
  int pnr; // current parameter number
  int ttyp; // temporary storage of parameter type
  string parstring;
  // create interpreter functions
  ofstream ifc("../c/ifunc.c");

  if (!ifc.is_open()) return false;
  header(ifc,"the interpreter functions");

  //	preamble
  ifc << "#include \"owbasic.h\"" << endl;
  ifc << "#include \"ifunc.h\"" << endl;
  ifc << "#include \"functions.h\"" << endl;
  // interpreter functions for each binary code
  for (code=0; code<bcodenumber; code++)
    {
      fnr=bt[code].fnr;
      if (INTER(ft[fnr]))
  {
    ifc << "void " << IFuncName(code) << "(void) { " << endl;

    switch (ft[fnr].rtype)
      {
      case DINTEGER:
        ifc << " int ret;" << endl;
        break;

      case DBOOLEAN:
        ifc << " int ret;" << endl;
        break;

      case DFLOAT:
        ifc << " FLOATT ret;" << endl;
        break;

      case DSTRING:
        ifc << " char ret[MAXSTRLEN];" << endl;
        break;

      }
    ic=0; fc=0; sc=0;
    parstring="";
    for (pnr=bt[code].parnumber-1;pnr>=0;pnr--)
      {
        ttyp=ft[fnr].paratype[pnr];
        if (parstring!="") parstring=","+parstring;
        if ((ttyp & KREFERENCE) > 0)
    {
      ifc << " int index" << hs(pnr) << "=*(i_stack_ptr-"+hs(ic)+");" << endl;
      switch (ttyp & TYPEMASK)
        {
        case DINTEGER:
        case DBOOLEAN:
          ifc << " int val" << hs(pnr) << "=xdata[index" << hs(pnr) << "];" << endl;
          break;
        case DFLOAT:
          ifc << " float val" << hs(pnr)  << "=*(double*)&xdata[index" << hs(pnr) << "]);" << endl;
          break;
        case DSTRING:
          break;
        }
      parstring="&val"+hs(pnr)+parstring;
      ic++;
    }
        else
    {
      switch (ttyp & TYPEMASK)
        {
        case DINTEGER:
        case DBOOLEAN:
          if (ic==0) parstring="*i_stack_ptr"+parstring;
          else parstring=string("*(i_stack_ptr-")+hs(ic)+")"+parstring;
          ic++;
          break;
        case DFLOAT:
          if (fc==0) parstring="*f_stack_ptr"+parstring;
          else parstring=string("*(f_stack_ptr-")+hs(fc)+")"+parstring;
          fc++;
          break;
        case DSTRING:
          if (sc==0) parstring="hstring[s_stack_ptr]"+parstring;
          else parstring=string("hstring[s_stack_ptr-")+hs(sc)+"]"+parstring;
          sc++;
          break;
        }
    }
      }

    switch (ft[fnr].rtype)
      {
      case DINTEGER:
      case DBOOLEAN:
      case DFLOAT:
        ifc << " ret=" ;
        break;

      case DSTRING:
        if (parstring!="") parstring=","+parstring;
        parstring="ret"+parstring;
        break;
      }

    ifc << " " << FunctionName(code) << "(" << parstring << ");" << endl;

    for (pnr=bt[code].parnumber-1;pnr>=0;pnr--)
      {
        ttyp=ft[fnr].paratype[pnr];
        if ((ttyp & KREFERENCE) > 0)
    {
      switch (ttyp & TYPEMASK)
        {
        case DINTEGER:
        case DBOOLEAN:
          ifc << " xdata[index" << hs(pnr) << "]=val" ;
          ifc << hs(pnr) << ";" << endl;
          break;
        case DFLOAT:
          ifc << " *((double*)&xdata[index" << hs(pnr) << "])=val" ;
          ifc << hs(pnr) << ";" << endl;
          break;
        case DSTRING:

          break;
        }
    }
      }

    switch (ft[fnr].rtype)
      {
      case DINTEGER:
      case DBOOLEAN:
        ifc << " ipush(ret)" << endl ; break;
      case DFLOAT: ifc << " fpush(ret)" << endl ; break;
      }

    if (ic>0) ifc << " ipopn(" << hs(ic) << ");" << endl;
    if (fc>0) ifc << " fpopn(" << hs(fc) << ");" << endl;
    if (sc>0) ifc << " spopn(" << hs(sc) << ");" << endl;

    ifc << "}" << endl;
  }
    }
  return true;
}
//-------------------------------------------------------------------
int main(int argc,char **argv)
{
  string s;

  cout << "interface builder for OWB compiler/interpreter" << endl;

  if (!readdescription())
    {
      cout << "Read description failed" << endl;
      return 1;
    }

  if (!preparedata())
    {
      cout << "Data not consistent" << endl;
      return 1;
    }

  if (!Createbcodeheader())
    {
      cout << "Can't create header file with binary codes" << endl;
      return 1;
    }

  if (!Createjumptable())
    {
      cout << "Can't create jump table assembler file" << endl;
      return 1;
    }

  if (!Createfnames())
    {
      cout << "Can't create function name files" << endl;
      return 1;
    }
  if (!Createftables())
    {
      cout << "Can't create compiler tables" << endl;
      return 1;
    }

  if (!CreateIFuncHeader())
    {
      cout << "Can't create interpreter function declarations" << endl;
      return 1;
    }

  if (!CreateInterpreterFunctions())
    {
      cout << "Can't create interpreter functions" << endl;
      return 1;
    }
  return 0;
}
