#ifndef IBUILD_H
#define IBUILD_H
// Header for interface builder for OWB compiler/interpreter

#include <string>
#include <fstream>
#include <iostream>

using namespace std;

/* datatypes */
#define DUNKNOWN         0x0000
#define DINTEGER         0x0001
#define DFLOAT           0x0002
#define DBOOLEAN         0x0003
#define DSTRING          0x0004
#define DCHAR            0x0005
#define DFPTR            0x0006
#define DENUM            0x0007

#define DCONSTANT        0x000A
#define DPROCFUNC        0x000B
#define DTYPEDEF         0x000C
#define DMACRO           0x000D
#define DLABEL           0x000E
#define DREFERENCE       0x000F
#define DPOINTER         DREFERENCE
#define DDATA            DCONSTANT
#define DPROCEDURE       DCONSTANT

#define TYPEMASK         0x000F
#define EVALMASK         (TYPEMASK | KREFERENCE | KARRAY | KPOINTER)
#define LASTVARTYPE      0x0009
#define LASTCVARTYPE     0x000A


/* special kind of data (bit mask) */
#define KREFERENCE       0x0010
#define KLAST            0x0020
#define KARRAY           0x0040
#define KPOINTER         0x0080

#define KCONST           0x0100


/* kind of entry */
// define then name as string
#define E_NAME 0x01
#define NAME(x) (x.kind & E_NAME)
// assign an interpreter code
#define E_CODE 0x02
#define CODE(x) (x.kind & E_CODE)
// insert into compiler tables
#define E_TABLE 0x04
#define TABLE(x) (x.kind & E_TABLE)
// insert in interpreter table
#define E_INTER 0x08
#define INTER(x) (x.kind & E_INTER)
// generate interpreter function
#define E_RT 0x10
#define RT(x) (x.kind & E_RT)

#define MAXPARAM 20
#define MAXFUNC 1000
#define MAXCODE 1000

// one entry for list of procedures and functions
typedef struct keywordfn
{
  int kind;
  string  name; // procedure/function name (lower case letters)
  string  uname; // procedure/function name (upper case letters)
  char    rtype; // return type (functions) or DUNKNOWN for procedures
  char    paratype[MAXPARAM]; // parameter types
  string  def[MAXPARAM];
  int     parnumber; // number of parameters
  int     bincode;   // first binary code for this proc/func
  int     versions;  // number of versions (different numbers of parameters)
  string  cname;  // basename interpreter function
  string  cuname; // basename interpreter function (upper case letters)
  string  clname; // basename interpreter function (lower case letters)
  string  condition;
} KeywordFn;

// structure for binary codes
typedef struct bcodeentry
{
  int   fnr; // pointer to procedure
  bool  full;
  int   parnumber; // number of parameters (for this version)
} BCodeEntry;
#endif

