/*
 * ModeCall - Library for Pocketviewer for terminal emulation
 *
 * Copyright (C) 2002 Wolfgang Ortmann
 * Contact: noo@uni-jena.de
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <modecall.h>

/*
 * read/write modecall parameter
 */

static uchar FAR *gptr;
#ifdef __HITACHI__
uchar paramstr[128];
#endif

int mc_Init()
{
  if (*MCP_VERSION!=MC_VERSION) return FALSE;
  gptr=MCP_START;
  return TRUE;
}

int mc_InitWrite()
{
  word mode_code, state;
#ifndef __HITACHI__
  word seg,off;
  LibGetMode(&mode_code,&state,&seg,&off);
#else
  void *seg;
  LibGetMode(&mode_code,&state,&seg);
#endif
  *MCP_START=0; 
  *MCP_VERSION=MC_VERSION; /* set version identifier */
  *MCP_MODEWORD=mode_code;
  gptr=MCP_START;
  return TRUE;
}

int mc_WriteInt(int val,uchar delim)
{
  char is[7]={0,0,0,0,0,0,0};
  int n=0;
  if (val>=0)
    {
      is[0]=(val % 10)+'0'; n++;
      val=val/10;
      while (val>0) 
	{
	  is[n]=(val % 10)+'0'; n++;
	  val=val/10;
	}
      n--;
    }
  while ((gptr>MCP_END)&&(n>=0)) *(gptr--)=is[n--];
  *(gptr--)=delim;
  return gptr>MCP_END;
}

int mc_WriteString(const uchar *s,uchar delim)
{
  while ((gptr>MCP_END)&&(*s!=0))
    {
      *(gptr--)=*(s++);
    }
  *(gptr--)=delim;
  return gptr>MCP_END;
}

int mc_WriteEmpty()
{
  if (! (gptr>MCP_END)) return FALSE;
  *(gptr--)=0;
  return TRUE;
}

int mc_WriteParam(int cat,const uchar *name,int cursorpos)
{
  if (!mc_WriteInt(cat,MC_DELIM)) return FALSE;
  if (!mc_WriteString(name,MC_DELIM)) return FALSE;
  return mc_WriteInt(cursorpos,0);
}

static int NEAR isdelim(uchar c) 
{
  return (c==MC_DELIM)||(c==0);
}

int mc_ReadInt(int *i)
{
  *i=-1;
  while (!isdelim(*gptr))
    {
      if (*i<0) *i=0;
      *i=*i*10+*(gptr--)-'0';
    }
  gptr--;/* skip delimiter */
  return TRUE;
}

int mc_ReadString(uchar *st)
{
  while (!isdelim(*gptr))
    {
      *st=*gptr;
      st++; gptr--;
    }
  gptr--;/* skip delimiter */
  *st=0; /* terminate string */
  return TRUE;
}

int mc_isEmpty()
{
  return *gptr==0;
}

int mc_Skip()
{
  while ((gptr>MCP_END)&&(*gptr!=0)) gptr--;
  gptr--;
  return gptr>MCP_END;
}

int mc_ReadParam(int *cat,uchar *name,int *curpos)
{
  if (!mc_ReadInt(cat)) return FALSE;
  if (!mc_ReadString(name)) return FALSE;
  return mc_ReadInt(curpos);
}

