/*
 * ModeCall - Library for Pocketviewer
 *
 * Copyright (C) 2002 Wolfgang Ortmann
 * Contact: noo@uni-jena.de
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <modecall.h>
#include <string.h>
/*
  general functions - status.c
  functions for calling program - status1.c
  functions for called program - status2.c
*/

/*
 * find (user) mode with given name
 * returns mode_code, usable with LibJumpMode
 */

word mc_FindMode(const uchar pname[])
{
  word condition;
  word mode, submode, status;
  word mode_code;
  uchar addinname[16];
  bool found;
  
  condition=IB_DLFIRST_SRCH;                  /* find addin pname */
  do {
    LibGetUserMode(&mode_code,&status,condition); /* get addin mode_code */
    if (mode_code==0xffff) return mode_code;  /* not found */
    condition=IB_DLNEXT_SRCH;                 /* continue search */
    submode=mode_code & 0xff;                 /* split mode_code to */
    mode=mode_code >> 8;                      /* mode and submode */
    LibGetProgramName(addinname,mode,submode);/* get mode name */
    found=strcmp((char*)addinname,(char*)pname)==0;         /* compare */
  } while (!found);
  return mode_code;
}

/*
 * is current mode returning from ModeCall 
 */
bool mc_isReturn()
{
  if ((getstate() & MC_CALLMASK)!=MC_RETURNVALUE) return FALSE;
  return (*MCP_VERSION==MC_VERSION);
}
