/*
 * ModeCall - Library for Pocketviewer
 *
 * Copyright (C) 2002 Wolfgang Ortmann
 * Contact: noo@uni-jena.de
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#ifndef MODECALL_H
#define MODECALL_H

#ifndef __HITACHI__
#include <stdrom.h>
#else
#include <hitachi.h>
#include <stdio.h>
#endif

#include <libc.h>
#include <l_libc.h>

#ifdef __HITACHI__
#define FAR /* nothing */
#define NEAR /* nothing */
#else
#define FAR far
#define NEAR near
#endif

#ifndef UCHAR
#define UCHAR
typedef unsigned char uchar;
#endif

#define MC_CALLMASK 0xF0
#define MC_CALLVALUE 0xE0
#define MC_RETURNVALUE 0xF0
#define MC_LEVELMASK 0x0F

int getstate(void);

bool mc_isCall(void); /* returns TRUE if called via ModeCall */
bool mc_isReturn(void); /* returns TRUE if return from ModeCall */

int mc_CallLevel(void); /* returns call level :
			   0 - no call, but jump
			   1 - first level call
			   2 - second level call (call from call)
			   ...
			*/

word mc_FindMode(const uchar pname[]);
word mc_CallerMode(void);

/* delimiter for parts of file-parameter */
#define MC_DELIM 0xfe

/* functions to read/write parameters */
int mc_Init(void);

int mc_Skip(void);
int mc_isEmpty(void);

int mc_ReadInt(int *i);
int mc_ReadString(uchar *d);

int mc_InitWrite(void);

int mc_WriteInt(int i,uchar delim);
int mc_WriteString(const uchar *s,uchar delim);
int mc_WriteEmpty(void);

int mc_WriteParam(int cat,const uchar *name,int cursorpos);

int mc_ReadParam(int *cat,uchar *name,int *curpos);

int mc_Call();
int mc_Return();

#define MC_VERSION 0x02708

/* as defined in libc.h
   extern volatile byte	SYS_INP_BUF[2049]; /+ Copy/Paste Buffer +/
*/

#define PARAMETER_END 2046

#define MCP_VERSION (word FAR *)&(SYS_INP_BUF[PARAMETER_END])
#define MCP_MODEWORD (word FAR *)&(SYS_INP_BUF[PARAMETER_END-2])
#define MCP_START (uchar FAR *)&(SYS_INP_BUF[PARAMETER_END-3])
#define MCP_END (uchar FAR *)&(SYS_INP_BUF[0])

#endif
