/*
 * SharedLib - Library for Pocketviewer for access to shared file
 *
 * Copyright (C) 2002 Wolfgang Ortmann
 * Contact: noo@uni-jena.de
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "modeldef.h"
#include <string.h>
#include <shared.h>

static uchar filename[]=FILENAME;

static byte mode=0;
static byte submode=0;

#ifdef __HITACHI__
FILE_BUF fbuf;
#endif

int sffindfile(FILE_BUF *fb)
{
  /* find or create shared data file */
  if (mode==0) /* only if not yet done */
    {
      if (!LibGetAllEntry(filename,&mode,&submode)) /* find shared file */
	{
	  /* if not exists, create new one */
	  if (!LibSubEntrySave(filename,&submode)) return FALSE;
	  LibGetAllEntry(filename,&mode,&submode); /* get mode and submode */
	}
    }
  /* Set main and sub entry mode in fb */
  fb->fsb_main_entry_ = mode;
  fb->fsb_sub_entry_  = submode;
  fb->fsb_scrt_info_  = 0x80; /* Open Mode */
  return TRUE;
}

int sffindrecord(const uchar signature[],FILE_BUF *fb,FILE_INFO *fi)
{
  /* search named record and load it*/
  int found=FALSE;
  fi->fp = 0xffff;             /* first search */    
  fi->kind = FILE_KIND_BIN;    /* binary file */
  
  while ((!found) && (LibFileFindNext(fb,fi,0)))
    {
      if (!LibFileReadEx(fb,fi,FREEBYTES/FILE_BLOCK_SIZE)) 
	return FALSE; /* Read error */
      found = strcmp((char*)fb->fbuf.bin.bin_buf,(char*)signature)==0; /*compare signature*/
    }
  
  if (!found) 
    {
      fi->fp = 0xffff;             /* prepare FILE_INFO for "write new" */  
      fi->kind = FILE_KIND_BIN;
    }
  return found;
}

int sfLoadRecord(const uchar *signature,int size,byte FAR *buffer)
{
  int i;
  int bytes;
  FILE_INFO fi;
#ifndef __HITACHI__
  FILE_BUF fbuf;                    /* ! requires very much stack space ! */
#endif
  byte *sptr;
  byte FAR *dptr;
  
  if (! sffindfile(&fbuf)) return FALSE;   /* find shared file */
  if (! sffindrecord(signature,&fbuf,&fi)) /* find and load record */   
    return FALSE;
  
  bytes = fbuf.fbuf.bin.char_num;            /* number of data bytes */
  sptr = fbuf.fbuf.bin.bin_buf;
  while (*sptr!=0) { sptr++; bytes--; }    /* find end of signature */
  sptr++; bytes--;                         /* skip null byte */

  dptr = buffer;

  if (size>bytes) size=bytes;
  for (i=0;i<size;i++)         /* move data to buffer */
    {
      *(dptr++)=*(sptr++);
    }
  return TRUE;
}

int sfSaveRecord(const uchar *signature,int size,byte FAR *buffer)
{
  int i;
  FILE_INFO fi;
#ifndef __HITACHI__
  FILE_BUF fbuf;                  /* requires very much stack space */
#endif
  byte FAR *sptr;
  const uchar *usptr;
  byte *dptr;
  int freebytes=FREEBYTES;

  if (!sffindfile(&fbuf)) return FALSE;      /* find or create shared file */
  sffindrecord(signature,&fbuf,&fi);         /* find record, but ignore error*/
 
  dptr=fbuf.fbuf.bin.bin_buf;
  usptr=signature;
  while (*usptr!=0)                          /* Write signature to filebuffer*/
    {
      *(dptr++)=*(usptr++); 
      freebytes--;
    }
  *(dptr++)=0; freebytes--;

  if (size>freebytes) return FALSE;          /* size too big error */

  sptr=buffer;
  for (i=0;i<size;i++)                       /* move data to filebuffer */
    *(dptr++)=*(sptr++);

  fbuf.fbuf.bin.char_num = dptr-fbuf.fbuf.bin.bin_buf; /* store size in filebuffer */
  return LibFileWrite(&fbuf,&fi);            /* write record */
}



