/*
 * TermLib - Library for Pocketviewer for terminal emulation
 *
 * Copyright (C) 2002 Wolfgang Ortmann
 * Contact: noo@uni-jena.de
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/* input.c */
#include "stdio.h"
#include "define.h"
#include "libc.h"
#include "trm.h"

int InputTrm(uchar *ib,int cnr)
{
  uchar *hp;
  byte aline, acol;
  int ready;
  int cnt;
  TCHSTS tsts;
  byte kycd;
  ModeTrm(M_SHORT);
  LibKeyInit();
  LibDispKey();
  ready=FALSE;
  spline=cursorline;
  spcol=cursorcol;
  while (!ready)
    {
      DisplayTrm();
      kycd=LibGetKeyM(&tsts); /* get key or other events */
      if (kycd!=KEY_NONE)     /* any Key */
	{
	  if ((kycd==0xF0)||(kycd==0xF7)) ready=TRUE;
	  else 
	    if (kycd!=0xf8) OutTrm(kycd);
	}
      else 
	{                     /* no key */
	  switch (tsts.obj)
	    {
	    case OBJ_HIC_ESC:
	      ready=TRUE;     /* ESC pressed */
	      break;
	    } /* switch (tsts.obj) */
	} /* if (kycd!=KEY_NONE) */
    } /* while (!ready) */

  acol=cursorcol; cursorcol=spcol;
  aline=cursorline; cursorline=spline; /* start position */
  cnt=0;
  hp=ib;
  while ( (aline>cursorline)||
	     ((aline==cursorline)&&(acol>cursorcol))
	   )
	{
	if (cnt<cnr) *(hp++)=trmbuffer[cursorline*TRMCOLS+cursorcol];
	cnt++;
	cursornext();
     }

  *(hp++)=0;
  OutTrm(0xF0);
  ModeTrm(M_LONG);
  DisplayTrm(); /* update display */
  switch (kycd)
    {
    case 0xF0: return TRM_ENTER;
    case 0xF7: return TRM_NEXT;
    default: return TRM_ESC;
    }
}
