/*
 * TermLib - Library for Pocketviewer for terminal emulation
 *
 * Copyright (C) 2002 Wolfgang Ortmann
 * Contact: noo@uni-jena.de
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* trm.c */
#include <stdio.h>
#include <libc.h>
#include <string.h>
#ifndef __HITACHI__
#include <memory.h>
#endif
#undef max
#include "modeldef.h"
#include "trm.h"

uchar trmbuffer[TRMCHARS+1];
uchar nflag[TRMLINES];
int trmsize;
int firstline;
int cursorline, cursorcol;
int spline, spcol; /* place to store current cursor position */

#define LINEDIST 9
#define DLINES (160/9)
#define CHARWIDTH 6
#define LINECHARS TRMCOLS

static void NEAR NFlags (void)
{
  /* mark all lines as NEW */
  /*  int i;
      for (i=0;i<TRMLINES;i++) nflag[i]=TRUE;*/
  memset (nflag, 1, TRMLINES);
}

void ModeTrm (int mode)
{
  /* switch mode and set globals */
  switch (mode)
  {
  case M_SHORT:
    trmsize = 9;
    firstline = 0;
    while (cursorline > firstline + 8)
      ++firstline;
    break;
  case M_LONG:
    trmsize = 17;
    firstline = 0;
    break;
  }
  LibClrBox (0, 0, 160, 160 /*trmsize*LINEDIST+4*/);
  LibBox (0, 0, LINECHARS * CHARWIDTH + 4, trmsize * LINEDIST + 4, 1);
  NFlags ();
}

void DisplayTrm(void)
{
  /* display current alpha screen (update) */
  int i;
  int buffer_haddr1;
  int buffer_haddr2;
  uchar hc;
  int line;
  int y_line;

  buffer_haddr1=-TRMCOLS+firstline*LINECHARS;
  buffer_haddr2=0+firstline*LINECHARS;
  line=firstline-1;
  y_line=2-LINEDIST;
  for (i=0; i<trmsize; i++)
    {
      buffer_haddr1+=TRMCOLS;
      buffer_haddr2+=TRMCOLS;
      line++;
      y_line+=LINEDIST;
      if (nflag[line]) /* is line changed ? */
  {
    LibClrBox(2,y_line,LINECHARS*CHARWIDTH,LINEDIST);
    hc=trmbuffer[buffer_haddr2];
    trmbuffer[buffer_haddr2]=0;
    LibPutProStr(IB_CG57FONT,2,y_line,
           &(trmbuffer[buffer_haddr1]),156);
    trmbuffer[buffer_haddr2]=hc;
    nflag[line]=FALSE;
  }
    }
  LibPutDisp ();
}

void InitTrm(void)
{
  /* int i;
     for (i=0;i<TRMCHARS;i++) trmbuffer[i]=' ';*/
  register word* tbp;
  register word* max;
  for (tbp = (word*) trmbuffer, max = tbp + (TRMCHARS / 2); tbp < max; *(tbp++) = 0x2020);
  /*memset (trmbuffer, ' ', TRMCHARS);*/
  NFlags ();
  firstline = 0;
  cursorline = 0;
  cursorcol = 0;
  trmsize = 17;
  firstline = 0;
  /*ModeTrm(M_LONG);*/
  /* DisplayTrm ();*/
}

/*
  byte hex(byte c)
  {
  byte hchar[16]={'0','1','2','3','4','5','6','7',
  '8','9','A','B','C','D','E','F'};
  return hchar[c %16];
  }
*/

void cursornext (void)
{
  ++cursorcol;
  if (cursorcol >= TRMCOLS)
  {
    ++cursorline;
    cursorcol = 0;
  }
}

static void NEAR cursorprev (void)
{
  if (cursorcol > 0)
    --cursorcol;
  else
    if (cursorline > 0)
    {
      --cursorline;
      cursorcol = TRMCOLS - 1;
    }
}

static void NEAR rollup (void)
{
  int i;
  cursorline--;
  for (i=0;i<(TRMLINES-1)*TRMCOLS;i++)
    trmbuffer[i]=trmbuffer[i+TRMCOLS]; /* one line up */
  for (;i<TRMCHARS;i++) trmbuffer[i]=' '; /* clear last line */
  NFlags(); /* all is new */
  if (spline>0) spline--; /* correct saved position */
}

void GotoTrm(int col,int line)
{
  cursorline=line;
  cursorcol=col;
  if (cursorline-firstline>=trmsize)
    {
      NFlags();
      firstline=cursorline-trmsize+1;
    }
}

void InfTrm(int *col,int *line)
{
  *line = cursorline;
  *col = cursorcol;
}

void OutTrm(uchar c)
{
  int i;
  switch (c)
  {
  case 0x0D:
  case '\n':
  case 0xF0: /* CRLF */
    ++cursorline;
    cursorcol = 0;
    break;
  case 0x0C: /* Page = ClearScreen */
    for (i = 0; i < TRMCHARS; ++i)
      trmbuffer[i] = ' ';
    NFlags ();
    firstline = 0;
    cursorline = 0;
    cursorcol = 0;
    break;
  case 0xF1: /* DEL */
    cursorprev ();
    trmbuffer[cursorline * TRMCOLS + cursorcol] = ' ';
    nflag[cursorline] = TRUE;
    if (cursorline < firstline)
    {
      firstline = cursorline;
      NFlags ();
    }
    break;
  case 0xF2: /* deleol */
    for (i = cursorcol; i < TRMCOLS; ++i) trmbuffer[cursorline * TRMCOLS + i] = ' ';
    nflag[cursorline] = TRUE;
    break;
  case 9: /* tab */
    trmbuffer[cursorline * TRMCOLS + cursorcol] = ' ';
    nflag[cursorline] = TRUE;
    cursornext ();
    while (cursorcol & 0x03)
    {
      trmbuffer[cursorline * TRMCOLS + cursorcol] = ' ';
      cursornext ();
    }
    break;
  case 0xF4: /* space */
    c = ' ';   /* ! must run through to default */
  default:
    trmbuffer[cursorline * TRMCOLS + cursorcol] = c;
    nflag[cursorline] = TRUE;
    cursornext ();
  }
  if (cursorline >= TRMLINES)
    rollup ();
  if (cursorline - firstline + 1 > trmsize)
  {
    NFlags ();
    firstline = cursorline - trmsize + 1;
  }
}

void OutputTrm (uchar* ob)
{
  int i;
  i = 0;
  while (ob[i] != 0)
    OutTrm (ob[i++]); /* output all chars */
  DisplayTrm ();    /* update display */
}
